#ifndef lint
static char rcsid_example_c[] = "\
@(#)  $Header: example.c,v 1.10 91/11/26 23:07:44 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint

/************************************************************************
 *									*
M*	Example of using ST-II Application Programming Interface	*
M*									*
M* Allows target (listener) to be created, and an origin to open an	*
M* ST-II stream to one or more targets, send data, and disconnect from	*
M* the stream.								*
M*									*
M* Burried in the mass of lines is the code of interest; it is marked	*
M* by comments delimited by / * * / (you know why the spaces are there	*
M* right :-).  The example is roughly divided into five parts:		*
M* 1) code to process the specification of a Target (process_target),	*
M* 2) coed to setup an ST-II socket, 3) code that a Target might use,	*
M* 4) code that an Origin might use, and 5) code to close things doen.	*
M*									*
M* Surrounding it is boiler plate for error processing, displaying the	*
M* interactions (data structures passed) between the example and ST-II,	*
M* interacting with the user for data or commands, waiting for the next	*
M* event that should be processed (select loop).			*
M*									*
 ************************************************************************/

/*
m* Status:
m*   Features:
m*	socket, bind, listem, a,cept, connect, accept, refuse, disconnect,
m*	reconnect via sendmsg, recvmsg, sendmsg, close, UserData on connect.
m*   Untested Features:
m*	UserData with accept, refuse, disconnect.
m*   Restrictions/Bugs:
m*	Receiver looses data at high rates because printout takes too long.
m*   Things to do:
m*	Illustrate more features.
m*
 */

#include <errno.h>		/* System error codes */
#include <stdio.h>		/* STDIN, etc. */
#include <sys/time.h>		/* for struct timeval for select () */

/* Needed by ST-II applications */

#include <sys/types.h>		/* Required by sys/socket.h, fd_set */
#include <sys/socket.h>		/* Socket defns, including struct msg */
#include <sys/uio.h>		/* iov for struct msg */
#include "st2_api.h"		/* <netinet/st2_api.h> */

/* From ST2Print, only needed for verbose printouts */

extern char *LookupSTError (/*enum ST2Errors error*/);
extern int PrintMsg (/*struct msghdr *msgp, int datalen, char *prefix*/);
extern int PrintParms (/*struct aParameter *parmp, int len, char *prefix*/);
extern int PrintSockaddrST (/*struct sockaddr_st2 *sap, int len, char *prefix*/);
extern int PrintTarget (/*struct aTarget *tp, int len, char *prefix*/);

/* Locally defined routines */

struct aParameter *FindParm (/*pcode, parmsp, parmslen*/);
octet4	 lookupIPAdr (/*hostp, textp*/);
void	 process_args (/*ac, app */);
int	 process_target ( /* tsp */ );
void	 usage ();

/* From some include file which we don't need */

char	*index ();
void	 exit ();

/* Misc parameters for this example program */

#define MAX_SR 5	/* Max number of Source Route parameters per Target */
#define MAX_SR_HOPS 10	/* Max number of Source Route hops per Target */
#define MAX_TARGETS 10	/* Max number of Targets per stream */
#define MAX_SOCKETS 20	/* Max number of open sockets */

#define MIN_BYTES_PER_PACKET 1		/* Data packet limits */
#define DEFAULT_BYTES_PER_PACKET 512
#define MAX_BYTES_PER_PACKET 8192

#define MIN_MSEC_BETEEN_PKTS 1		/* Inter-packet interval limits */
#define DEFAULT_MSEC_BETEEN_PKTS 100
#define MAX_MSEC_BETEEN_PKTS 10000

#define MIN_PKTS 1			/* Packet burst limits */
#define DEFAULT_PKTS 100
#define MAX_PKTS 10000

#define MIN_PROTOCOL 0			/* Valid protocol numbers */
#define DEFAULT_PROTOCOL 63
#define MAX_PROTOCOL 255

/* The example uses two byte SAPs which are identical to IP ports */
#define	SAPLEN 2			/* Length of SAP in this example */
#define Bytesof(x) Bytes2 (x)		/* !@#$ cpp lacks an "eval" */

#define MIN_PORT 0			/* Valid SAP values */
#define DEFAULT_PORT 1234
#define MAX_PORT ((1 << (8*SAPLEN))-1)


/* Datastructures for Example - ST-II interaction */

/*****************************************************************************
 * The local structure is the sockaddr structure for the bind system call.   *
 * It creates a sockaddr_st2 structure named "local" containing one          *
 * aApplEntity parameter (named "local.here").  It is initialized as an      *
 * STReqBind request; the other fields (IPAdr, NextPcol, and SAP) will be    *
 * filled in after the command line has been processed, but are initialized  *
 * here to the default values as an example.		                     *
 *****************************************************************************/

Instsockaddr_st2 (local,InstaApplEntity (here,SAPLEN,/*no srcrut*/);) = {
    Initsockaddr_st2 (STReqBind,0/*opt_xxx*/),
    InitaApplEntity (STLclAppEnt,local.here,
	INADDR_ANY,DEFAULT_PROTOCOL,SAPLEN,Bytesof (DEFAULT_PORT),/*no srcrut*/)
};


/*****************************************************************************
 * The remote structure is the sockaddr structure for the connect system     *
 *  call.  It creates a sockaddr_st2 structure named "remote" containing one *
 * aFlowSpec3 parameter to specify the desired resources and one aTargetList *
 * parameter containing a single aTarget parameter whose length is large     *
 * enough to hold the targets and userdata specified by the user.  The       *
 * targets are filled in dynamically as the command line is parsed, adding   *
 * targets and source route parameters as required.  Later, the number of    *
 * targets and length of the target list is filled in and an optional        *
 * aUserData parameter will be added if required.  The total length of the   *
 * remote structure is computed.  The minimum resources requested are 128    *
 * byte packets at a rate of two packets a second (256 bytes per second) but *
 * can generate packets of 1024 bytes at a rate of ten packets per second.   *
 *****************************************************************************/

Instsockaddr_st2 (remote,
	InstaFlowSpec3 (tos);
	InstaTargetList (targlst,
	    InstaTarget (targ1,4*MAX_TARGETS*(2 + MAX_SR + MAX_SR_HOPS + 1),);
	);
) = {
    Initsockaddr_st2 (STReqConnect,0/*opt_xxx*/),
    InitaFlowSpec3 (STpFlowSpec,
		    /*min*/  128/*bytes*/,  2/*pps*/, 128*2/*bandwidth*/,
		    /*des*/ 1024/*bytes*/, 10/*pps*/),
    InitaTargetList (remote.targlst,1/*targets*/,
	InitaTarget (remote.targlst.targ1,
		INADDR_ANY,SAPLEN,Bytesof (DEFAULT_PORT),)
    )
};


Instsockaddr_st2 (addon,
	InstaFlowSpec3 (adtos);
	InstaTargetList (adtarglst,
	    InstaTarget (adtarg1,4*MAX_TARGETS*(2 + MAX_SR + MAX_SR_HOPS + 1),);
	);
) = {
    Initsockaddr_st2 (STReqConnect,0/*opt_xxx*/),
    InitaFlowSpec3 (STpFlowSpec,
		    /*min*/  128/*bytes*/,  2/*pps*/, 128*2/*bandwidth*/,
		    /*des*/ 1024/*bytes*/, 10/*pps*/),
    InitaTargetList (addon.adtarglst,1/*targets*/,
	InitaTarget (addon.adtarglst.adtarg1,
		INADDR_ANY,SAPLEN,Bytesof (DEFAULT_PORT),)
    )
};

/*****************************************************************************
 * The sendname structure is the sockaddr structure for the sendmsg system   *
 * call.  It creates a sockaddr_st2 structure named "sendname".  No          *
 * parameters are necessary since data sent through an ST-II stream is       *
 * delivered to all the targets which have accepted the connection.          *
 *****************************************************************************/

Instsockaddr_st2 (sendname,) = {
    Initsockaddr_st2 (STReqUnspec,0/*opt_xxx*/)
};


/*****************************************************************************
 * The refuse structure is the cmsghdr structure used by a target to refuse  *
 * an ST-II connection request from an origin, or later by a target which    *
 * has previously accepted a connection to disconnect itself from the        *
 * connection.  The example below uses the ApplDisconnect reason code.       *
 *****************************************************************************/

Instcmsghdr (refuse,InstaReasonCode (why1);) = {
    Initcmsghdr (refuse,STReqRefuse),
    InitaReasonCode (ApplDisconnect)
};


/*****************************************************************************
 * The discname structure is the sockaddr structure for the sendmsg system   *
 *  call.  It creates a sockaddr_st2 structure named "discname" containing a *
 * single Remote Application Entity parameter which the origin will use to   *
 * identify the target it wishes to disconnected from the stream.            *
 *****************************************************************************/

Instsockaddr_st2 (discname,InstaApplEntity (targ,SAPLEN,);) = {
    Initsockaddr_st2 (STReqDisconnect,0/*opt_xxx*/),
    InitaApplEntity (STRmtAppEnt,discname.targ,INADDR_ANY,DEFAULT_PROTOCOL,
		     SAPLEN,Bytesof (DEFAULT_PORT),/*no srcrut*/)
};


/*****************************************************************************
 * The disc structure is the cmsghdr structure used by an origin to specify  *
 * the reason code to be passed to the target which the origin is            *
 * disconnecting.							     *
 *****************************************************************************/

Instcmsghdr (disc,InstaReasonCode (why2);) = {
    Initcmsghdr (disc,STReqDisconnect),
    InitaReasonCode (ApplDisconnect)
};


/* ST-II related Variables */

struct msghdr		 msg;		/* For recvmsg/sendmsg ... */
	
union {					/* First component of msg */
    struct sockaddr_st2	 sa_st2;
    char		 buf[ MAX_ST_NAM ]; /* For socket name info */
}			 name;
struct			 sockaddr_st2 *namep; /* Contents of namebuf */

					/* Second component of msg */
struct iovec		 dataiov,	/* For send/receive data descriptor */
			*iovecp;	/* Ptr to a data descriptor */

					/* For receive data */
char			 databuf[ 2048 ],
					 /* Send data */
			 pktgenbuf[ MAX_BYTES_PER_PACKET ]
			 = "00000 This is a data buffer.";

					/* Third component of msg */
union {					/* First component of msg */
    struct cmsghdr	 cm;
    char		 buf[ MAX_ST_CTL ]; /* For control info from ST-II */
}			 ctl;
struct cmsghdr		*ctlp;		/* Contents of ctlbuf */

struct aApplEntity	*aep;		/* Ptr to an Application Entity parm */
struct aDetectorIPAdr	*adiap;
struct aParameter	*parmsp,	/* Ptr to parameter list */
			*parmp;		/* Ptr to a parameter */
struct aReasonCode	*arcp;
struct aTarget		*tp;
struct aTargetList	*tlp;


/* Example/boiler plate Variables */

char			*arg0p,		/* Pointer to program name */
			*cmdp,		/* User command */
			 ourname[256],	/* Name of local host */
			 ttybuf[2048],	/* User response */
			*userdatap;	/* User data, if userdatalen > 0 */
fd_set			 rfds;		/* Fds selected for reading */
int			 Bflag = 0,	/* Omit bind() switch -b */
			 Hflag = 0,	/* Detailed usage switch -h */
			 Vlevel = 0,	/* Level of verbosity, -v switch */
			 listener = 0,	/* We are a listener */
			 nconnected = 0, /* Number of connected targets */
			 no_recovery = 0,/* No recovery on stream -N switch */
			 nsockets = 0,	/* Number of active sockets */
			 ntargets = 0,	/* Number of targets */
			 origin = 0,	/* We are the origin switch -o */
			 pktbytes = DEFAULT_BYTES_PER_PACKET,
			 pktcount = 0,	/* Count of data packets to be sent */
			 pktmsec = DEFAULT_MSEC_BETEEN_PKTS,
			 port = DEFAULT_PORT,
			 protocol = DEFAULT_PROTOCOL,
			 reuseaddr = 0,	/* default is to not SO_REUSEADDR */
			 single_socket = 0, /* Single socket mode (not yet) */
			 sockets[ MAX_SOCKETS+1 ], /* Table of open sockets */
			/* [0] is listener & [1] is unused, listener only */
			/* [0] is unused   & [1] is origin, origin only */
			/* [0] is listener & [1] is origin, if both */
			 Tflag = 0,	/* Use ST-II Timestamps */
			 userdatalen = 0, /* Length of optional User Data */
			 useselect = 1,	/* Use select instead of waiting */
			 wflag = 0;	/* Use write () instead of sendmsg () */
octet4			*freeTargetp = (octet4 *) &( remote.targlst.targ1 ),
					/* Where to build the Target List */
			*protoendp = (octet4 *) ((long) &( remote )
						 + sizeof (remote) - 4);
					/* Where Target List space ends */
			 ouraddr = INADDR_ANY; /* Our host address */

struct aremote {			/* Parallel sockets[] */
    unsigned long	 name[4];	/* Remote Application Entity param */
}			 remotes[ MAX_SOCKETS+1 ];

enum targstate {
    TargStateNew,
    TargStateConnecting,
    TargStateConnected,
    TargStateClosing,
    TargStateClosed
};

struct targinfo {			/* Filter duplicates at origin */
    enum targstate	 state;
    struct aTarget	*p;
}			targets[ MAX_TARGETS + 1 ];

struct timeval	 poll =	{		/* For select & packet burst timing */
    DEFAULT_MSEC_BETEEN_PKTS / 1000, (DEFAULT_MSEC_BETEEN_PKTS % 1000) / 1000
};


/*
|* int main ( argc, argpp, envpp )
|* Usage:
|*	Example of how to use ST-II via a socket interface.
|* Arguments:
|*	argc	argument count
|*	argpp	pointer to arguemnt list
|*	envpp	pointer to environment variable list (unused)
|*
*/

int
main( argc, argpp, envpp )		/* ARGSUSED */
int			  argc;		/* Count of arguments */
char			**argpp,	/* Pointer to argument list */
			**envpp;	/* Pointer to environment variables */
{
    char		*cp,		/* Ptr into command string */
			*prompt = "No prompt.", /* For replies from user */
			*sendp;		/* Ptr to data to be sent */
    unsigned char	*ucp;
    int			 error,		/* Error, from system call */
			 i,		/* Counter */
			 j,		/* Counter */
			 namelen,	/* Length of a name (sockaddr) */
			 pkth = 0,	/* Packet burst count, hundreds */
			 pktj = 0,	/* Packet burst count, ten thousands */
			 pktk = 0,	/* Packet burst count, thousands */
			 pktt = 0,	/* Packet burst count, tens */
			 pktu = 0,	/* Packet burst count, ones */
    			 recvlen,	/* Length of received info */
			 remotelen,	/* Length of remote parameter */
#define NOCONNECTOUT 99
			 repliestogo,	/* # outstanding connects */
			 s = -1,	/* Origin or Target/listening socket */
			 sendlen,	/* Length of data to be sent */
			 si,		/* Index into sockets[] table */
			 stdinclosed = 0, /* STDIN was closed */
			 tx,		/* A target index */
			 wait = 1;	/* Wait for user to abort connection */


    arg0p = *argpp;			/* Remember command name */

    process_args( argc, argpp );	/* Process command line arguemnts */

    if ( ouraddr == 0 )
    {
	/* Find name/address of local host */
	if ( gethostname( ourname, sizeof (ourname) ) < 0 )
	{
	    perror( "\nGethostname failed" );
	    exit (1);
	}

	ouraddr = lookupIPAdr( ourname, "Local agent: " );
    }


    if ( single_socket )		/* Not yet supported */
    {
	local.st_options |= STOpt1socket;
    }

/****  Fill in sockaddr_st2 structure to be used later with bind call.	****/
/**/									 /**/
/**/    local.here.IPAdr = ouraddr;					 /**/
/**/    local.here.NextPcol = protocol;					 /**/
/**/    local.here.SAP[0] = port >> 8;					 /**/
/**/    local.here.SAP[1] = port & 0xFF;				 /**/
/**/									 /**/
/***************************************************************************/

    /* Set FlowSpec parameters based on packet size and rate.  Units in
     * aFlowSpec3 are bytes and packets per 10 seconds.
     */

    if ( (i = pktbytes) <= 0 )		i = 1;
    if ( (j = 10000 / pktmsec) <= 0 )	j = 1;

    remote.tos.LimitOnPDUBytes = i;
    remote.tos.DesPDUBytes     = i;
    remote.tos.LimitOnPDURate  = j;
    remote.tos.DesPDURate      = j;
    remote.tos.MinBytesXRate   = i * j;

    if ( Tflag != 0 )			/* Timestamp PDUs if possible */
	remote.st_options = (remote.st_options & ~ STTSR) | STTSRYes;

    /* If targets were specified (origin), finish filling in the sockaddr_st2
     * structure (remote) to be specified with the connect call.  If user data
     * was specified, include it, too.
     */

    if ( origin )
    {
/**********  Finish parameter for connect, including target list.  *********/
/**/									 /**/
/**/	remote.targlst.TargetCount = ntargets;				 /**/
/**/	remote.targlst.plen =     (long) freeTargetp			 /**/
/**/				- (long) &( remote.targlst );		 /**/
/**/									 /**/
	/**** Append UserData if specified ****/
	if ( userdatalen > 0 )
	{
	    struct aUserData *udp = (struct aUserData *)freeTargetp;	    
	    char *cp = (char *) &( udp->userdata[0] );
	    int i = (long) protoendp - (long) cp;
	    if ( (userdatalen + 3) > i )
	    {
		(void) printf(
		   "\nError: Too much user data for connect, must split.\n" );
		exit ( 1 );
	    }
	    bcopy( userdatap, /* -> */ cp, userdatalen );
	    udp->userlen = /*hton2*/(userdatalen);
	    /* Maintain alignment */
	    i = (cp - (char *) udp) + userdatalen + 3;
	    i &= ~ 3;
	    if ( i > 252 )
	    {
		(void) printf(
	"\nError: Too much user data for connect, must split target list.\n" );
		exit ( 1 );
	    }
	    udp->plen = i;
	    udp->pcode = STpUserData;
	    freeTargetp = (octet4 *) ((long) freeTargetp + i);
	}
/**/									 /**/
/**/	remotelen = (long) freeTargetp - (long) &( remote );		 /**/
/**/	if ( remotelen >= MAX_ST_NAM )					 /**/
/**/	{								 /**/
/**/	    (void) printf(						 /**/
/**/ "\nError: Too many targets or userdata for connect, must split target list.\n" );
/**/	    exit ( 1 );							 /**/
/**/	}								 /**/
/**/									 /**/
/***************************************************************************/


	/* Standard socket interface will not allow a connect on a
	 * listening socket.  We would like to do that someday.
	 * For now, open a second socket here for the connect call.
	 */

/************************  Get an ST-II socket.  ***************************/
/**/									 /**/
/**/	s = socket( PF_COIP, SOCK_RAW, /*prot*/0 );			 /**/
/**/									 /**/
/***************************************************************************/
	if ( s == -1 )
	{
	    perror( "\nSocket failed" );
	    exit ( 1 );
	}
	else
	{
	    if ( Vlevel > 1 )
	    {
		(void) printf( "\nSocket for origin is %d.", s );
	    }
	}

	sockets[1] = s;			/* Socket for connect */
	nsockets++;

	/* ---------------------------------------------------------------- */
	if ( reuseaddr )
	{
	    error = setsockopt( s, SOL_SOCKET, SO_REUSEADDR, &reuseaddr,
			        sizeof(reuseaddr) );
	    if ( error == -1 )
	    {
		perror( "\nsetsockopt SO_REUSEADDR failed" );
		exit ( 1 );
	    }
	}

	/* ---------------------------------------------------------------- */
	if ( ! Bflag )			/* If user wants to bind, do so */
	{
	    if ( Vlevel > 1 )
	    {
		(void) printf(
	"\nBinding local address: %d.%d.%d.%d protocol %d, port %d, options %02x",
			(0xFF & (ouraddr >> 24)), (0xFF & (ouraddr >> 16)),
			(0xFF & (ouraddr >> 8)), (0xFF & ouraddr),
			protocol, port, local.st_options );
	    }

	    if ( Vlevel > 2 )
	    {
		(void) PrintSockaddrST( (struct sockaddr_st2 *) &local,
				       sizeof (local), "    " );
	    }

	    if ( Vlevel > 1 )
	    {
		(void) fflush( stdout );
	    }

/**************************  Bind ST-II socket.  ***************************/
/**/									 /**/
/**/	    error = bind( s, (struct sockaddr *) &local,		 /**/
/**/			 sizeof (local) );				 /**/
/**/									 /**/
/***************************************************************************/
	    if ( error == -1 )
	    {
		perror( "\nBind failed" );
		exit ( 1 );
	    }
	} /* end of want to bind origin */
    } /* end of origin */

    repliestogo = NOCONNECTOUT;		/* Assume just listening */


    /* Listener (or neither specified defaults to listener) */
    if ( ! origin || listener )
    {
/************************  Get an ST-II socket.  ***************************/
/**/									 /**/
/**/	s = socket( PF_COIP, SOCK_RAW, /*prot*/0 );			 /**/
/**/									 /**/
/***************************************************************************/
	if ( s == -1 )
	{
	    perror( "\nSocket failed" );
	    exit ( 1 );
	}
	else
	{
	    if ( Vlevel > 1 )
	    {
		(void) printf( "\nSocket returned %d.", s );
		(void) fflush( stdout );
	    }
	}

	sockets[0] = s;			/* Save listen socket info */
	nsockets++;

	/* ---------------------------------------------------------------- */
	if ( reuseaddr )
	{
	    error = setsockopt( s, SOL_SOCKET, SO_REUSEADDR, &reuseaddr,
			        sizeof(reuseaddr) );
	    if ( error == -1 )
	    {
		perror( "\nsetsockopt SO_REUSEADDR failed" );
		exit ( 1 );
	    }
	}

	/* ---------------------------------------------------------------- */
	if ( ! Bflag )			/* If user wants to bind, do so */
	{

	    if ( Vlevel > 1 )
	    {
		(void) printf(
  "\nBinding local address: %d.%d.%d.%d protocol %d, port %d, options %02x",
			(0xFF & (ouraddr >> 24)), (0xFF & (ouraddr >> 16)),
			(0xFF & (ouraddr >> 8)), (0xFF & ouraddr),
			protocol, port, local.st_options );
		(void) fflush( stdout );
	    }

	    if ( Vlevel > 2 )
	    {
		(void) PrintSockaddrST( (struct sockaddr_st2 *) &local,
				       sizeof (local), "    " );
	    }

/**************************  Bind ST-II socket.  ***************************/
/**/									 /**/
/**/	    error = bind( s, (struct sockaddr *) &local,		 /**/
/**/			  sizeof (local) );				 /**/
/**/									 /**/
/***************************************************************************/
	    if ( error == -1 )
	    {
		perror( "\nBind failed" );
		exit ( 1 );
	    }
	} /* end of wanted to bind */



	if ( ! single_socket )		/* Example sanity chaecking */
	{
	    if ( ! useselect )
	    {
		(void) printf(
   "\nWarning: Must use selects when using multiple sockets; continuing ..." );
	    }
	    useselect = 1;
	}

/***********************  Listen on ST-II socket.  *************************/
/**/									 /**/
/**/	error = listen( s, 5 );						 /**/
/**/									 /**/
/***************************************************************************/
	if ( error == -1 )
	{
	    perror( "\nListen failed" );
	    exit ( 1 );
	}

	if ( Vlevel > 0 )
	{
	    (void) printf( "\nListen returned." );
	}

	prompt = "Waiting for receipt of a connect, data, or a command ... ";
	(void) printf( "\n%s", prompt );
	(void) fflush( stdout );

	/* Now loop for events to be processed */
	while (1)
	{
	    /* Use select to wait for next event */
	    if ( useselect )
	    {
		while (1)		/* A real OS wouldn't have EINTRs */
		{
		    FD_ZERO(&rfds);	/* Clear select mask */
		    if ( ! stdinclosed )
		    {
			FD_SET(0,&rfds); /* STDIN - user input */
		    }
		    for ( i = nsockets - 1 ; i >= 0 ; i-- )
		    {
			if ( sockets[i] <= 0 )
			{
			    continue;	/* unused (closed) entry */
			}
			FD_SET(sockets[i],&rfds); /* ST-II socket */
		    }
		    error = select( 32, &rfds, (fd_set *) NULL,
				   (fd_set *) NULL, (struct timeval *) NULL );
		    if ( (error == -1) && (errno == EINTR) )
		    {
			if ( Vlevel > 2 )
			{
			    (void) printf( "." ); /* Maybe the mouse moved */
			    (void) fflush( stdout );
			}
		    }
		    else
		    {
			break;		/* out of EINTR loop */
		    }
		} /* end of EINTR loop */

		if ( error == -1 )
		{
		    perror( "\nSelect failed" );
		    exit ( 1 );
		}

		/* STDIN - Maybe user typein ------------------------------ */
		if ( FD_ISSET( 0, &rfds ) )
		{
		    /* EOF ------------------------------------------------ */
		    if ( feof(stdin) )
		    {
			stdinclosed = 1;
			(void) printf(
		"\n*** Command input closed, now what are you going to do?\n" );
			(void) printf( "*** We'll now accept anything!\n" );
			(void) fflush( stdout );
			continue;
		    }

		    /* CR ------------------------------------------------- */
		    (void) gets( ttybuf );	/* Read user typein */
		    i = strlen( ttybuf );

		    if ( i <= 0 )
		    {
			(void) printf( prompt ); /* Just hit return so reprompt */
			(void) fflush( stdout );
			continue;
		    }

		    /* Command: ?, help ----------------------------------- */
		    if (   (    (cmdp = "?") /* side effect */
			    && (strncmp( ttybuf, cmdp, i ) == 0))
			|| (    (cmdp = "help" ) /* side effect */
			    && (strncmp( ttybuf, cmdp, i ) == 0)) )
		    {
			(void) printf( "\nCommands are:\n\
 ?                    Print this message.\n\
 close                Close all connections.\n\
 disconnect           Print a list of sockets.\n\
 disconnect <socket>  Disconnect specified socket.\n\
 exit                 Close all connections.\n\
 help                 Print this message.\n\
 quit                 Close all connections.\n%s", prompt );
			(void) fflush( stdout );
			continue;
		    }

		    /* Command: close ------------------------------------- */
		    if (   (cmdp = "close") /* side effect */
			&& (strncmp( ttybuf, cmdp, i ) == 0) )
		    {
			if ( Vlevel > 0 )
			{
			    (void) printf( "\nReceived %s command.", cmdp );
			}
			break;		/* out of while loop */
		    }

		    /* Command: exit, quit -------------------------------- */
		    if (   (   (cmdp = "exit") /* side effect */
			    && (strncmp( ttybuf, cmdp, i ) == 0))
			|| (   (cmdp = "quit") /* side effect */
			    && (strncmp( ttybuf, cmdp, i ) == 0)) )
		    {
			if ( Vlevel > 0 )
			{
			    (void) printf( "\nReceived %s command.", cmdp );
			}
			wait = 0;
			break;		/* out of while loop */
		    }

		    /* Command: connect ----------------------------------- */
		    /* User may want to connect (not yet) */
		    if (    origin && (repliestogo == NOCONNECTOUT)
			&& (strncmp( ttybuf, "connect", i ) == 0) )
		    {
			if ( Vlevel > 0 )
			{
			    (void) printf( "\nConnecting." );
			}
			if ( Vlevel > 1 )
			{
			    (void) PrintSockaddrST(
					(struct sockaddr_st2 *) &remote,
						   remotelen, "    " );
			}

/****************************  ST-II Connect.  *****************************/
/**/									 /**/
/**/			error = connect( sockets[1],			 /**/
/**/					(struct sockaddr *) &( remote ), /**/
/**/					remotelen );			 /**/
/**/									 /**/
/***************************************************************************/
			if ( error == -1 )
			{
			    perror( "\nConnect failed" );
			    (void) fflush( stdout );
			    continue;
			}

			for ( i = ntargets - 1 ; i >= 0 ; i-- )
			{
			    targets[ i ].state = TargStateConnecting;
			}

			repliestogo = ntargets;

			ttybuf[0] = '\0';
			prompt = "Waiting for connect reply or command ... ";
			(void) printf( prompt );
			(void) fflush( stdout );
			continue;	/* Back for next event */
		    }

		    /* Command with argument ------------------------------ */
		    cp = index( ttybuf, ' ' );
		    if ( cp != (char *) NULL )
		    {
			i = cp - ttybuf - 1; /* characters to first space */
		    }

		    /* Command: disconnect ("refuse") --------------------- */
		    if (   (   (cmdp = "disconnect") /* side effect */
			    && (strncmp( ttybuf, cmdp, i ) == 0)) )
		    {
			if ( Vlevel > 0 )
			{
			    (void) printf( "\nReceived %s ", cmdp );
			}

			cmdp = index( ttybuf, ' ' );
			/* may specify which origin (socket) to disconnect */
			if (   (cp++ != (char *) NULL)
			    && ('0' <= *cp) && (*cp <= '9') )
			{
			    s = atoi( cmdp );
			    if ( Vlevel > 0 )
			    {
				(void) printf( "socket %d command.", s );
			    }
			    if ( s > 0 ) /* Specified a socket */
			    {
				/* Look for specified socket to disconnect */
				for ( i = nsockets ; i >= 0 ; i-- )
				{
				    if ( sockets[ i ] == s )
				    {
					break; /* out of for loop */
				    }
				} /* end of for all sockets loop */
				if ( i < 0 )
				{
				    (void) fflush( stdout );
				    continue; /* Back for next event */
				}

				/**** ST-II Target Disconnect setup ****/
				namep = &( name.sa_st2 );
				namep->sa_family = AF_COIP;
				namep->st_options = 0;
				namep->st_request = STReqRefuse;
				parmp = (struct aParameter *)
					&( remotes[ i ].name[0] );
				(void) bcopy( (char *) parmp, /*->*/
					     (char *) &( namep->st_parms[0] ),
					     parmp->plen );
				InitMsg (&(msg),namep,sizeofsockaddr_st2header + parmp->plen,
					0,0,&(refuse),sizeof (refuse));

				if ( Vlevel > 2 )
				{
				    (void) printf( "\nSending: " );
				    (void) PrintMsg( &msg, -1, "    " );
				}

				/****  ST-II Target disconnect (a refuse)  ****/
				error = sendmsg( s, &msg, 0/*flags*/ );
				if ( error == -1 )
				{
				    perror( "\nSendmsg failed" );
				    (void) printf( "\n%s", prompt );
				    (void) fflush( stdout );
				}

				(void) close( s );

				sockets[ i ] = -1;
				remotes[ i ].name[0] = 0;
				if ( (i > 0)  &&  (i == nsockets) )
				{
				    --nsockets;
				}
			    }
			}
			else		/* Not specify which to disconnect */
			{
			    if ( nsockets > 0 )
			    {
				(void) printf( "Connected sockets are:" );
				for ( i = nsockets - 1 ; i >= 0 ; i-- )
				{
				    if ( sockets[ i ] > 0 )
				    {
					parmp = (struct aParameter *)
						&( remotes[ i ].name[0] );
					if ( (int) parmp->pcode == STRmtAppEnt )
					{
					    (void) PrintParms( parmp, parmp->plen, "" );
					    (void) printf( " --> %d", sockets[ i ] );
					}
				    }
				} /* end of for loop */
			    }
			    else
				(void) printf( "\nNo connected sockets" );
			}
			(void) printf( "\n%s", prompt );
			(void) fflush( stdout );
			continue;	/* Back for next event */
		    }

		    /* Not a command, so send it as data ------------------ */
		    if ( repliestogo > 0 )
		    {
			ttybuf[0] = '\0';
			if ( ntargets > 0 )
			    (void) printf(
		"Cannot send data until replies have been received.\n%s",
					  prompt );
			else
			    (void) printf( "A half-duplex target cannot send data.\n%s",
					  prompt );
			(void) fflush( stdout );
			continue;	/* Back for next event */
		    }

		    /* Send ST-II data */
		    if ( wflag )
		    {
/*******************  Send data using write via ST-II.  ********************/
/**/									 /**/
/**/			error = write( s, ttybuf, strlen( ttybuf ) );	 /**/
/**/									 /**/
/***************************************************************************/
			if ( error == -1 )
			{
			    perror( "\nWrite failed" );
			    for ( i = nsockets ; i >= 0 ; i-- )
			    {
				if ( sockets[ i ] == s )
				{
				    (void) close( s );
				    sockets[ i ] = -1;
				    remotes[ i ].name[0] = 0;
				    break;	/* out of for loop */
				}
			    } /* end of for all sockets loop */
			}
			else if ( error != strlen( ttybuf ) )
			{
			    (void) printf( "Write: Sent %d of %d.",
					  error, strlen( ttybuf ) );
			}
		    }
		    else
		    {
/******************  Send data using sendmsg via ST-II.  *******************/
/**/									 /**/
/**/			/* Setup msg structure and data io vector */	 /**/
/**/			dataiov.iov_base = (caddr_t) ttybuf;		 /**/
/**/			dataiov.iov_len = strlen( ttybuf );		 /**/
/**/			InitMsg (&(msg),&(sendname),sizeof (sendname),	 /**/
/**/				&dataiov,1, NULL,0);			 /**/
/**/									 /**/
			if ( Vlevel > 2 )
			{
			    (void) printf( "Sending: " );
			    (void) PrintMsg( &msg, -1, "    " );
			}
/**/									 /**/
/**/			error = sendmsg( s, &msg, 0/*flags*/ );		 /**/
/**/									 /**/
/***************************************************************************/
			if ( error == -1 )
			{
			    perror( "\nSendmsg failed" );
			    for ( i = nsockets ; i >= 0 ; i-- )
			    {
				if ( sockets[ i ] == s )
				{
				    (void) close( s );
				    sockets[ i ] = -1;
				    remotes[ i ].name[0] = 0;
				    break;	/* out of for loop */
				}
			    } /* end of for all sockets loop */
			}
			else if ( error != strlen( ttybuf ) )
			{
			    (void) printf( "\nSendmsg: Sent %d. of %d.",
					  error, strlen( ttybuf ) );
			}
		    }

		    ttybuf[0] = '\0';	/* Flush user typein buffer */

		    prompt = "Waiting for data or command ... ";
		    (void) printf( prompt );
		    (void) fflush( stdout );
		    continue;		/* Back for next event */
		}

		/* Not STDIN, find socket with activity ------------------- */
		if ( Vlevel > 2 )
		{
		    (void) printf( "\nSelect returns %d, r=%08x",
				  error, * (int*) &( rfds ) );
		}

		/* Find ST-II socket with input and position in our table */
		for ( i = nsockets - 1 ; i >= 0 ; i-- )
		{
		    s = sockets[i];
		    if ( s <= 0 )
		    {
			continue;
		    }
		    if ( ! FD_ISSET( s, &rfds ) )
		    {
			continue;
		    }
		    si = i;		/* found socket in table */
		    break;		/* out of for loop */
		} /* end of for loop */

	    } /* end of using select */

	    /* slot 0 in table is for listener */
	    /* multiple socket listener */
	    if ( (! single_socket) && (si == 0) )
	    {
/**************  Target accepts connection from an origin.  ****************/
/**/									 /**/
/**/		namep = &(  name.sa_st2 );				 /**/
/**/		namelen = sizeof (name);				 /**/
/**/		s = accept( s, (struct sockaddr *) namep, &namelen );	 /**/
/**/									 /**/
/***************************************************************************/
		if ( s == -1 )
		{
		    perror( "\nAccept failed" );
		    exit ( 1 );
		}

		if ( Vlevel >= 1 )
		{
		    (void) printf( "\nAccept returns socket %d.", s );
		}

		if ( Vlevel > 1 )
		{
		    (void) PrintSockaddrST( &( name.sa_st2 ),
					   namelen, "    " );
		}

		/* find empty slot in table for new connection */
		for ( i = 1 ; i < MAX_SOCKETS ; i++ )
		{
		    if ( sockets[ i ] <= 0 )
		    {
			sockets[ i ] = s;
			si = i;
			if ( i >= nsockets )
			{
			    nsockets = i + 1;
			}
			break;		/* out of for loop */
		    }
		} /* end of seraching for free socket entry loop */
		if ( i >= MAX_SOCKETS )
		{
		    remotes[ i ].name[0] = 0;
		    (void) printf(
		"\nError: Too many open sockets (%d), closing new one.",
				  error );
		    (void) fflush( stdout );

/******************* Target aborts and ST-II connection ********************/
/**/									 /**/
/**/		    (void) close( s );					 /**/
/**/									 /**/
/***************************************************************************/
		    continue;		/* on to next event */
		}
	    }

/********************* Receipt of ST-II Control or Data ********************/
/**/									 /**/
/**/	    /* Setup msg structure and data io vector */		 /**/
/**/	    dataiov.iov_base = (caddr_t) databuf;			 /**/
/**/	    dataiov.iov_len = sizeof (databuf);				 /**/
/**/	    InitMsg (&(msg),&( name.sa_st2 ),sizeof (name),		 /**/
/**/		    &dataiov,1, &( ctl.cm ),sizeof (ctl));		 /**/
/**/									 /**/
/**/	    recvlen = recvmsg( s, &msg, 0/*flags*/ );			 /**/
/**/									 /**/
/***************************************************************************/
	    if ( recvlen == -1 )
	    {
		perror( "\nRecvmsg failed" );
		(void) printf( "\n%s", prompt );
		(void) fflush( stdout );
		for ( i = nsockets - 1 ; i >= 0 ; i-- )
		{
		    if ( sockets[ i ] == s )
		    {
			(void) close( s );
			sockets[ i ] = -1;
			remotes[ i ].name[0] = 0;
			break;		/* out of for loop */
		    }
		} /* end of searching for socket table entry loop */
	    }

	    if ( Vlevel > 1 )
	    {
		(void) printf( "\nReceived message: " );
		(void) PrintMsg( &msg, recvlen, "    ");
	    }

	    if ( msg.msg_namelen < sizeof (struct sockaddr_st2) )
	    {
		(void) printf( "\nRecvmsg: namelen too small (%d.)",
			      msg.msg_namelen );
		(void) fflush( stdout );
		continue;		/* Inconsistence, on to next event */
	    }

	    /**** Identify origin ****/
	    namep = &( name.sa_st2 );
	    namelen = msg.msg_namelen - sizeofsockaddr_st2header; /* Length of parameters */

	    /**** Unspecified info is data ****/
	    if ( namep->st_request == STReqUnspec )
	    {
		iovecp = msg.msg_iov;	/* Data vector */

		/* Make sure some data */
		if ( iovecp->iov_len > 0 )
		{
					/* We only suppied one buffer, */
					/* but code is more general */
		    for ( ; msg.msg_iovlen > 0 ; msg.msg_iovlen--, iovecp++ )
		    {
			i = iovecp->iov_len;
			if ( iovecp->iov_len < recvlen )
			{
			    (void) printf( "\nData from %s: %-*.*s", sockets[si],
				   i, i, iovecp->iov_base );
			    /* Vlevel */
			}
			else
			{
			    i = recvlen;
			    (void) printf( "\nData from %d: %-*.*s", sockets[si],
				   i, i, iovecp->iov_base );
			    /* Vlevel */
			    break;	/* out of for loop */
			}
			recvlen -= i;
		    } /* end of for all io vectors */
		}
		(void) printf( "\n%s", prompt );
	    }
	    else			/* Not STReqUnspec must be Control */
	    {
/*****  Processing ST-II Control Information in the "accrights" Field.  ****/
/**/									 /**/
/**/		/* "accrights" contains a cmsghdr structure */		 /**/
/**/		ctlp = &( ctl.cm );					 /**/
/**/									 /**/
/**/		if (   (msg.msg_accrightslen < sizeof (struct cmsghdr))	 /**/
/**/		    || (ctlp->cmsg_level != SOL_STII) )			 /**/
/**/		{							 /**/
/**/		    (void) close( s );					 /**/
/**/									 /**/
/***************************************************************************/
		    (void) printf(
	"\nError: recvmsg control len (%d) not msg_accrightslen(%d),\n\
       or cmsg_level (%d) not SOL_STII (%d).",
			   ctlp->cmsg_len, msg.msg_accrightslen,
				  ctlp->cmsg_level, SOL_STII );
		    remotes[ si ].name[0] = 0;
		    sockets[ si ] = -1;
		}

		if ( Vlevel >= 0 )
		{
		    (void) printf( "\nReceived control %d ", namep->st_request );
		}

		/**** ST-II parameter list follows cmsghdr ****/

/*******  Identity of Remote Application Entity is in sockaddr_st2.  *******/
/**/									 /**/
/**/		/* Locate Remote Entity parameter */			 /**/
/**/		parmsp = (struct aParameter *) &( namep->st_parms[0] );	 /**/
/**/		parmp = FindParm( STRmtAppEnt, parmsp, namelen );	 /**/
/**/									 /**/
/**/		/* Dispatch on Control request */			 /**/
/**/		switch ( namep->st_request )				 /**/
/**/		{							 /**/
/**/		case STReqConnect:					 /**/
/**/									 /**/
/***************************************************************************/
		    /* target receiving connect from an origin  */
		    if ( Vlevel >= 0 )
		    {
			/* Identify origin to user */
			(void) printf( "(STReqConnect) from" );
			if ( parmp == (struct aParameter *) NULL )
			{
			    (void) printf( " unspecified target" );
			}
			else
			{
			    (void) PrintParms( parmp, parmp->plen,
					      "        " );
			}
		    }

		    if ( feof(stdin) )
		    {
			(void) printf( "\n" );
			ttybuf[0] == 'n';
		    }
		    else
		    {
			/* Ask user if the connection should be accepted */
			if ( Vlevel >= 0 )
			{
			    (void) printf( "\nAccept connection? (y,n): " );
			}
			(void) gets( ttybuf );
		    }
		    if ( (ttybuf[0] == '\0') || (ttybuf[0] == 'y') )
		    {
/***********************  Accept ST-II connection.  ************************/
/**/									 /**/
/**/			namep->st_request = STReqAccept; /* Yes */	 /**/
/**/									 /**/
/***************************************************************************/

			if ( parmp != (struct aParameter *) NULL )
			{
			    (void) bcopy( (char *) parmp, /*->*/
					 (char *) &( remotes[ si ].name[0] ),
					 sizeof (remotes[ si ].name) );
			}		/* else trouble ... */
		    }
		    else
		    {
/***********************  Refuse ST-II connection.  ************************/
/**/									 /**/
/**/			namep->st_request = STReqRefuse; /* No */	 /**/
/**/			/* replace control info with refuse info */	 /**/
/**/			msg.msg_accrights = (caddr_t) &refuse;		 /**/
/**/			msg.msg_accrightslen = sizeof (refuse);	 /**/
/**/									 /**/
/***************************************************************************/
		    }
/**********************  Send ST-II Accept or Refuse.  *********************/
/**/									 /**/
/**/		    /* No data for control message. */			 /**/
/**/		    msg.msg_iovlen = 0;					 /**/
/**/		    error = sendmsg( s, &msg, 0/*flags*/ );		 /**/
/**/									 /**/
/***************************************************************************/
		    if ( error == -1 )
		    {
			perror( "\nSendmsg failed" );
			(void) printf( "\n%s", prompt );
			(void) fflush( stdout );
			(void) close( s );
			remotes[ si ].name[0] = 0;
			sockets[ si ] = -1;
		    }

		    /* If refusing connection, close the socket */
		    if ( namep->st_request == STReqRefuse )
		    {
			(void) close( s );
			remotes[ si ].name[0] = 0;
			sockets[ si ] = -1;
		    }
		    (void) printf( prompt );
		    break;		/* out of switch */

/***************************************************************************/
/**/									 /**/
/**/		case STReqDisconnect:					 /**/
/**/									 /**/
/***************************************************************************/
		    /* target receiving disconnect from the origin */
		    if ( si == 0 )
		    {
			if ( Vlevel >= 0 )
			{
			    (void) printf( "(STReqDisconnect) from" );
			}
			--nconnected;
		    }

		    /* target being disconnected by origin */
		    else
		    {
			if ( Vlevel >= 0 )
			{
			    (void) printf( "(STReqDisconnect) on socket %d from",
					  sockets[si] );
			}
		    }

		    (void) close( sockets[si] );

		    remotes[ si ].name[0] = 0;
		    sockets[ si ] = -1;
		    goto printfrom;

/***************************************************************************/
/**/									 /**/
/**/		case STReqAccept:					 /**/
/**/									 /**/
/***************************************************************************/
		    /* origin receiving accept from a target */
		    if ( Vlevel >= 0 )
		    {
			(void) printf( "(STReqAccept) from" );
		    }
		    ++nconnected;
		    --repliestogo;
		    goto printfrom;

/***************************************************************************/
/**/									 /**/
/**/		case STReqRefuse:					 /**/
/**/									 /**/
/***************************************************************************/
		    /* origin receiving refuse (initial or later) from target */
		    if ( Vlevel >= 0 )
		    {
			(void) printf( "(STReqRefuse) from" );
		    }
		    --repliestogo;	/* Initial vs subsequent counting bug */
		    goto printfrom;

		default:	/* Ought to inclue Notify ... */
		    (void) printf( " unexpected control ignored from" );
printfrom:	    /* Identify sender of control messagte */
		    if ( Vlevel >= 0 )
		    {
			adiap = (struct aDetectorIPAdr *) FindParm( STDetectorIPAdr,
			       (struct aParameter *) &( ctlp->cmsg_data[0] ),
			       ctlp->cmsg_len - sizeofCMsgHdr );

			if ( adiap != (struct aDetectorIPAdr *) NULL )
			{
			    ucp = (unsigned char *) &( adiap->IPAdr );
			    (void) printf( " %u.%u.%u.%u", ucp[0],ucp[1],
					   ucp[2],ucp[3] );
			}

			arcp = (struct aReasonCode *) FindParm( STReasonCode,
			       (struct aParameter *) &( ctlp->cmsg_data[0] ),
			       ctlp->cmsg_len - sizeofCMsgHdr );

			if ( arcp != (struct aReasonCode *) NULL )
			{
			    (void) printf( " (%u %s)", ntoh2 (arcp->reason),
				LookupSTError( (enum ST2Errors) ntoh2 (arcp->reason)) );
			}

			if ( parmp == (struct aParameter *) NULL )
			{
			    (void) printf( " unspecified target." );
			}
			else
			{
			    (void) PrintParms( parmp, parmp->plen, "        " );
			}
		    }
		    (void) printf( "\n%s", prompt );
		    break;		/* out of switch */
		} /* end of switch */

	    } /* end of control received */

	    (void) fflush( stdout );

	} /* end of while (1) receiving events loop */

    } /* end of listening */



    /* Here is an example of how an ST-II origin might be written */

    if ( origin )
    {
	/* turn on no recovery for a stream - error action to be determined */
	/* by individual parties on error detection                         */
	if ( no_recovery )  
	    remote.st_options |= STOptSBit;

	if ( Vlevel >= 0 )
	{
	    (void) printf( "\nConnecting." );
	}
	if ( Vlevel > 1 )
	{
	    (void) PrintSockaddrST( (struct sockaddr_st2 *) &remote, remotelen,
				   "    " );
	}

/************  ST-II Origin sending connect request to Targets.  ************/
/**/									 /**/
/**/	/* Setup of remote occurred in process_target and above */	 /**/
/**/	error = connect( s, (struct sockaddr *) &( remote ), remotelen );/**/
/**/									 /**/
/***************************************************************************/
	if ( error == -1 )
	{
	    perror( "\nConnect failed" );
	    exit ( 1 );
	}

	for ( i = ntargets - 1 ; i >= 0 ; i-- )
	{
	    targets[ i ].state = TargStateConnecting;
	}

	repliestogo = ntargets;		/* Number of replies we expect */
	prompt = "Waiting for replies to connect ... ";
	(void) printf( "\n%s", prompt );
	(void) fflush( stdout );

	/* Loop waiting for events to be processed */
	for (;;)		/* (nconnected > 0) || (repliestogo > 0) */
	{
	    /* Wait for a real event using select () */
	    while (1)			/* A real OS wouldn't have EINTRs */
	    {
		/* Setup for select */
		FD_ZERO(&rfds);		/* Clear select mask */
		if ( ! stdinclosed )
		{
		    FD_SET(0,&rfds);	/* STDIN - user input */
		}
		FD_SET(s,&rfds);	/* Origin's socket */

		/* If able to send data (received accept/refuse replies
		   from all targets) and sending canned data, wait for
		   no more than inter packet interval */

		if ( (pktcount > 0) && (repliestogo <= 0) )
		    error = select( 32, &rfds, (fd_set *) NULL, (fd_set *) NULL,
				   &poll );
		else			/* Otherwise, wait forever */
		    error = select( 32, &rfds, (fd_set *) NULL, (fd_set *) NULL,
				   (struct timeval *) NULL );

		if ( (error == -1) && (errno == EINTR) )
		{
		    if ( Vlevel > 2 )
		    {
			(void) printf( "." );	/* Maybe the mouse moved */
			(void) fflush( stdout );
		    }
		}
		else
		{
		    break;		/* out of EINTR loop */
		}
	    } /* end of EINTR loop */

	    if ( error == -1 )
	    {
		perror( "\nSelect failed" );
		break;			/* out of forever loop */
	    }
	    if ( Vlevel > 2 )
	    {
		(void) printf( "\nSelect returns %d, r=%08x", error,
			      * (int*) &( rfds ) );
		(void) fflush( stdout );
	    }

	    sendp = (char *) NULL;	/* Assume not sending any data */

	    if ( error == 0 )		/* Timer for canned data expired */
	    {
		pktcount--;		/* One less packet to send */
		if ( ++pktu > 9 )	/* Insert ascii packet number in data */
		{
		    pktu = 0;
		    if ( ++pktt > 9 )
		    {
			pktt = 0;
			if ( ++pkth > 9 )
			{
			    pkth = 0;
			    if ( ++pktk > 9 )
			    {
				pktk = 0;
				if ( ++pktj > 0 )
				{
				    pktj = 0;
				}
				pktgenbuf[0] = '0' + pktj;
			    }
			    pktgenbuf[1] = '0' + pktk;
			}
			pktgenbuf[2] = '0' + pkth;
		    }
		    pktgenbuf[3] = '0' + pktt;
		}
		pktgenbuf[4] = '0' + pktu;

		/* Set data pointer to canned data */
		sendp = (char *) pktgenbuf;
		sendlen = pktbytes;	/* User requested data length */
	    }

	    else if ( FD_ISSET( s, &rfds ) ) /* received ST-II Control */
	    {
/***********************  Receipt of ST-II Control.  ***********************/
/**/									 /**/
/**/		/* Setup msg structure control only */			 /**/
/**/		InitMsg (&(msg),&( name.sa_st2 ),sizeof (name),NULL,0,	 /**/
/**/			&( ctl.cm ),sizeof (ctl));			 /**/
/**/									 /**/
/**/		error = recvmsg( s, &msg, 0/*flags*/ );			 /**/
/**/									 /**/
/***************************************************************************/
		if ( error == -1 )
		{
		    perror( "\nRecvmsg failed" );
		    (void) printf( "\n%s", prompt );
		    (void) fflush( stdout );
		    for ( i = nsockets - 1 ; i >= 0 ; i-- )
		    {
			if ( sockets[ i ] == s )
			{
			    (void) close( s );
			    sockets[ i ] = -1;
			    remotes[ i ].name[0] = 0;
			    break;	/* out of for loop */
			}
		    } /* end of searching for socket loop */
		    continue;		/* on to next event */
		}

		if ( Vlevel > 1 )
		{
		    (void) printf( "\nResponse: " );
		    (void) PrintMsg( &msg, error, "    " );
		}

		/* Consistency check for contro info */
		if ( msg.msg_accrightslen > 0 )
		{
/*****  Processing ST-II Control Information in the "accrights" Field.  ****/
/**/									 /**/
/**/		    /* "accrights" contains a cmsghdr structure */	 /**/
/**/		    ctlp = &( ctl.cm );					 /**/
/**/									 /**/
/**/		    if ( ctlp->cmsg_level != SOL_STII )			 /**/
/***************************************************************************/
		    {
			(void) printf(
			"\nError: recvmsg control (%d) not SOL_STII (%d)",
				      ctlp->cmsg_level, SOL_STII );
			(void) fflush( stdout );
			continue;
		    }

		    /* ctlp->cmsg_type is identical to namep->st_request */
		    /**** Dispatch on Control request ****/
		    switch ( ctlp->cmsg_type )
		    {
		    case STReqAccept:	/* Target accepted connection */
		    case STReqRefuse:	/* Target refused connection */
			if ( Vlevel >= 0 )
			{
			    (void) printf( "\n%s received from",
				(ctlp->cmsg_type == STReqAccept)
				? "ACCEPT" : "REFUSE" );
			}

			adiap = (struct aDetectorIPAdr *) FindParm( STDetectorIPAdr,
			       (struct aParameter *) &( ctlp->cmsg_data[0] ),
			       ctlp->cmsg_len - sizeofCMsgHdr );

			if ( adiap != (struct aDetectorIPAdr *) NULL )
			{
			    ucp = (unsigned char *) &( adiap->IPAdr );
			    (void) printf( " %u.%u.%u.%u", ucp[0],ucp[1],
					   ucp[2],ucp[3] );
			}

			arcp = (struct aReasonCode *) FindParm( STReasonCode,
			       (struct aParameter *) &( ctlp->cmsg_data[0] ),
			       ctlp->cmsg_len - sizeofCMsgHdr );

			if ( arcp != (struct aReasonCode *) NULL )
			{
			    (void) printf( " (%u %s)", ntoh2 (arcp->reason),
				LookupSTError( (enum ST2Errors) ntoh2 (arcp->reason)) );
			}

			tlp = (struct aTargetList *) FindParm( STpTargetList,
			       (struct aParameter *) &( ctlp->cmsg_data[0] ),
			       ctlp->cmsg_len - sizeofCMsgHdr );

			if ( tlp == (struct aTargetList *) NULL )
			{
			    if ( Vlevel >= 0 )
				(void) printf( " unspecified target." );
			}
			else
			{
			    tp = &( tlp->targets[0] );
			    for ( j = tlp->TargetCount ; j > 0 ; j-- )
			    {
				if ( Vlevel >= 0 )
				    (void) PrintTarget( tp,tp->TargetBytes,  "        " );

				for ( i = ntargets - 1 ; i >= 0 ; i-- )
				{
				    struct aTarget *tip = targets[ i ].p;

				    if (   (tp->TargetBytes != tip->TargetBytes)
					|| (bcmp( (char *) tp, (char *) tip,
						 tp->TargetBytes ) != 0) )
					continue; /* not match */

				    if ( ctlp->cmsg_type == STReqAccept )
				    {
					if ( targets[ i ].state
					    == TargStateConnecting )
					{
					    targets[ i ].state = TargStateConnected;
					    --repliestogo;
					    ++nconnected;
					} /* else duplicate, ignored */
				    }
				    else /* STReqRefuse */
				    {
					if ( targets[ i ].state
					    == TargStateConnecting )
					{
					    targets[ i ].state = TargStateClosed;
					    --repliestogo; /* Target did not accept */
					}
					else if ( targets[ i ].state == TargStateConnected )
					{
					    targets[ i ].state = TargStateClosed;
					    --nconnected; /* Target disconnected */
					}
				    }

				    break; /* out of searching for loop */
				} /* end of searching for loop */

				tp = (struct aTarget *) ((char *) tp + tp->TargetBytes);
			    } /* end of for all reported targets loop */
			}
			break;

		    case STReqDisconnect: /* Origin sent this Target a disconnect */
			/* (only get here if full-duplex) */
			/**** Name contains a sockaddr(_st2) ****/
			namep = &( name.sa_st2 );
			/**** which contains ST-II parameters ****/
			parmsp = &( namep->st_parms[0] );
			namelen = msg.msg_namelen - sizeofsockaddr_st2header;

			/**** Locate Local Entity parameter which ****/
			/**** identifies identifies the origin  ****/
			parmp = FindParm( STLclAppEnt, parmsp, namelen );

			if ( si == 1 )
			{
			    if ( Vlevel >= 0 )
			    {
				if ( parmp == (struct aParameter *) NULL )
				{
				    (void) printf( "\nSTReqDisconnect received from unspecified target." );
				}
				else
				{
				    (void) printf( "\nSTReqDisconnect received from" );
				    (void) PrintParms( parmp, parmp->plen, "        " );
				}
			    }
			    --nconnected;
			}
			else
			    if ( Vlevel >= 0 )
			    {
				if ( parmp == (struct aParameter *) NULL )
				{
				    (void) printf( "\nSTReqDisconnect on socket %d from unspecified target.",
						  sockets[si] );
				}
				else
				{
				    (void) printf( "\nSTReqDisconnect on socket %d from",
						  sockets[si] );
				    (void) PrintParms( parmp, parmp->plen, "        " );
				}
			    }
			(void) close( sockets[si] );
			remotes[ si ].name[0] = 0;
			sockets[ si ] = -1;
			break;

		    default:
			if ( Vlevel >= 0 )
			{
			    (void) PrintMsg( &msg, -1, "    " );
			    (void) printf( "\n unexpected control ignored" );
			}
			break;
		    } /* end of request switch */

		    /* If all targets have replied, may send data */
		    if ( repliestogo == 0 )
		    {
			repliestogo = -1; /* only tell once */
			if ( nconnected > 1 )
			    (void) printf( "\nAll %d connect replies received.",
					  nconnected );
			else
			    (void) printf( "\nConnect reply received." );
			if ( pktcount > 0 )
			    prompt = "Sending data ... ";
			else
			    prompt = "Waiting for data or command ... ";
		    }
		} /* end of received control */

		else			/* Full-duplex data received */
		{
		    (void) printf( "\nReceived %d. data bytes on origin socket.",
				  error );
		}

	    } /* end of received ST */

	    /* STDIN - user typein ---------------------------------------- */
	    else if ( FD_ISSET( 0, &rfds ) )
	    {
		/* EOF ---------------------------------------------------- */
		if ( feof(stdin) )
		{
		    stdinclosed = 1;
		    (void) printf(
		"\n*** Command input closed, now what are you going to do?\n" );
		    (void) fflush( stdout );
		    continue;
		}

		/* CR ----------------------------------------------------- */
		(void) gets( ttybuf );	/* Get user typein */
		i = strlen( ttybuf );

		if ( i <= 0 )		/* Just <cr> so reprompt  */
		{
		    (void) printf( "%s", prompt );
		    (void) fflush( stdout );
		    continue;		/* for next event */
		}

		/* Command: ?, help --------------------------------------- */

		if (   ((cmdp = "?") && (strncmp( ttybuf, cmdp, i ) == 0))
		    || ((cmdp = "help" ) && (strncmp( ttybuf, cmdp, i ) == 0)) )
		{
		    (void) printf( "\nCommands are:\n\
 ?                    Print this message.\n\
 close                Close the connection.\n\
 connect <target>     Connect to the specified target.\n\
 disconnect           Print a list of connected targets.\n\
 disconnect <target>  Disconnect specified target.\n\
 exit                 Close the connection.\n\
 help                 Print this message.\n\
 quit                 Close the connection.\n\
Anything else is data to be sent.\n%s", prompt );
		    (void) fflush( stdout );
		    continue;
		}

		/* Command: close ----------------------------------------- */
		if ( (cmdp = "close") && (strncmp( ttybuf, cmdp, i ) == 0) )
		{
		    if ( Vlevel > 0 )
		    {
			(void) printf( "\nReceived %s command.", cmdp );
		    }
		    break;		/* out of forever loop */
		}

		/* Command: exit, quit ------------------------------------ */
		if (   ((cmdp = "exit" ) && (strncmp( ttybuf, cmdp, i ) == 0))
		    || ((cmdp = "quit" ) && (strncmp( ttybuf, cmdp, i ) == 0)) )
		{
		    if ( Vlevel > 0 )
		    {
			(void) printf( "\nReceived %s command.", cmdp );
		    }
		    wait = 0;
		    break;		/* out of forever loop */
		}

		/* Command with argument ---------------------------------- */
		cp = index( ttybuf, ' ' );
		if ( cp != (char *) NULL )
		{
		    i = cp - ttybuf - 1; /* Characters to first space */
		}

		/* Command: disconnect ------------------------------------ */
		if (   (   (cmdp = "disconnect") /* side effect */
			&& (strncmp( ttybuf, cmdp, i ) == 0)) )
		{
		    cp = (char *) index( ttybuf, ' ' );
		    if (   (cp++ != (char *) NULL)
			&& ('1' <= *cp) && (*cp <= '9') )
		    {
			tx = atoi( cp );
			if ( Vlevel > 0 )
			{
			    (void) printf( "\nReceived %s target %d command; ",
					  cmdp, tx );
			}
			if ( targets[ --tx ].state != TargStateClosed )
			{
			    si = 1;	/* entry 1 is origin */
			    s = sockets[ si ];
			    if ( s <= 0 )
			    {
				(void) printf(
					"\nError: stream is already closed.\n%s",
					      prompt );
				(void) fflush( stdout );
				continue;
			    }

			    aep = (struct aApplEntity *) &( discname.targ );
			    aep->NextPcol = protocol;
			    aep->IPAdr = targets[ tx ].p->IPAdr;
			    (void) bcopy( (char *) &( targets[ tx ].p->SAP[0] ),
					 /* -> */ (char *) &( aep->SAP[0] ),
					 aep->SAPBytes );

			    namep = &( name.sa_st2 );
			    (void) bcopy( (char *) &( discname ) /*->*/,
					 (char *) namep, sizeof (discname) );

/**************  Origin sending a Disconnect to a target.  *****************/
/**/									 /**/
/**/			    InitMsg (&(msg),namep,sizeof (discname),	 /**/
/**/				    0,0,&(disc),sizeof (disc));		 /**/
/**/									 /**/
			    if ( Vlevel > 2 )
			    {
				(void) printf( "Sending: " );
				(void) PrintMsg( &msg, -1, "    " );
			    }
/**/									 /**/
/**/			    error = sendmsg( s, &msg, 0/*flags*/ );	 /**/
/**/									 /**/
/***************************************************************************/
			    if ( error == -1 )
			    {
				perror( "\nSendmsg failed" );
				remotes[ si ].name[0] = 0;
				sockets[ si ] = -1;
				if ( (si > 1)  &&  (si == nsockets) )
				{
				    --nsockets;
				}
				for ( i = ntargets - 1 ; i >= 0 ; i-- )
				{
				    targets[ i ].state = TargStateClosed;
				}
			    }
			    targets[ tx ].state = TargStateClosed;
			}
		    }
		    else
		    {
			(void) printf( "\nConnected targets are:" );
			for ( i = 0 ; i < ntargets ; i++ )
			{
			    if (   (targets[ i ].state == TargStateConnecting)
				|| (targets[ i ].state
				    == TargStateConnected) )
			    {
				struct aTarget *tp  = targets[ i ].p;
				unsigned char *ucp  = (unsigned char *)
							&( tp->IPAdr );
				unsigned char *uc2p = (unsigned char *)
							&( tp->SAP[0] );
				(void) printf(
					"\n %d: %u.%u.%u.%u  SAP(%d) = %u.%u",
					(i + 1), ucp[0],ucp[1],ucp[2],ucp[3],
					tp->SAPBytes, uc2p[0],uc2p[1] );
			    }
			} /* end of for all targets loop */
		    }
		    (void) printf( "\n%s", prompt );
		    (void) fflush( stdout );

		    continue;
		}

		/* Command: connect --------------------------------------- */
		if (   (   (cmdp = "connect") /* side effect */
			&& (strncmp( ttybuf, cmdp, i ) == 0)) )
		{
		    char *endp;

		    cp = (char *) index( ttybuf, ' ' );	/* target spec */

		    /* ??? only one at a time */
		    if ( (tx = process_target( cp + 1 )) == -1 )
		    {
			(void) fflush( stdout );
			continue;
		    }

		    /* Use current FlowSpec */
		    (void) bcopy( (char *) &( remote.tos ),
				 (char *) &( addon.adtos ),
				 sizeof (addon.adtos) );

		    /**** Copy specified target to addon ****/
		    endp = (char *) &( addon.adtarglst.adtarg1 );
		    remotelen = targets[ tx ].p->TargetBytes;
		    (void) bcopy( (char *) targets[ tx ].p, endp, remotelen );
		    endp += remotelen;
		    addon.adtarglst.plen = 4 + remotelen;

		    /**** Append UserData if specified ****/
		    if ( userdatalen > 0 )
		    {
			struct aUserData *udp = (struct aUserData *) endp;
			char *cp = (char *) &( udp->userdata[0] );
			int i = (long) protoendp - (long) cp;
			if ( (userdatalen + 3) > i )
			{
			    (void) printf(
		  "\nError: Too much user data for connect, must split.\n" );
			    (void) fflush( stdout );
			    continue;
			}
			bcopy( userdatap, /* -> */ cp, userdatalen );
			udp->userlen = /*hton2*/(userdatalen);
			/* Maintain alignment */
			i = (cp - (char *) udp) + userdatalen + 3;
			i &= ~ 3;
			if ( i > 252 )
			{
			    (void) printf(
	"\nError: Too much user data for connect, must split target list.\n" );
			    (void) fflush( stdout );
			    continue;
			}
			udp->plen = i;
			udp->pcode = STpUserData;
			endp += i;
		    }

		    remotelen = (long) endp - (long) &( addon );
		    if ( remotelen >= MAX_ST_NAM )
		    {
			(void) printf(
"\nError: Too many targets or userdata for connect.\n" );
			(void) fflush( stdout );
			continue;
		    }

		    if ( Vlevel >= 0 )
		    {
			(void) printf( "\nAdding target.\n" );
		    }

/**********************  Origin Adding a new target.  **********************/
/**/									 /**/
/**/		    InitMsg (&(msg),&addon,remotelen,0,0,0,0);		 /**/
/**/									 /**/
		    if ( Vlevel > 2 )
		    {
			(void) printf( "Sending: " );
			(void) PrintMsg( &msg, -1, "    " );
			(void) printf( "\n" );
			(void) fflush( stdout );
		    }
/**/									 /**/
/**/		    error = sendmsg( s, &msg, 0/*flags*/ );		 /**/
/**/									 /**/
/***************************************************************************/
		    if ( error == -1 )
		    {
			perror( "\nSendmsg of Connect failed" );
			(void) fflush( stdout );
			targets[ tx ].state = TargStateClosed;
			continue;
		    }

		    targets[ tx ].state = TargStateConnecting;
		    repliestogo++;
		    continue;
		}

		/* Not command is data ------------------------------------ */
		if ( nconnected <= 0 )
		{
		    ttybuf[0] = '\0';
		    (void) printf(
		"Cannot send data until an accept has been received.\n%s",
				  prompt );
		    (void) fflush( stdout );

		    continue;
		}

		prompt = "Waiting for data or command ... ";

		sendp = (char *) ttybuf;
		sendlen = strlen( ttybuf );

	    } /* end of received STDIN */

	    if ( sendp == (char *) NULL )
	    {
		(void) printf( "\n%s", prompt );
		(void) fflush( stdout );
		continue;		/* to wait for next event */
	    }

	    if ( wflag )
	    {
/*******************  Send data using write via ST-II.  ********************/
/**/									 /**/
/**/		error = write( s, sendp, sendlen );			 /**/
/**/									 /**/
/***************************************************************************/
		if ( error == -1 )
		{
		    perror( "\nWrite failed" );
		    for ( i = nsockets - 1 ; i >= 0 ; i-- )
		    {
			if ( sockets[ i ] == s )
			{
			    (void) close( s );
			    sockets[ i ] = -1;
			    remotes[ i ].name[0] = 0;
			    break;	/* out of for loop */
			}
		    } /* end of searching for socket loop */
		}
		else if ( error != sendlen )
		{
		    (void) printf( "Write: sent %d. of %d.", error, sendlen );
		}
	    }
	    else
	    {
/******************  Send data using sendmsg via ST-II.  *******************/
/**/									 /**/
/**/		dataiov.iov_base = (caddr_t) sendp;			 /**/
/**/		dataiov.iov_len = sendlen;				 /**/
/**/		InitMsg (&(msg),&(sendname),sizeof (sendname),		 /**/
/**/			&dataiov,1, NULL,0);				 /**/
/**/									 /**/
		if ( Vlevel > 2 )
		{
		    (void) printf( "\nSending: " );
		    (void) PrintMsg( &msg, sendlen, "    " );
		}
/**/									 /**/
/**/		error = sendmsg( s, &msg, 0/*flags*/ );			 /**/
/**/									 /**/
/***************************************************************************/
		if ( error == -1 )
		{
		    perror( "\nSendmsg failed" );
		    for ( i = nsockets - 1 ; i >= 0 ; i-- )
		    {
			if ( sockets[ i ] == s )
			{
			    (void) close( s );
			    sockets[ i ] = -1;
			    remotes[ i ].name[0] = 0;
			    break;	/* out of for loop */
			}
		    } /* end of searching for socket loop */
		}
		else if ( error != sendlen )
		{
		    (void) printf( "\nSendmsg: sent %d. of %d.", error,
				  sendlen );
		}
	    }

	    if ( pktcount <= 0 )
	    {
		prompt = "Waiting for data or command ... ";
		(void) printf( "\n%s", prompt );
		(void) fflush( stdout );
	    }

	} /* end of forever loop */

	if ( (nconnected <= 0) && (repliestogo <= 0) )
	{
	    (void) printf( " No targets remain\n" );
	}

    } /* end of origin */


    if ( wait )
    {
	(void) printf( "\nWaiting for confirm before aborting: " );
	(void) fflush( stdout );
	(void) gets( ttybuf );
    }

    if ( Vlevel > 0 )
    {
	(void) printf( "\nClosing.\n" );
    }

    for ( i = nsockets - 1 ; i >= 0 ; i-- )
    {
	if ( sockets[i] < 0 )
	{
	    remotes[ i ].name[0] = 0;

/**************************  Abort ST-II socket.  **************************/
/**/									 /**/
/**/	(void) close( sockets[i] );					 /**/
/**/									 /**/
/***************************************************************************/
	}
    }

    exit ( 0 );
    return ( 0 );

} /* end of main */


/*
|* struct aParameter
|* FindParm ( pcode, parmsp, parmslen )
|* Usage:
|*	Scan a list of parameters with given length for first
|*	parameter with specified parameter code.
|* Arguments:
|*	pcode	desired parameter code
|*	parmsp	pointer to parameter list
|*	parmslen length of parameter list
|* Returns:
|*	pointer to desired parameter if found, or NULL
|*
 */


struct aParameter *
FindParm( pcode, parmsp, parmslen )
octet1			 pcode;		/* Desired parameter code */
struct aParameter	*parmsp;	/* Pointer ot parameter list */
int			 parmslen;	/* Length of parameter list */
{
    struct aParameter	*parmp = parmsp; /* Current parameter */

    /* Move it beyond last parameter */
    parmsp = (struct aParameter *) ((long) parmsp + parmslen);

    /* Scan through list */
    while ( parmp < parmsp )
    {
	if ( ntoh1 (parmp->pcode) == pcode )
	{
	    return ( parmp );		/* Found match */
	}

	/* move to next parameter, assuming well formatted */
	parmp = (struct aParameter *) ((long) parmp + parmp->plen);
    }

    return ( (struct aParameter *) NULL ); /* desired parameter not in list */

} /* end of FindParm */


/*
|* octet4 lookupIPAdr ( hostp, textp )
|* Usage:
|*	Used to lookup a host name, find its internet address(es) and ask the
|*	user to specify which address should be used.
|* Arguments:
|*	hostp	pointer to host name/address string
|*	textp	pointer to prompt string
|* Returns:
|*	internet address (net order)
|*
 */

#include <netdb.h>			/* for struct hostent */


octet4 lookupIPAdr( hostp, textp )
char *hostp;
char *textp;
{
    static char		 spaces[] = "                                        ";
    int			 i, a, b, c, d;
    int			 error;
    int			 len;
    unsigned long	 addr;
    struct hostent	*hostentp;


    /* skip over leading spaces or tabs */
    while ( (*hostp == ' ') || (*hostp == '\t') ) hostp++;

    /* gethostbyname doesn't check for address & skip lookup, but we will */
    if ( ('0' <= *hostp) && (*hostp <= '9') )
    {
	i = sscanf( hostp, "%u.%u.%u.%u", &a, &b, &c, &d );
	if ( i == 4 )
	{
	    addr = (  ((0xFF & a) << 24)
		    | ((0xFF & b) << 16)
		    | ((0xFF & c) <<  8)
		    |  (0xFF & d) );

	    return ( (octet4) htonIP (addr) );
	}
    }

    (void) printf( "\n%sHost %s has addresses:", textp, hostp );

    hostentp = gethostbyname( hostp );
    if ( hostentp == (struct hostent *) NULL )
    {
	perror( "\nGethostbyname failed" );
	exit ( 1 );
    }

    /* Print out address(es) for host */
    for ( i = 0 ; hostentp->h_addr_list[i] != (char *) NULL ; i++ )
    {
	(void) printf( " %u.%u.%u.%u",
	       (0xFF & hostentp->h_addr_list[i][0]),
	       (0xFF & hostentp->h_addr_list[i][1]),
	       (0xFF & hostentp->h_addr_list[i][2]),
	       (0xFF & hostentp->h_addr_list[i][3]) );
    }

    len = strlen( textp );
    (void) printf( "\n%*.*sWhich do you want to use (1-%d, a.b.c.d)? : ",
		  len,len,spaces, i );

    (void) gets( ttybuf );		/* Read response */

    if ( ttybuf[0] == '\0' )
    {
	error = 1, a = 1;		/* <cr> means first */
    }
    else
    {
	error = sscanf( ttybuf, "%d.%d.%d.%d", &a, &b, &c, &d );
    }

    if ( error == 4 )			/* Already have octets in a,b,c,d */
    {
	addr = (  ((0xFF & a) << 24)
		| ((0xFF & b) << 16)
		| ((0xFF & c) <<  8)
		|  (0xFF & d) );
	addr = htonIP (addr);
    }
    else if ( error == 1 )
    {
	if ( (a < 1) || (i < a) )
	{
	    (void) printf( "%*.*sInvalid response: %d, exiting.\n",
			  len,len,spaces, a );
	    exit ( 0 );			/* Didn't enter 1 to i */
	}
	a--;

	/* Get octets corresponding to selected (net order) address */
	d = hostentp->h_addr_list[a][3];
	c = hostentp->h_addr_list[a][2];
	b = hostentp->h_addr_list[a][1];
	a = hostentp->h_addr_list[a][0];

	addr = (  ((0xFF & a) << 24)
		| ((0xFF & b) << 16)
		| ((0xFF & c) <<  8)
		|  (0xFF & d) );
    }
    else
    {
	(void) printf( "%*.*sInvalid response: %s, exiting.\n",
		      len,len,spaces, ttybuf );
	exit ( 0 );

    }

    return ( (octet4) addr );

} /* end of lookupIPAdr */


/* Datastructures variables, and routines to parse command line */

char		 nuls[4] = {0,0};	/* Point to here to break out of loop */

struct aswitch {			/* info for bcinp numeric switches */
    int		 min,			/* minimum value */
		 def,			/* default value */
		 max,			/* maximum value */
		*var;			/* variable holding selected value */
    char	*name,			/* name */
		*range;			/* error message for invalid values */
}
		 bsw = {MIN_BYTES_PER_PACKET,DEFAULT_BYTES_PER_PACKET,
			MAX_BYTES_PER_PACKET,&pktbytes,"-b",
			"invalid packet size %d, must be %u:%u bytes\n" },
		 csw = {MIN_PKTS,DEFAULT_PKTS,MAX_PKTS,&pktcount,"-c",
			"invalid packet count %d, must be %u:%u bytes\n" },
		 isw = {MIN_MSEC_BETEEN_PKTS,DEFAULT_MSEC_BETEEN_PKTS,
			MAX_MSEC_BETEEN_PKTS,&pktmsec,"-i",
			"invalid inter-packet interval %d, must be %u:%u msec\n" },
		 nsw = {MIN_PROTOCOL,DEFAULT_PROTOCOL,MAX_PROTOCOL,&protocol,"-n",
			"invalid protocol %d, must be %u:%u\n" },
		 psw = {MIN_PORT,DEFAULT_PORT,MAX_PORT,&port,"-p",
			"invalid port %d, must be %u:%u\n" };

struct aSrcRutDesc {			/* info for source route types */
    char		*abbrev,	/* abbreviation for prompt */
			*name,		/* name of type */
			 open,		/* opening delimiter */
			 close;		/* closing delimiter */
    octet1		 pcode;		/* COrersponding ST-II parameter coed */
}			 srdesc[] = {
  { "Ls-ST", "Loose  ST", '{', '}', STpSTLSrcRut },
  { "St-ST", "Strict ST", '[', ']', STpSTSSrcRut },
  { "Ls-IP", "Loose  IP", '(', ')', STpIPLSrcRut },
  { "St-IP", "Strict IP", '<', '>', STpIPSSrcRut },
};


/*
|* int process_target ( tsp )
|* Usage:
|*	Used to process a target specification:
|*
|*	[@hop[,hop]*@]target_agent[/port]	NO SPACES
|*	where:   @  ==  {}	Loose ST Source Route Hop(s)  (need "...")
|*			[]	Strict ST Source Route Hop(s) (need "...")
|*			()	Loose IP Source Roure Hop(s)  (need "...")
|*			<>	Strict IP Source Roure Hop(s) (need "...")
|* Arguments:
|*	tsp	pointer to target specification string
|* Returns:
|*	error:		-1
|*	success:	target[] index
|*
 */

  int
process_target( tsp )
char		*tsp;			/* Target specification string */
{
    char		 prefix[16],	/* For prompt */
			*cp,		/* Position of comma in source route */
			*ep,		/* Pointer to end of a field */
			*wp;		/* Current position in tsp */
    int			 i,		/* Counter, misc int */
			 len;		/* Length of target */
    long		 l;		/* Used for alignment calculation */
    octet4		*ipadrwp;	/* Pointer to area for IPAdr in srwp */
    struct aSrcRutDesc	*srdp;		/* Ptr into Source Route type table */
    struct aSrcRut	*srwp;		/* Working pointer to Source Route */
    struct aTarget	*wtp;		/* Working Target pointer */


    wtp = (struct aTarget *) freeTargetp; /* Area for next Target */
    wp = tsp;				/* Start at the beginning */

    /* do variable length saps here */
    wtp->SAPBytes = 2;			/* Example uses 2-octet SAPs */

    /* SAP, if specified, follows "/" */
    if ( (ep = (char *) index( wp, '/' )) != (char *) NULL )
    {
	i = atoi( (ep+1) );
	if ( (i < psw.min) || (psw.max < i) )
	{
	    (void) printf( "\n" );
	    (void) printf( psw.range, i, psw.min, psw.max );
	    return ( -1 );
	}
	wtp->SAP[0] = 0xFF & (i >> 8);	/* User specified SAP */
	wtp->SAP[1] = 0xFF &  i;
    }
    else
    {
	wtp->SAP[0] = 0xFF & (port >> 8);
	wtp->SAP[1] = 0xFF &  port;	/* Default (or -p) SAP value */
    }

    /* ST-II wants everything 4-octet aligned to keep sparcs happy */
    l = (long) &( wtp->SAP[ wtp->SAPBytes - 1 ] ) - (long) wtp + 3;
    l &= ~3;
    l += (long) wtp;
    srwp = (struct aSrcRut *) l;	/* Where a source route should go */
    
    /* Process any Source Route specifications */
    while (1)
    {
	while ( (*wp == ' ') || (*wp == '\t') ) wp++;
	srdp = &( srdesc[0] );		/* First table entry */
	for ( i = 0 ; i < DimensionOf (srdesc) ; i++, srdp++ )
	{
	    if ( *wp == srdp->open )	/* pointing to opening character? */
		break;			/* yes, found source route */
	}

	if ( DimensionOf (srdesc) <= i )
	    break;			/* while, no (more) source routes */

	/* find end of source route specification */
	ep = (char *) index( wp, srdp->close );
	if ( ep == (char *) NULL )
	{
	    (void) printf( "\nError: Missing %c in %s\n", srdp->close, tsp );
	    return ( -1 );
	}

	srwp->pcode = srdp->pcode;	/* Type of source route */
	srwp->plen = 4;			/* Overhead */
	srwp->fill2 = 0;
	ipadrwp = &( srwp->IPAdr[0] );	/* Slot for IPAdr */
	i = 1;				/* count IPAdrs */

	/* hops are separated by commas */
	while ( (cp = (char *) index( ++wp, ',' )) != (char *) NULL )
	{
	    (void) sprintf( prefix, " %s  %3d: ", srdp->abbrev, i++ ); /* ??? */
	    *cp = '\0';			/* Replace comma to end hop name */
	    *ipadrwp = lookupIPAdr( wp, prefix ); /* Get IPAdr */
	    *cp = ',';			/* Put comma back */
	    wp = cp;			/* Move over to it */

	    if ( *ipadrwp++ == INADDR_ANY )
	    {
		(void) printf( "\nError: INADDR_ANY is not a valid source route hop\n\
           in %s\n", tsp );
		return ( -1 );
	    }

	    /* Make sure have room for entry */
	    if ( protoendp <= ipadrwp )
	    {
		(void) printf( "\nError: Too many source route hops\n" );
		return ( -1 );
	    }

	    srwp->plen += 4;		/* IPAdr adds 4 octets */
	}

	/* Process last hop as above */
	(void) sprintf( prefix, " %s  %3d: ", srdp->abbrev, i++ ); /* ??? */
	*ep = '\0';			/* Replace closing character */
	*ipadrwp = lookupIPAdr( wp, prefix ); /* Get IPAdr */
	*ep++ = srdp->close;		/* Put it back */
	wp = ep;			/* Move after source route */

	if ( *ipadrwp++ == INADDR_ANY )
	{
	    (void) printf( "\nError: INADDR_ANY is not a valid source route hop\n\
           in %s\n", tsp );
	    return ( -1 );
	}

	/* Make sure have room for entry */
	if ( protoendp <= ipadrwp )
	{
	    (void) printf( "\nError: Too many source route hops\n" );
	    return ( -1 );
	}

	srwp->plen += 4;		/* IPAdr adds 4 octets */

	srwp = (struct aSrcRut *) ipadrwp; /* For next source route */

    } /* end of all src ruts */

    /* Compute length of Target parameter */
    wtp->TargetBytes = (long) srwp - (long) wtp;

    /* find end of Target name */
    if ( (ep = (char *) index( wp, '/' )) != (char *) NULL )
	*ep = '\0';
    while ( (*wp == ' ') || (*wp == '\t') ) wp++;
    (void) sprintf( prefix, "Target  %3d: ", ntargets+1 );
    wtp->IPAdr = lookupIPAdr( wp, prefix ); /* Get Target IPAdr */

    /* Restore input string */
    if ( ep != (char *) NULL )
	*ep++ = '/';

    /* Simple validate of Target address */
    if ( wtp->IPAdr == INADDR_ANY )
    {
	(void) printf( "\nINADDR_ANY is not a valid target\n" );
	return ( -1 );		/* Without updating freeTargetp */
    }

    /* Make sure it all fits */
    if (   ((octet4 *) srwp >= protoendp)
	|| (ntargets >= MAX_TARGETS - 1) )
    {
	(void) printf( "\nError: Too many targets (%d) with %s\n", MAX_TARGETS,
		      tsp );
	return ( -1 );
    }

    for ( i = ntargets - 1 ; i >= 0 ; i-- )
    {
	if (   (wtp->TargetBytes == targets[ i ].p->TargetBytes)
	    && (bcmp( (char *) wtp, (char *) targets[ i ].p,
		      wtp->TargetBytes ) == 0) )
	  break;		/* Found a duplicate */
    } /* end of for all targets loop */

    if ( i >= 0 )
    {
	switch ( targets[ i ].state )
	{
	case TargStateConnecting:
	case TargStateConnected:
	case TargStateClosing:
	    break;

	case TargStateNew:
	    (void) printf( "  ***  Duplicates target %d, target %d ignored.\n",
			  i + 1, ntargets + 1 );
	    /* Fall through */

	case TargStateClosed:
	    targets[ i ].state = TargStateNew;
	    return ( i );
	} /* end of old target state switch */

	(void) printf( "  ***  Duplicates active target %d, ignored.\n",
		      i + 1 );
	return ( -1 );
    }

    /* Append new target to end of list */
    targets[ ntargets ].state = TargStateNew;
    targets[ ntargets ].p = (struct aTarget *) freeTargetp;
    freeTargetp = (octet4 *) srwp;	/* For next Target */
    ntargets++;			/* Count this one */

    return ( ntargets - 1 );

} /* end of process_target */


/*
|* void process_args ( ac, app )
|* Usage:
|*	Used to process command line arguments
|* Arguments:
|*	ac	argument count
|*	app	pointer to argument list
|* Returns:
|*	on success, calls usage which exits on error
|*
 */


void
process_args( ac, app )
int			  ac;		/* argument count */
char			**app;		/* ptr to next argument in list */
{
    char		 *ap,		/* pointer to current argument */
			  ch;		/* character in argument */
    int			  a,b,c,d,
			  num;
    struct aswitch	 *swp;


    ac--, app++;			/* Skip over command name */

    /* Process all arguments */
    while ( (ac > 0) && (*app != (char *) NULL) )
    {
	ap = *app++;			/* Get next argument */

	switch ( *ap )			/* Check for switches */
	{
	case '-':			/* Argument is switch (list) */
	    for ( ++ap ; *ap != '\0' ; ) /* Process each switch */
	    {
		ch = *ap++;		/* switch name character */
		switch ( ch )
		{			/* Simple flag switches */
		case '1':	single_socket = 1;	break;
		case '?':	Hflag = 1;
		case 'h':	usage();		break;
		case 'B':	Bflag = 1;		break;
		case 'l':	listener = 1;		break;
		case 'N':	no_recovery = 1;	break;
		case 'o':	origin = 1;		break;
		case 'R':	reuseaddr = 1;		break;
		case 'S':	useselect = 0;		break;
		case 'T':	Tflag = 1;		break;
		case 'v':	Vlevel++;		break;
		case 'w':	wflag = 1;		break;

		case 'a':
		  if ( *ap == '\0' )
		  {
		      if (   (ac-- <= 0)
			  || (*(ap = *app++) == '\0') )	/* Get next argument */
			  usage();
		  }
		  num = sscanf( ap, "%u.%u.%u.%u", &a, &b, &c, &d );
		  if (   (num != 4)
		      || (a > 255) || (b > 255) || (c > 255) || (d > 255) )
		  {
		      (void) printf( "Invalid IP Address: %s\n", ap );
		      usage();
		  }
		  (void) strncpy( &( ourname[0] ), ap, sizeof( ourname ) -1 );
		  ouraddr =   ((0xFF & a) << 24) | ((0xFF & b) << 16)
			    | ((0xFF & c) <<  8) |  (0xFF & d);
		  ap = nuls;
		  break;
					/* Numeric switches */
		case 'b':	swp = &bsw;	goto numeric;
		case 'c':	swp = &csw;	goto numeric;
		case 'i':	swp = &isw;	goto numeric;
		case 'n':	swp = &nsw;	goto numeric;
		case 'p':	swp = &psw;	goto numeric;
numeric:
		    if ( (*ap < '0') || ('9' < *ap) )
		    {
			if ( (ac-- <= 0) || (*app == (char *) NULL) )
			{
			    (void) printf(
				"Error: No value argument for %s switch.",
					  swp->name );
			    usage();
			}
			num = atoi( *app++ ); /* No value means next argument */
		    }
		    else
		    {
			num = atoi( ap ); /* Get value */
			ap = nuls;	/* abandon this argument */
		    }

		    /* Validate numberic range */
		    if ( (num < swp->min) || (swp->max < num) )
		    {
			(void) printf( swp->range, num, swp->min, swp->max );
			usage();
		    }

		    *(swp->var) = num;	/* Accept new value */

		    if ( swp == &isw )
		    {
			poll.tv_sec  =  pktmsec / 1000;
			poll.tv_usec = (pktmsec % 1000) * 1000;
		    }
		    break;

		case 't':		/* Process target specifications */
		    origin = 1;		/* a Target implies were the origin */

		    if ( *ap != '\0' )	/* spec appended to -t */
		    {
			do
			{
			    char *ep;

			    /* drop leading spaces or tabs */
			    while ( (*ap == ' ') || (*ap == '\t') ) ap++;

			    /* may be several targets separated by spaces */
			    ep = (char *) index( ap, ' ' );
			    if ( ep != (char *) NULL )
				*ep = '\0';

			    num = process_target( ap ); /* Do the work */
			    if ( num == -1 )
				usage();

			    if ( ep != (char *) NULL )
				*ep++ = ' ';
			    else
				ep = nuls; /* abandon this argument */

			    ap = ep;	/* position for next Target */

			} while ( *ap != '\0' );
		    }
		    else		/* spec in next argument(s) */
		    {
			if ( (ac <= 0) || (*app == (char *) NULL) )
			    usage();

			/* everything up to next switch is a Target */
			while (   (ac-- > 0)
			       && (*app != (char *) NULL) && (**app != '-') )
			{
			    num = process_target( *app++ ); /* Do the work */
			    if ( num == -1 )
				usage();
			} /* end of while loop */
		    }
		    break;

		case 'u':		/* UserData info follows switch */
		    userdatap = ap;
		    userdatalen = strlen( ap );
		    ap = nuls;		/* abandon this argument */
		    break;

		default:
		    (void) printf( "\nError: Unknown switch '%c'\n", ch );
		    usage();
		    break;

		} /* end of switch on command switch characters */
	    } /* end of for all characters in switch loop */
	    break;

	default:			/* Treat it as a target */
	    origin = 1;			/* a Target implies were the origin */
	    num = process_target( ap );	/* Do the work */
	    if ( num == -1 )
		usage();
	    break;

	} /* end of switch on arguments first character */

    } /* end of for all arguments loop */
 
   return;

} /* end of process_args */


/*
|* void usage ()
|* Usage:
|*	Called on command line error to print usage information & exit
|* Arguments:
|*	none
|* Returns:
|*	never, always exits
|*
 */

void
usage()
{
    /* Summary usage information */
    (void) printf( "usage: %s [-1?BhlSwv] [-b<bytes>] [-c<#pkts>] [-i<msec>]\n\
       %*.*s [-n<protocol>] [-p<port>] [-u<userdata] [-t target_spec*]\n",
    arg0p, strlen( arg0p ), strlen( arg0p ), " " );

    /* Optional detailed information */
    if ( Hflag )
    {
	(void) printf( "\
  -1            single socket mode (no)\n\
  -?            print detail usage\n\
  -a<ipadr>     use specified local host address\n\
  -B            Omit bind\n\
  -b<bytes/pkt> bytes per packet (%d)\n\
  -c<# pkts>    number of pktets to send (%d)\n\
  -h            print summary usage\n\
  -i<msec/pkt>  #  msec between packets (%d)\n\
  -l            listener\n\
  -n<protocol>  next protocol (%d)\n\
  -N            use No Recovery option\n\
  -o            origin\n\
  -p<port>      port number (%d)\n\
  -R		allow local address reuse (aka SO_REUSEADDR)\n\
  -S            don't use select\n\
  -t<targets>   target specifications (implies -o)\n\
        [@hop[,hop]*@]target_agent[/port]\n\
        where:   @  ==  %c%c  %s Source Route Hop(s) (use \"...\")\n\
                        %c%c  %s Source Roure Hop(s) (use \"...\")\n\
                        %c%c  %s Source Route Hop(s) (use \"...\")\n\
                        %c%c  %s Source Roure Hop(s) (use \"...\")\n\
  -T            use ST-II timestamp\n\
  -u<userdata>  user data\n\
  -v		increment verbosity (0)\n\
                0 - user interactions only\n\
                1 - API interactions\n\
                2 - API data structures\n\
                3 - everything\n\
  -w            use write, not sendmsg",
	       DEFAULT_BYTES_PER_PACKET, DEFAULT_PKTS,
	       DEFAULT_MSEC_BETEEN_PKTS,
	       DEFAULT_PROTOCOL, DEFAULT_PORT,
	       srdesc[0].open, srdesc[0].close, srdesc[0].name,
	       srdesc[1].open, srdesc[1].close, srdesc[1].name,
	       srdesc[2].open, srdesc[2].close, srdesc[2].name,
	       srdesc[3].open, srdesc[3].close, srdesc[3].name );
	(void) printf(
		"\nE.g., Target: %s\n      Origin: %s -t <target> -p 4321\n",
		      arg0p, arg0p );
    }
    else
	(void) printf( "       %s '-?'    for more details\n", arg0p );

    exit ( 0 );

} /* end of usage */

/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
