#ifndef lint
static char rcsid_st2_routing_c[] = "\
@(#)  $Header: ststat.c,v 1.10 91/11/26 23:07:58 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint


/*
M* ststat.c  Main routine to dump ST-II kernel data structures.
M*
 */

/*
m* Status:
m*   Features:
m*	Print major data structures.
m*   Untested Features:
m*   Restrictions/Bugs:
m*	Doesn't work well with crash dumps -- need to figure out how
m*	virtual memory mapping works.
m*	Vulnerable to bad structures - loops or faults.
m*   Things to do:
m*	Add one-line per stream summary a la netstat.
m*	Add more switches.
m*
 */

/* Module Revision History
 *
 * $Log:	ststat.c,v $
 * Revision 1.10  91/11/26  23:07:58  clynn
 * Added -v switch to print_vlinks.
 * 
 * Revision 1.9  91/11/04  09:16:10  clynn
 * Added support for generic interfaces, neighbors, retransmission queue,
 * and traps.
 * 
 * Revision 1.8  91/05/28  16:18:05  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 */

#include <stdio.h>
#include <kvm.h>			/* kvm_t */
#include <fcntl.h>			/* O_RDONLY */
#include <nlist.h>
#include <sys/stdtypes.h>		/* time_t */
#include <sys/time.h>			/* struct timeval */

extern int g_dump;
extern int g_mf;
extern struct nlist g_nl[];

struct nlist		 l_nl[] = { { "_time" }, 0 };
struct timeval tv;
time_t now;
char hostname[256];
char prefix[] = "         ";

#define Ident(x)x

#define SwitchList \
 aSwch(a,Print ARP table) \
 aSwch(c,Print Configuration table) \
 aSwch(D,Print Dump ST-II tables) \
 aSwch(g,Print Generic Interfaces) \
 aSwch(h,Print HID Hash table) \
 aSwch(n,Print Neighbor table) \
 aSwch(r,Print Resource tables) \
 aSwch(s,Print Statistics table) \
 aSwch(t,Print Trap table) \
 aSwch(v,Print VLink table) \
 aSwch(x,Print Retransmission queue) \




  int
main(argc, argpp)
  int			 argc;
  char			**argpp;
{
    char		*prognamep = *argpp++,
			*argp = *argpp,
			*sname = "/vmunix",
			*cname = "/dev/mem";

#define aSwch(l,desc) int Ident (l)flag = 0;
    SwitchList
#undef aSwch

    kvm_t		*kfd;


    gethostname( &( hostname[0] ), sizeof (hostname) );

    /* -------------------------------------------------------------------- */
    /* Process arguments */

    while ( (--argc > 0) && (* (argp = *argpp) == '-') )
    {
	while ( *++argp != '\0' )
	{
	    switch ( *argp )
	    {
	    case 'f':
		if (argc > 1)
		{
		    sname = argpp[1];
		    argpp++; argc--;
		    break;
		} else
		    goto bad;

	    case 'd':
		if (argc > 1)
		{
		    g_dump++;
		    cname = argpp[1];
		    sprintf ( &( hostname[0] ), "file %s", cname );
		    argpp++; argc--;
		    break;
		} else
		    goto bad;

#define aSwch(l,desc)	case 'l': Ident (l)flag++;  break;
	    SwitchList
#undef aSwch

	    case '?':
	    default:
  bad:
		printf ("usage: %s [-", prognamep );
#define aSwch(l,desc) printf( "l" );
		SwitchList 
#undef aSwch
		printf ( "] [-f <sysname>] [-d <corename>]\n" );

		if ( argp[1] == '?' )
		{
#define aSwch(l,desc) printf ( "    l   desc\n" );
		    SwitchList
#undef aSwch
		}
		return;

	    } /* end of switch */

	} /* end of all characters in switch loop */

	argpp++; argc--;

    } /* end of all args loop */

    /* .................................................................... */
    /* Default to steram dump */

#define aSwch(l,desc) && (Ident (l)flag == 0)
    if ( (0 == 0) SwitchList )
#undef aSwch
	Dflag++;

    /* -------------------------------------------------------------------- */
    /* Open appropriate memory file */

    kfd = kvm_open ( sname, cname, (char *) NULL, O_RDONLY, "kvm_open" );
    if ( kfd == (kvm_t *) NULL )
    {
	fprintf ( stderr, "not able to access %s / %s", sname, cname );
	exit(1);
    }

    g_mf = (int) kfd;			/* publish memory file */

    /* -------------------------------------------------------------------- */
    /* Read interesting symbols */

    if ( kvm_nlist( kfd, g_nl ) == -1 )
    {
	fprintf (stderr, "no %s namelist", sname);
	exit(1);
    }

    /* -------------------------------------------------------------------- */
    /* Get current time or time from dump & print title */

    if (    g_mf
	&& (kvm_nlist( kfd, l_nl ) != -1)
	&& ((long) l_nl[ 0 ].n_value != 0) )
    {
	(void) kvm_read ( g_mf, (long) l_nl[ 0 ].n_value, &tv, sizeof (tv) );
	now = tv.tv_sec;
    }
    else
	(void) time( &now );

    printf ( "Kernel structure dump of %s started at %s", &( hostname[0] ), ctime( &now ) );

    /* -------------------------------------------------------------------- */
    /* Print ST-II Data Structures */

    if ( aflag )
	print_arptable( prefix );	/* ARP Table */


    if ( Dflag )			/* Active Connections */
    	print_aST2pcb( (struct aST2pcb *) NULL, prefix );


    if ( hflag )
	print_ConHshTable( prefix );    /* ConHsh Table */


    if ( vflag )
	print_vlinks( prefix );		/* VLink Table */


    if ( xflag )
	print_rxqueue( prefix );	/* Rx Queue */


    if ( rflag )
	print_netrsrcs( prefix );	/* Network resource usage */


    if ( cflag )
	print_config( prefix );		/* Configuration table */


    if ( gflag )
	print_gifcs( prefix );		/* Generic interfaces */


    if ( nflag )
	print_aNeighborList( prefix );	/* Neighbor table */


    if ( sflag )
	print_stats( prefix );		/* Statistics */


    if ( tflag )
	print_traps( prefix );		/* Traps */

    /* -------------------------------------------------------------------- */
    /* Print trailer */

    if ( strcmp (cname,"/dev/mem") == 0 )
	(void) time( &now );
    printf ( "\nKernel structure dump of %s stopped at %s", &( hostname[0] ), ctime( &now ) );


    return ( 0 );

} /* end of main */

/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
