#ifndef lint
static char rcsid_st2_c[] = "\
@(#)  $Header: st2.c,v 1.10 91/11/26 23:09:22 clynn Rel $  \n\
@(#) /*-------------------------------------------------------------------* \n\
@(#)  |       Copyright (c) 1991 by BBN Systems and Technologies,         | \n\
@(#)  |            A Division of Bolt Beranek and Newman Inc.             | \n\
@(#)  |                                                                   | \n\
@(#)  |  Permission to use, copy, modify, distribute, and sell this       | \n\
@(#)  |  software and its documentation for any purpose is hereby         | \n\
@(#)  |  granted without fee, provided that the above copyright notice    | \n\
@(#)  |  and this permission appear in all copies and in supporting       | \n\
@(#)  |  documentation, and that the name of Bolt Beranek and Newman      | \n\
@(#)  |  Inc.  not be used in advertising or publicity pertaining to      | \n\
@(#)  |  distribution of the software without specific, written prior     | \n\
@(#)  |  permission. BBN makes no representations about the suitability   | \n\
@(#)  |  of this software for any purposes.  It is provided ``AS IS''     | \n\
@(#)  |  without express or implied warranties.                           | \n\
@(#)  *-------------------------------------------------------------------*/ \n";
#endif lint


/*
M* st2.c  ST2 Data Forwarding and Interface to Networks and SCMP Routines.
M*
 */

/*
m* Status:
m*   Features:
m*   Untested Features:
m*   Restrictions/Bugs:
m*   Things to do:
m*	Make more use of memory buffer abstraction and less of mbufs.
m*	Priorities.
m*
 */


/* Module Revision History
 *
 * $Log:	st2.c,v $
 * Revision 1.10  91/11/26  23:09:22  clynn
 * Minor changes to some traps.
 * 
 * Revision 1.9  91/11/04  09:19:36  clynn
 * Updated for Public Domain Release.  Major changes: addition of Source
 * Routing, IP Encapsulation, HELLO protocol between neighbors, tracking
 * of neighbors, detection of component or agent failures & notification
 * to applications.  More consistant naming and format, including making
 * all external names begin with "st2_".  Moved some routines and data
 * structures to reduce external references.
 * 
 * Added IP Encapsulation support.
 * 
 * Revision 1.8  91/05/28  16:17:36  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 * Revision 1.1  91/03/15  18:33:21  clynn
 * Initial revision
 */

#define _FILE_ Strst2			/* Module bugid 1,2 */

/*
m* Overview
m*
m*	Received packets are dispatched by the receiving local
m*	network interface to one of several protocol families, based
m*	on information in the link-level header.  For ST-II, this may
m*	take one of two forms: ST packets delivered to the IP family,
m*	using IP Version 5, or through a specific link-level
m*	identifier that specifies ST directly.  For now, the packets
m*	will be queued for ST; unresolved is the issue of whether the
m*	call to do the enqueueing must return quickley in order not to
m*	loose the next packet, or whether it can push a data packet
m*	through to the output queue before returning.  Packets
m*	containing ST-II Data and Control Messages are placed into the
m*	ST-II input queue (st2_intrq) by the local network interfaces,
m*	which also cause the st2_intr routine to be invoked.
m*
m*	When called, st2_intr removes packets from the input queue,
m*	validates the ST Header and extractes the connection id (HID).
m*	The HID is used (by ANDing with ConfigParm (hid_mask)) to index
m*	into the st2ConHsh table.  The table contains the full HID,
m*	that is verified, and a pointer to the associated stream (or,
m*	more precisely, meander of a stream) data structure (aST2pcb)
m*	that specifies a routine to which the incoming packet is
m*	passed.  ST-II Control Messages all carry HID zero, which
m*	causes the SCMP packets to be passed to st2_CMPInput (see
m*	st2_cmp .c).  ST-II Data packets are passed, with a pointer to
m*	the aST2pcb, to the st2_Forward routine.
m*
m*	The st2_Forward routine replicates ST-II Data packets as
m*	required, replaces the HID, updates the ST Header checksum,
m*	and forwards the copies to all of a stream's (meander's)
m*	next-hops.
m*
m*	Several databases are accessed when dispatching an ST-II
m*	packet:
m*
m*	st2_ConHsh:	This table maps incoming Hop-By-Hop
m*			Identifiers (HIDs) to the corresponding
m*			aST2pcb data structure.
m*
m*	aST2pcb:	Primary per meander data structure.  Pointed
m*			to from st2_ConHsh (by cp).  It contains a
m*			pointer (nxthopsp) to a list of aNxtHop
m*			structures tha specify how the Data packets
m*			should be forwarded.
m*
m*	aNxtHop:	Per meander per "next hop" local network data
m*			structure that contains next-hop local network
m*			address information.  Multiple next-hops may
m*			be reached through a single aNxtHop when local
m*			network multicast/broadcast is being used.
m*
 */

/* External definitions */

#include <sys/types.h>		/* for <sys/mbuf.h> and  <netinet/in.h> */
#include <sys/mbuf.h>		/* needs <sys/types.h> & <sys/param.h> */
#include <sys/socket.h>		/* for <net/if.h> */
#include <net/if.h>		/* if_output, for <netinet/if_ether.h>,
				   needs <sys/socket.h> */
#include <netinet/in.h>		/* before st2.h -> st2_api.h */

typedef unsigned long n_long;	/* Include disease */
#include <netinet/ip.h>		/* IP Header for encapsulation */

#include "st2_api.h"		/* Global Definitions, before st2.h */
#include "st2.h"		/* ST-II Implementation definitions,
				 * includes st2_api.h, before st2_cmp.h */

#ifndef STIINOST1		/* Pass ST-I packets back to Raw IP */
extern int		 rip_input (/*ip mbuf */);
#endif  STIINOST1

/* Local routines */

enum ST2Errors		 st2_Forward (/* gifcp, cp, msgp, st2p, datap,
					 encaplen */);
int			 st2_intr ();


/*
S* st2_Forward ( gifcp, cp, msgp, st2p, datap, encaplen )
s*
s*	A dspin routine, called for HIDs >= HID_DYNAMIC to forward data.
s*
 */

  enum ST2Errors			/* Next bugid 0x10106 */
st2_Forward( gifcp, cp, msgp, st2p, datap, encaplen ) /* from st2_intr api_send */
					/* set in st2_PCBGet HandleConnect */
  struct aNetIf		*gifcp;		/* Receiving interface (UNUSED) */
  struct aST2pcb	*cp;		/* Connection entry, via HID */
  struct mbuf		*msgp,		/* First aPktDesc (with ifaddr/header),
					   we dispose */
			*st2p;		/* aPktDesc containing ST2Header+... */
  octet1		*datap;		/* Ptr after ST-II Header */
					/* ARGSUSED */
  int			 encaplen;	/* Bytes PRECEEDING ST-II Header */
{
    enum ST2Errors	 retcod = NoError;
    int			 cpyflg,
			 i,
			 dups = 0,
			 errcod,	/* OS error */
			 st2len;
    struct aNxtHop	*nxthopp;
    struct aNxtHopList	*nxthopsp;
    struct aST2Header	*st2hdrp;
    struct mbuf		*cpyp,
			*hedp;


    st2hdrp = mtod (st2p,struct aST2Header *);
    st2len = ntoh2 (st2hdrp->TotalBytes);

    if ( (nxthopsp = cp->nxthopsp) EQ NO_NHLISTP )
    {
	Count (frwd_nonxthops,1);	/* no listed next hops */
	cp->prvhop.thruput.data_in_drop_bytes += st2len + encaplen;
	cp->prvhop.thruput.data_in_drop_pkts++;
	FreePkts ((struct aPktDesc *) msgp);
	BUGRETURN (-,DropNoRoute,0x10101,enum ST2Errors);
    }

    /* ??? Validate sender's local net address for denial of service vs
       multiple senders (via multicast group) for single stream
       PrvHopValidate & prvhop.lclhdrlen, prvhop.lclsockaddr */

    cp->prvhop.thruput.data_in_bytes += st2len + encaplen;
    cp->prvhop.thruput.data_in_pkts++;

    for ( ; nxthopsp NE NO_NHLISTP ; nxthopsp = nxthopsp->nxtNxtHopp )
    {
	nxthopp = &( nxthopsp->NxtHops[0] );
        for ( i = nxthopsp->nxtfree ; i > 0 ; --i, ++nxthopp )
        {
	    if ( ! IsSet32 (nxthopp->flags, NxtHopActive) )
	        continue;		/* Skip inactive/deleted entries */

	    gifcp = nxthopp->gifcp;
	    if ( gifcp EQ NO_NETIFP )
	    {
	        Count (nofwdifc,1);	/* forwarder not ready */
	        continue;
	    }

/* ??? Need own loaned buffer mgmt here ... ether is using one of its LOANED
   mbufs, & m_copy is using a system LOANED mbuf.  We abstract above multiple
   use of ether buffer & supply our free routine.  Also lets us know when
   packet has been sent!!
 */
	    cpyflg = 0;
	    if (   (i > 1)
		OR (nxthopsp->nxtNxtHopp NE NO_NHLISTP) )
	    {
		/* Must duplicate the ST header + data */
	        cpyp = m_copy ( st2p, /*skip*/ 0, M_COPYALL );
		cpyflg = 1;
	    }
	    else				/* Last time through */
	    {
		/* break ST header + data out of msg */
	        if ( msgp EQ st2p )
		{
		    msgp = (struct mbuf *) NULL;
		    cpyp = st2p;	/* Original becomes last copy */
		    st2p = (struct mbuf *) NULL;
		}
	        else if ( msgp->m_next EQ st2p )
		{
		    msgp->m_next = (struct mbuf *) NULL;
		    cpyp = st2p;	/* Original becomes last copy */
		    st2p = (struct mbuf *) NULL;
		}
	        else if ( msgp->m_next->m_next EQ st2p )
		{
		    msgp->m_next->m_next = (struct mbuf *) NULL;
		    cpyp = st2p;	/* Original becomes last copy */
		    st2p = (struct mbuf *) NULL;
		}
	        else
		{			/* Just copy here, too */
		    BUGSTOP (-,Inconsistency,0x10102);
		    cpyp = m_copy ( st2p, /*skip*/ 0, M_COPYALL );
		    cpyflg = 1;
		}
	    }

	    if ( cpyp EQ (struct mbuf *) NULL )
	    {
	        BUGSTOP (-,5354,0x10103);
	        nxthopp->thruput.data_out_drop_pkts++;
		nxthopp->thruput.data_out_drop_bytes += st2len + encaplen;
	        continue;
	    }

	    dups++;

	    st2hdrp = mtod (cpyp,struct aST2Header *);

	    if ( cpyflg OR (nxthopp->encaplen > 0) )
	    {
		unsigned char *ucp;

		cpyflg = 2 + nxthopp->lclhdrlen + nxthopp->encaplen
			+ sizeofaST2Header;
		/* Data structures ought to be correct! */
		if ( ((cpyflg & 3) != 0) OR (cpyflg > MLEN) )
		{
		    Count (other,1);
		    m_free ( cpyp );
		    continue;
		}

		MGET ( hedp, M_DONTWAIT, cpyflg );
		if ( hedp EQ (struct mbuf *) NULL )
		{
		    IfLog (DBG_MEM_ALO)
			(void) printf ( "st2_Forward (no free mbufs)\n" );
		    Count (scmp_nobuf1,1);	/* count no small bufs */
		    BUGSTOP (-,CantGetResrc,0x10104);
		    m_free ( cpyp );
		    continue;
		}

		ucp = mtod (hedp,unsigned char *);
		Bcopy (&( nxthopp->lclsockaddr ),ucp,cpyflg - sizeofaST2Header);
		ucp += 2 + nxthopp->lclhdrlen;
		hedp->m_off += 2 + nxthopp->lclhdrlen;
		hedp->m_len = cpyflg - 2 - nxthopp->lclhdrlen;

		/* Copy ST header (w/o timestamp) to our copy for HID update */
		{   struct aST2Header *ost2hdrp = st2hdrp;
		    st2hdrp = Mkp (struct aST2Header *,ucp,nxthopp->encaplen);
		    *st2hdrp = *ost2hdrp;
		}

		cpyp->m_off += sizeofaST2Header;
		cpyp->m_len -= sizeofaST2Header;
		hedp->m_next = cpyp;
		cpyp = hedp;

		/* if IP encapsulation, must add ST length to IP total length
		   and update IP checksum (computed with (IP total length = 0))
		   NB: nxthopp->encaplen EQ IP total length before adding ST */

		if ( nxthopp->encaplen > 0 )
		{
		    struct ip		*ipp = Mkp (struct ip *,ucp,0);
		    unsigned long	 cksum;

		    ipp->ip_len = hton2 (nxthopp->encaplen + st2hdrp->TotalBytes);

		    cksum = ipp->ip_sum + ONES_NEGATE (ipp->ip_len);
		    ipp->ip_sum = cksum + (cksum >> 16);
		}
	    }

	    /* adjust ST header */
	    /*
	     * Incremental checksum update: header checksum is
	     * negative of the other header fields so we add original
	     * positive HID back to leave negative of remaining
	     * fields.  Then, we add one's comp negative of the new
	     * HID to yeild correct (negative) checksum for new
	     * header.  Note that we have to be sure that negate takes
	     * place in 16 bits; sum needs 18 to catch carry out of
	     * high bits, which we end-around add back into low to
	     * yeild one's comp sum.
	     */

	    {   unsigned long cksum = st2hdrp->STChecksum + st2hdrp->HID
					+ ONES_NEGATE (nxthopp->HID);
	        st2hdrp->STChecksum = cksum + (cksum >> 16);
	    }
	    st2hdrp->HID = nxthopp->HID;

	    if ( gifcp->ifc_output EQ st2_ToAppl )
	    {
		errcod = st2_ToAppl( gifcp, (struct aPktDesc *) cpyp,
			(struct sockaddr_st2 *) &( nxthopp->lclsockaddr ) );
		/* EINVAL 22, ENETDOWN 50, ENOBUFS 55 */
	    }
	    else
	    {
		errcod = gifcp->ifc_output( gifcp->osifcp, cpyp,
					    &( nxthopp->lclsockaddr ) );
		/* ??? how to pass resource id -- sockaddr w/ AF_RESOURCE ??*/
	    }

	    if ( errcod EQ 0 )
	    {
	        nxthopp->thruput.data_out_pkts++;
		nxthopp->thruput.data_out_bytes += st2len + encaplen;
	    }
	    else
	    {
	        BUGSTOP (-,errcod,0x10105);
	        nxthopp->thruput.data_out_drop_pkts++;
		nxthopp->thruput.data_out_drop_bytes += st2len + encaplen;

		if ( retcod EQ NoError )
	            retcod = IntfcFailure; /* ??? better translation */
	    }

          } /* end of for all next hops loop */
    } /* end of all exparray segments */


    if ( msgp NE (struct mbuf *) NULL )
    {
	if ( dups EQ 0 )
	{
	    Count (pkts_in_drop,1);	/* count packets dropped */
	    cp->prvhop.thruput.data_in_drop_bytes += st2len + encaplen;
	    cp->prvhop.thruput.data_in_drop_pkts++;
	}
	FreePkts ((struct aPktDesc *) msgp);
    }

    return ( NoError );

} /* end of st2_Forward */


/*
S* st2_intr ( mbufp, ifnetp )
s*
s*	COIP-K input interrupt routine.  Process queue if received packets.
s*
 */

  int					/* Next bugid 0x10221 */
st2_intr( ambufp, aifnetp )		/* from */
  struct mbuf		*ambufp;	/* NULL		st mbuf	ip mbuf */
  struct ifnet		*aifnetp;	/* <na> 	ifnetp	ifnetp	*/
{					/*  |              |       |    */
    enum Invoker {			BySoftcall, 	ByIPv5, ByIPv4p5 }
			 invoker;	/* How we got here */
    enum ST2Errors	 retcod = NoError; /* What failed */
 /* int			 buglen = st2_intrq.ifq_len; */
    int			 iplen,		/* Bytes in IP header */
			 netlen,	/* Bytes from network */
			 oldpri,
			 st2len;	/* Bytes in ST2 PDU */
    struct aConHshEntry	*conHashp;	/* Ptr from hashed HID */
    struct aNetIf	*gifcp = NO_NETIFP;
    struct aST2Header	*st2hdrp;	/* Ptr to ST2 Header */
    struct aST2pcb	*cp;		/* Ptr to aST2pcb for packet */
    struct ifnet	*ifnetp = (struct ifnet *) NULL;
    struct mbuf		*mp,		/* Ptr to sum m_lens */
			*msgp,		/* Ptr to first mbuf in chain */
			*st2p;		/* Ptr to mbuf w/ST2 Header in it */
    unsigned char	*ucp;		/* Ptr to first (IP)-layer byte */
    octet2		*cksump;	/* Ptr to shorts in ST2 Header */
    unsigned long	 cksum;		/* ST2 Header checksum */
    static int		 trapoff = 0;



    if ( (msgp = ambufp) EQ (struct mbuf *) NULL )
    {
	invoker = BySoftcall;		/* Process st2_intrq */
	if ( st2_stats.pkts_in_qfull NE st2_intrq.ifq_drops )
	    st2_stats.pkts_in_max = 0;
    }
    else
    {
	ucp = mtod (msgp,unsigned char *); /* Process single IP packet */

	if ( (*ucp & 0xF0) EQ 0x40 )
	{
	    if ( ucp[9] NE IPPROTO_ST )	/* Make sure Protocol 5 for ST */
	    {
		Count (pkts_in_ipv4px,1); /* count messages received */
		st2_stats.pkts_in_ipv4p = ucp[9];
		(void) m_free ( msgp );
		return ( 0 );
	    }

	    invoker = ByIPv4p5;		/* IP Version 4 Encapsulation */
	    Count (pkts_in_ipv4p5,1);	/* count messages received */
	}
	else
	{
	    invoker = ByIPv5;		/* AF_INET IP V5 packet */
	    Count (pkts_in_ipv5,1);	/* count messages received */
	}

	st2p = msgp;
	gifcp = Map2Gifc (aifnetp);
    }

    while ( TRUE )
    {
	iplen = 0;			/* Assume not encapsulated */

	/* ---------------------------------------------------------- */
	/* Via st2_intrq / PF_COIP */

	if ( invoker EQ BySoftcall )
	{
	    oldpri = splimp ();		/* Lock out other drivers */
	    {
		if ( st2_stats.pkts_in_max < st2_intrq.ifq_len )
		    st2_stats.pkts_in_max = st2_intrq.ifq_len;

		/* Kludge to get local net headers */
		IF_DEQUEUE ( &( st2_intrq ), msgp );
	    }
	    (void) splx ( oldpri );

	    if ( msgp EQ (struct mbuf *) NULL )
	    {
		st2_stats.pkts_in_qfull = st2_intrq.ifq_drops;
		return ( (int) retcod );
	    }

	    Count (pkts_in,1);	/* count messages received */
/*	    BUGSTOP (,msgp,0x10201);	/* ??? debugging */

	    /* ...................................................... */
	    /* Look for mbuf with st2 header in it */
	    st2p = msgp;		/* Assume the first */

	    if ( st2p->m_type EQ MT_HEADER )
	    {
		gifcp = (mtod (st2p,struct LocalNetInfo *))->gifcp;
		st2p = st2p->m_next;	/* Maybe the next */
	    }

	    if ( st2p EQ (struct mbuf *) NULL )
	    {				/* Inconsistency, st2_from* broken */
		Count (no_mt_data,1);	/* count no data */
		BUGGOTO (-,gifcp,0x10202,dropit);
	    }

	    if ( gifcp EQ NO_NETIFP )
	    {
		if (   (   (ifnetp = * (mtod (st2p,struct ifnet **)))
			EQ (struct ifnet *) NULL)
		    OR ((gifcp = Map2Gifc (ifnetp)) EQ NO_NETIFP) )
		{
		    Count (no_ifnet,1);	/* count no ifnet */
		    BUGGOTO (-,ifnetp,0x10203,dropit);
		}
	    }

	    if ( st2p->m_type EQ MT_IFADDR )
	    {
		st2p = st2p->m_next;	/* Maybe the next */
	    }
	    else			/* ??? Assume MT_DATA */
	    {
		/* drop interface pointer */
		st2p->m_off += sizeof (struct ifnet *);
		st2p->m_len -= sizeof (struct ifnet *);
	    }

	    if ( st2p EQ (struct mbuf *) NULL )
	    {
		Count (no_mt_data,1);	/* count no data */
		BUGGOTO (-,gifcp,0x10204,dropit);
	    }

	} /* end of BySoftcall */

	/* ---------------------------------------------------------- */
	/* Via IP Protocol 5 / PF_INET */

	else if ( invoker EQ ByIPv4p5 )
	{
	    iplen = (*ucp & 0x0F) * 4;

	    /* ??? make sure pulled up; MPULL_EXTRA 32 > ST-II = 16 */
	    /* max IP = 64 + ST-II = 16 should fit in first mbuf (112) */

#ifndef STIINOST1			/* Pass ST-I packets back to Raw IP */
	    if ( st2p->m_len < iplen + 1 )
	    {
		Count (msg_intrunc,1); /* count bad buffering */
		BUGGOTO (-,st2p->m_len,0x10205,dropit);
	    }

	    st2hdrp = Mkp (struct aST2Header *,ucp,iplen);

	    if ( st2hdrp->IPSTVer NE STV2 )
	    {
		Count (pkts_in_ipv4p5_rawip,1);
		rip_input(st2p);	/* ??? Give it back to (raw)ip */
		break;			/* Since not BySoftcall */
	    }
#endif  STIINOST1

	    if ( st2p->m_len < iplen + sizeofaST2Header )
	    {
		/* ??? fix up broken network drivers! */
		Count (msg_intrunc,1); /* count bad buffering */
		BUGGOTO (-,st2p->m_len,0x10206,dropit);
	    }

	    /* IP has already subtracted out it header length */
	    st2len = (Mkp (struct ip *,ucp,0))->ip_len;
	    st2p->m_off += iplen;	/* Drop ip header from mbuf */
	    st2p->m_len -= iplen;
	}

	/* ---------------------------------------------------------- */
	/* Via IP Version 5 == ST / PF_INET */

	else
	{
	    if ( st2p->m_len < sizeofaST2Header )
	    {
		/* ??? fix up broken network drivers! */
		Count (msg_intrunc,1); /* count bad buffering */
		BUGGOTO (-,st2p->m_len,0x10207,dropit);
	    }
	}

	/* ---------------------------------------------------------- */

	st2hdrp = mtod (st2p,struct aST2Header *);

	if ( st2p->m_type NE MT_DATA )
	{
	    Count (bad_mt_type,1);	/* count bad types */
	    BUGGOTO (-,(st2p->m_type << 16) | MT_DATA,0x10208,dropit);
	}

	if ( st2hdrp->IPSTVer NE STV2 )
	{
	    /* ??? give non-ST-II to raw IP */
	    Count (not_st2,1);		/* count not ST 2 */
	    BUGGOTO (-,(st2hdrp->IPSTVer << 16) | STV2,0x10209,dropit);
	}

	if ( ( (unsigned long) st2hdrp & ALIGNMENT ) NE 0 )
	{
	    Count (not_aligned,1);	/* count bad alignment */
	    BUGGOTO (-,gifcp,0x10210,dropit);
	}

	if ( st2p->m_len < sizeofaST2Header
			+ (HasSTTimestamp (st2hdrp)) ? sizeofaSTTimestamp : 0 )
	{
	    /* ??? fix it up */
	    Count (msg_intrunc,1);	/* count bad buffering */
	    BUGGOTO (-,st2p->m_len,0x10211,dropit);
	}

	/* ---------------------------------------------------------- */
	/* Inline ST2 header checksum verification */

	cksump = (octet2 *) st2hdrp;
	cksum  = *cksump++;
	cksum += *cksump++;
	cksum += *cksump++;
	cksum += *cksump++;
	if ( HasSTTimestamp ( st2hdrp ) )
	{
	    cksum += *cksump++;
	    cksum += *cksump++;
	    cksum += *cksump++;
	    cksum += *cksump++;
	}
	cksum = (cksum >> 16) + (cksum & 0xFFFF);
	cksum = (cksum >> 16) + (cksum & 0xFFFF);
	if (   (cksum NE 0xFFFF) AND (cksum NE 0)
#ifdef STIIDEBUG			/* (You do not see this :-) */
	    AND (st2hdrp->STChecksum NE 0) AND (NOT ConfigFlag (ST2FlgSTCksm))
#endif STIIDEBUG
	    )
	{
	    Count (bad_stcksum,1);	/* count bad checksum */
	    BUGGOTO (-,((st2hdrp->STChecksum << 16) | cksum),0x10212,dropit);
	}

	/* ---------------------------------------------------------- */
	/* Check for truncated packets ??? is truncated EQ data in error? */

	st2len = ntoh2 (st2hdrp->TotalBytes);
	netlen = 0;
	for ( mp = st2p ; mp NE (struct mbuf *) NULL ; mp = mp->m_next )
	{
	    /* Bug: some pkts being received with MCLBYTES of m_len */
	    /* Note "> st2len" case  might happen for short ethernet packets */
	    if (   (mp->m_len >= ConfigParm (api_MTU))
		OR (mp->m_len > (st2len - netlen + 60-14)) )
	    {
		if ( NOT trapoff++ )
		{
		    BUGSTOP (-,gifcp,0x10213);
		    BUGSTOP (,st2len,0x10219);
		    BUGSTOP (,mp->m_len,0x10220);
		}
		mp->m_len = st2len - netlen; /* Prevent subsequent errors */
	    }

	    netlen += mp->m_len;
	}

	if ( netlen < ntoh2 (st2hdrp->TotalBytes) )
	{
	    /* ??? drop or propagate truncated pkts, flag truncated
	           for nxt hops */

	    Count (msg_intrunc,1);	/* count truncated local net packets */
	    BUGGOTO (-,(netlen << 16) | ntoh2 (st2hdrp->TotalBytes),0x10214,
		     dropit);		/* ???send error message */
	}

	/* ---------------------------------------------------------- */
	/* Lookup aST2pcb based on HID */
					/* Use "our-order" HID */
	conHashp = st2_ConHsh + (ConfigParm (hid_mask) & st2hdrp->HID);
	if ( st2hdrp->HID NE conHashp->hid )
	{
	    Count (msg_inhidcoll,1);	/* count bad HIDs */
	    if ( conHashp->hid EQ (HID) 0 )
	    {
		BUGGOTO (-,(st2hdrp->HID << 16),0x10215,dropit);
	    }
	    BUGGOTO (-,(st2hdrp->HID << 16) | conHashp->hid,0x10216,dropit);
	}				/* ???error message, InvalidHID */

	if ( (cp = conHashp->cp) EQ NO_PCBP )
	{
	    Count (msg_innocon,1);	/* count no connection */
	    BUGGOTO (-,conHashp->hid,0x10217,dropit);
	}				/* ???error mesage, NoSuchConnection */

	if ( cp->dspin EQ (enum ST2Errors (*)()) NULL )
	{				/* I.e., st2_Forward, st2_CMPInput */
	    Count (no_dspin,1);	/* count no processing function */
	    BUGGOTO (-,conHashp->hid,0x10218,dropit);
	}				/* ???error message, ProtocolError */

					/* st2_Forward() st2_CMPInput() */
	retcod = cp->dspin ( gifcp, cp, msgp, st2p, (octet1 *) cksump, iplen );

	msgp = (struct mbuf *) NULL;	/* packet is gone */

	if ( invoker NE BySoftcall )
	    break;
	continue;

	/* ---------------------------------------------------------- */

 dropit:
	FreePkts ((struct aPktDesc *) msgp);

	if ( invoker NE BySoftcall )
	    break;			/* Only one packet to be processed */

    } /* end of while waiting packets loop */

    return ( 0 );

} /* end of st2_intr */

/* ??? send Notify on error, but not too often
 * Notify, RVLId = SVLId = 0, Ref = 0 or 1, lnkref = 0,
 * Sender = detector = our ifnet addr, Reason code = ?, HID parm = HID,
 * ErroredPDU = ... and
 * (*ifnetp->if_output)(ifnetp,pktp,&sockaddr???);
 */


/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
