#ifndef lint
static char rcsid_st2_api_c[] = "\
@(#)  $Header: st2_api.c,v 1.10 91/11/26 23:09:50 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint


/*
M* st2_api.c  BSD-style Sockets Application Interface to ST2 Services.
M* 
m* Overview
m*	
m*	The BSD-style socket API converts OS calls by the Application
m*	into various invocations of the st2_usrreq routine on the
m*	Application supplied argument (s) and the socket structure
m*	corresponding to the given file descriptor.  The socket
m*	structure in turn points to the ST-II protocol control block
m*	(aST2pcb).  We implement the control system calls (socket,
m*	bind, listen, accept, connect, and close) as well as the calls
m*	used for sending and receiving data (send, sendmsg, recvmsg).
m*
m*	For diagnostic and identification purposes, we record in the
m*	aST2pcb the Appllication pid and file descriptor associated
m*	with a stream; probably also ought to record the uid for
m*	accounting.
m*
 */

/*
m* Status:
m*   Features:
m*   Untested Features:
m*   Restrictions/Bugs:
m*   Things to do:
m*
 */


/* Module Revision History
 *
 * $Log:	st2_api.c,v $
 * Revision 1.10  91/11/26  23:09:50  clynn
 * Updated some documentation, minor improvements.
 * 
 * Revision 1.9  91/11/04  09:20:39  clynn
 * Updated for Public Domain Release.  Major changes: addition of Source
 * Routing, IP Encapsulation, HELLO protocol between neighbors, tracking
 * of neighbors, detection of component or agent failures & notification
 * to applications.  More consistant naming and format, including making
 * all external names begin with "st2_".  Moved some routines and data
 * structures to reduce external references.
 * 
 * Added SO_REUSEADDR support, pass both ReasonCode and DetectorIPAdr to
 * application, and fixed some timestamp bugs.
 * 
 * Revision 1.8  91/05/28  16:17:38  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 * Revision 1.1  91/03/15  18:33:22  clynn
 * Initial revision
 */

#define _FILE_ Strapi			/* Module bugid 3,12 */


/* External definitions */

#include <sys/errno.h>
#include <sys/types.h>		/* for netinet/in.h */
#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/systm.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <sys/file.h>		/* struct file */
#include <sys/user.h>		/* u.* uid = u.u_uid, pid = u.u_procp->p_pid */

#include <netinet/in.h>		/* before st2.h -> st2_api.h */
#include <net/if.h>
#include <net/route.h>		/* struct route for in_pcb.h */
#include <netinet/in_pcb.h>	/* struct inpcb for st2_output */

#include "st2_api.h"		/* Global Definitions, before st2.h */
#include "st2.h"		/* ST-II Implementation definitions,
				 * includes st2_api.h, before st2_cmp.h */
#include "st2_cmp.h"		/* SCM Protocol definitions */


extern octet4		 st2_parmManByOp[];

/* Local routines */

static int		 api_control (/* sop, cp, sast2p, sast2len, cmsghdrp,
					 datap */);
static int		 api_lbind (/* cp, sockadrp, maxsockbytes */);
static int		 api_makeifname (/* sp, pid */);
static int		 api_send (/* cp, gifcp, sast2p, sast2len, datap */);
static enum ST2Errors	 ApplIfcCheck (/* gifcp */);
static struct aNetIf	*ApplIfcCreate (/* cp */);
static int		 oopssubr ();
enum ST2Errors		 st2_ApplIfcBind (/* ifst2p, gwcp */);
int			 st2_ctlinput (/* "prc_req, sockaddrp */);
int			 st2_ctloutput (/* rop, so, level, optname, mp */);
int			 st2_FlowSpecFromAppl (/* apflowp, nflowp */);
int			 st2_input (/* mbufp, ifp */);
struct aST2pcb		*st2_ApplIfcCreate (/* listencp, arcp */);
int			 st2_output (/* inpcbp , mbufp */);
int			 st2_ToAppl (/* gifcp, pktp, ifsockadr */);
static int		 stsbappend (/* sb, pktp */);
static struct Target	*TargByApplIfc (/* gifcp, cp */);
/*macro			 UpdateApplId ();	*/


static char stifname[] = "ST-II.";
static char stiffree[] = "FreeST";


/* Needs #include <sys/socketvar.h> */

#define sotoST2pcb(sop) (Mkp (struct aST2pcb *, (sop)->so_pcb,0))


/* Define names for operations defined for st2_usrreq. */

#define USRREQLIST \
 aUReq(PRU_ATTACH,ATTACH,) \
 aUReq(PRU_DETACH,DETACH,) \
 aUReq(PRU_BIND,BIND,) \
 aUReq(PRU_LISTEN,LISTEN,) \
 aUReq(PRU_CONNECT,CONNECT,) \
 aUReq(PRU_ACCEPT,ACCEPT,) \
 aUReq(PRU_DISCONNECT,DISCONNECT,) \
 aUReq(PRU_SHUTDOWN,SHUTDOWN,) \
 aUReq(PRU_RCVD,RCVD,) \
 aUReq(PRU_SEND,SEND,) \
 aUReq(PRU_ABORT,ABORT,) \
 aUReq(PRU_CONTROL,CONTROL,) \
 aUReq(PRU_SENSE,SENSE,) \
 aUReq(PRU_RCVOOB,RCVOOB,) \
 aUReq(PRU_SENDOOB,SENDOOB,) \
 aUReq(PRU_SOCKADDR,SOCKADDR,) \
 aUReq(PRU_PEERADDR,PEERADDR,) \
 aUReq(PRU_CONNECT2,CONNECT2,) \
 aUReq(PRU_FASTTIMO,FASTTIMO,) \
 aUReq(PRU_SLOWTIMO,SLOWTIMO,) \
 aUReq(PRU_PROTORCV,PROTORCV,) \
 aUReq(PRU_PROTOSEND,PROTOSEND,) \


static char		*usrreqnames[] = {
#define aUReq(val,nam,desc) "nam",
    USRREQLIST
#undef aUReq
};


static char		*streqnames[ MAX_ST_REQUEST ] = {
#define aReq(value,name,comment) "name",
#define xReq(value,name,comment) (char *) NULL,
    STRequestList
#undef xReq
#undef aReq
};



/*
S* UpdateApplId ()
s*
s*	(BSD) Macro to update an Application's pid and file descriptor.
s*
 */


#define UpdateApplId()			/* from st2_usrreq*3 */		\
 {  int		  i;							\
    struct file	**fpp = u.u_ofile;					\
									\
    for ( i = 0 ; i <= u.u_lastfile ; i++, fpp++ )			\
    {									\
	if ( *fpp EQ (struct file *) NULL )				\
	    continue;							\
	if ( (long) (*fpp)->f_data EQ (long) sop )			\
	{								\
	    cp->usrfd = fpp - u.u_ofile;				\
	    break;			/* out of for loop */		\
	}								\
    } /* end of for loop */						\
    cp->usrpid = u.u_procp->p_pid;					\
    (void) api_makeifname( &( cp->gifcp->namebuf[0] ), cp->usrpid );	\
    cp->gifcp->osifcp->if_unit = cp->usrfd;				\
}

/*
S* api_control ( sop, cp, sast2p, sast2len, cmsghdrp, datap )
s*
s*	Routine to process application request to send control information.
s*	3 cases: sast2p w/o cmsghdrp, cmsghdrp w/o sast2p, both
s*
 */

  static int				/* Next bugid 0x30115 */
api_control( sop, cp, sast2p, sast2len, cmsghdrp, datap ) /* from st2_usrreq */
  struct socket		*sop;
  struct aST2pcb	*cp;
  struct sockaddr_st2	*sast2p;	/* ARGSUSED */
  int			 sast2len;
  struct cmsghdr	*cmsghdrp;
  struct aPktDesc	*datap;		/* we free it */ /* ARGSUSED */
{
    int			 errcod = 0;
    unsigned int	 arglen,
			 opts = 0;
    enum ST2Errors	 reason;
    enum STRequests	 cmsgtype = STReqUnspec,
			 func = STReqUnspec,
			 strequest = STReqUnspec;
    struct aApplEntity	*entp;
    struct aFlowSpec	*flowspecp = NULL;
    struct aNetIf	*gifcp;
    struct aParameter	*argp;
    struct aReasonCode	*reasoncodep = NULL;
    struct aUserData	*userdatap = NULL;
    struct Target	*targetp,
			*wtp;


    if ( datap NE NO_PKTDSCP )
	FreePkts (datap);		/* None currently allow data */

    /* -------------------------------------------------------------------- */
    /* Common cmsghdr processing */

    if ( cmsghdrp NE (struct cmsghdr *) NULL )
    {
	cmsgtype = (enum STRequests) cmsghdrp->cmsg_type;
	/* Save message type. */
	if ( cmsgtype NE STReqUnspec )
	    func = (enum STRequests) cmsghdrp->cmsg_type;

	/* Any additional information is a parameter list. */
	if ( (arglen = cmsghdrp->cmsg_len - sizeofCMsgHdr) > 0 )
	{
	    argp = Mkp (struct aParameter *, &( cmsghdrp->cmsg_data[0] ), 0);

	    while ( arglen >= 4 )
	    {
		if ( ((int) argp->pcode EQ 0)  OR  (argp->plen > arglen) )
		    break;

		switch ( argp->pcode )
		{
		case STpFlowSpec:	/* 2 */
		    flowspecp = (struct aFlowSpec *) argp; /* APIEQ? */
		    if ( flowspecp->FlowVer NE STFSVer3 )
		    {
			/* ??? why limit it */
			BUGRETURN (,EINVAL,0x30101,int);
		    }
		    break;

		case STReasonCode:	/* 30 */
		    reasoncodep = (struct aReasonCode *) argp;
		    break;

		case STpUserData:	/* 21 */
		    userdatap = (struct aUserData *) argp;
		    break;

		default:
		    break;
		} /* end of switch */

		arglen -= argp->plen;
		argp = Mkp (struct aParameter *, argp, argp->plen);

	    } /* while more parameters */

	    if ( arglen NE 0 )
	    {
		/* ill-formatted parameters */
		BUGRETURN (,EINVAL,0x30102,int);
	    }
	}
    }

    /* -------------------------------------------------------------------- */
    /* Initial sockaddr_st2 processing */

    targetp = NO_TRGP;

    if ( sast2p NE (struct sockaddr_st2 *) NULL )
    {					/* ??? 1sock; cp is @target, not @GW */
	strequest = (enum STRequests) sast2p->st_request;
	/* Save message type. */
	if ( strequest NE STReqUnspec )
	{
	    if (    (func NE STReqUnspec)
		AND (func NE strequest)
		AND (   (strequest NE STReqAccept)
		     OR (cmsgtype NE STReqConnect)) )
	    {
		/* Inconsistent requests */
		BUGSTOP (-,((sast2p->st_request << 8) | (int) func),0x30103);
	    }

	    func = strequest;
	    opts = sast2p->st_options;
	}

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */
	/* Scan to find the STRmtAppEnt binding that matches the sockaddr_st2
	 * passed to us.
	 */

	/* ??? ought to look through parameter list */
	entp = (struct aApplEntity *) &( sast2p->st_parms[0] );

	if (    (sast2len > sizeofsockaddr_st2header)
	    AND (entp->pcode EQ (int) STRmtAppEnt) )
	{
	    /* Scan targets if at Origin */
	    if ( (cp->flags & LocalOrigin) NE 0 )
	    {
		wtp = cp->targetsp;
		for ( ; wtp NE NO_TRGP ; wtp = wtp->nextp )
		{
		    if (   (wtp->targetIPAdr NE entp->IPAdr)
			OR (wtp->sap.SAPBytes NE entp->SAPBytes)
			OR (Bcmp (wtp->sap.SAPp,&( entp->SAP[0] ),
				  entp->SAPBytes) NE 0) )
			continue;	/* ??? check TS_BUSY too? */

		    targetp = wtp;	/* ??? check srcrut too? */
		} /* end of for all targets */
	    }
	    else			/* at target */
	    {
		ForAllGifc (gifcp, cp->gifcp,)
		{
		    if ( gifcp->ifc_gwcp NE NO_PCBP )
		    {
			/* ??? suspect Origin misuse, soname.applent */
			if ( 0 < st2_ApplEntityMatch( entp,
					&( gifcp->ifc_gwcp->Origin ) ) )
			{
			    targetp = TargByApplIfc( gifcp,
							  gifcp->ifc_gwcp );
			}
		    }			/* target is in gwcp */
		} /* end of for loop */
	    }
	} /* end of sast2 has parameter */
    } /* end of have a sast2p */

    /* -------------------------------------------------------------------- */
    /* If the sast2p was omitted, we can get the target only if
     * there is just one stream bound to this connection.
     */
    if ( targetp EQ NO_TRGP )
    {
	if (    (cp->gifcp NE NO_NETIFP)
	    AND (cp->gifcp->ifc_nextp EQ NO_NETIFP)
	    AND (cp->gifcp->ifc_gwcp NE NO_PCBP) )
	{
	    targetp = TargByApplIfc( cp->gifcp, cp->gifcp->ifc_gwcp );
					/* target is in gwcp */
	}
    }

    /* -------------------------------------------------------------------- */

    switch ( func )
    {
					/* We are Origin */
/*  case STReqChange:			/* CHANGE to make */
    case STReqConnect:			/* App->ST send CONNECT */
    case STReqDisconnect:		/* App->ST send DISCONNECT */
	/* Global forms don't require a target */
	if ( (cp->flags & LocalOrigin) EQ 0 )
	{
	    /* ??? not correct error - "wrong end" */
	    BUGRETURN (-,EOPNOTSUPP,0x30104,int);
	}
	break;
					/* We are Target */
    case STReqAccept:			/* App->ST ACCEPT connection */
/*  case STReqChangeRequest:		/* App->ST request origin to make change */
    case STReqRefuse:			/* App->ST send REFUSE */
	if ( targetp EQ NO_TRGP )
	{
	    /* ??? not correct error - "invalid peer" */
	    BUGRETURN (-,ENOTCONN,0x30105,int);
	}
	/* Fall through */

/*  case STReqNotify:			/* NOTIFY  sent/received */
	if ( (cp->flags & LocalTarget) EQ 0 )
	{
	    /* ??? not correct error - "wrong end" */
	    BUGRETURN (-,EOPNOTSUPP,0x30106,int);
	}
	break;
					/* We could be either */
    case STReqBind:			/* App->ST bind () */
    case STReqParms:			/* App->ST specify parametrers */

    case STCtlPending:			/* Control info waiting for appl. */
    case STReqUnspec:			/* App->ST request as per function */
    default:
	BUGSTOP (,sast2p->st_request,0x30107);
	return ( EOPNOTSUPP );

    } /* end of switch */

    /* -------------------------------------------------------------------- */

    switch ( func )
    {
    case STReqAccept:			/* App->ST ACCEPT connection */
	/* If flowspec is omitted here, "flowspec" will remain zero and
	 * accept_Message will use the (default) flowspec kept with the target
	 */

	/* Don't echo STReqConnect parm */
	if ( cmsgtype EQ STReqConnect )
	    userdatap = (struct aUserData *) NULL;

	/* ??? permit parms from sast2 also */

	reason = st2_DoAccept( targetp, flowspecp, opts, userdatap );
	if ( reason NE NoError )
	    errcod = ENOBUFS;		/* ??? better error */
	break;


    case STReqBind:			/* App->ST bind () */
	BUGSTOP (,sast2p->st_request,0x30108);
	return ( EOPNOTSUPP );

    case STReqChange:			/* CHANGE to make */
	BUGSTOP (,sast2p->st_request,0x30109);
	/* ??? may not have targetp */
	return ( EOPNOTSUPP );

    case STReqChangeRequest:		/* App->ST ask origin to STReqChange */
	BUGSTOP (,sast2p->st_request,0x30110);
	return ( EOPNOTSUPP );

    case STReqConnect:			/* App->ST send CONNECT */
	if (   (sast2p EQ (struct sockaddr_st2 *) NULL)
	    OR (sast2len < sizeof (struct sockaddr_st2)) )
	{
	    BUGRETURN (-,EINVAL,0x30111,int);
	}

	errcod = st2_DoConnect( cp, sast2p, sast2len, flowspecp, userdatap );
	/* NB: cp may no longer be valid */

	if (    (errcod EQ 0)
	    AND ((cp = sotoST2pcb (sop)) NE NO_PCBP ) )
	{
	    soisconnecting ( sop );
	    if ( IsState (PCB,cp,pcb_state,CINS_OPEN) )
	    {
		soisconnected ( sop );
	    }
	}

	return ( errcod );

    case STReqDisconnect:		/* App->ST send DISCONNECT */
					/* ??? G-Bit vs target list */
	if ( reasoncodep NE (struct aReasonCode *) NULL )
	    reason = (enum ST2Errors) reasoncodep->reason;
	else
	    reason = ApplDisconnect;		/* default */

	errcod = st2_DoDisconnect( cp, reason, targetp, userdatap );

	break;

    case STReqNotify:			/* NOTIFY  sent/received */
	BUGSTOP (,sast2p->st_request,0x30112);
	return ( EOPNOTSUPP );

    case STReqParms:			/* App->ST specify parametrers */
	BUGSTOP (,sast2p->st_request,0x30113);
	return ( EOPNOTSUPP );

    case STReqRefuse:			/* App->ST send REFUSE */
	if ( reasoncodep NE (struct aReasonCode *) NULL )
	    targetp->error = (enum ST2Errors) reasoncodep->reason;
	else
	    targetp->error = ApplDisconnect; /* default */


	if (    (userdatap NE (struct aUserData *) NULL)
	    AND (cmsgtype NE STReqConnect) )
	{
	    char *udp = AddSpace (char *, &( targetp->cp->fwdrfreespace ), 
				  PKT_MD_PROVHD, (int) userdatap->plen);

	    if ( udp NE (char *) NULL )
	    {
		Bcopy (userdatap,udp,userdatap->plen);
	    }
	    /* ??? else ignore it */
	}

	/*??? Other options for refuse: ErrdPDU, RecordRoute */

	/* ??? check TS_CONNECTED/TS_OPEN */

#ifdef NOTYET		/* no resources allocated between gateway and socket */
	SetTargState (targetp,TS_CLOSED);
#else
	SetTargState (targetp,TS_RELEASED);
#endif NOTYET
	SetTargState (targetp,TS_FAILED);

	targetp->cp->needupdate++;	/* Something to do */
	st2_PCBUpdate( targetp->cp );
	/* cp may be invalid */

	/* ??? how get error from st2_PCBUpdate -- targetp->error ???? */
	reason = NoError;
	break;

    /* ??? others: GroupName */
    case STCtlPending:			/* Control info waiting for appl. */
    case STReqUnspec:			/* App->ST request as per function */
    default:
	BUGSTOP (,sast2p->st_request,0x30114);
	return ( EOPNOTSUPP );

    } /* end of switch */

    return ( errcod );
}

/*
S* api_lbind ( cp, sockadrp, maxsockbytes )
s*
s*	Routine to bind a socket to a <IPAdr, NextPcol, SAPBytes, SAP.
s*
 */

  static int				/* Next bugid 0x30205 */
api_lbind( cp, sockadrp, maxsockbytes ) /* from st2_usrreq*3 */
					/* PRU_BIND PRU_CONNECT PRU_LISTEN */
  struct aST2pcb	*cp;
  struct sockaddr_st2	*sockadrp;	/* 	->	null	null */
  int			 maxsockbytes;	/* 	#	 0	 0 */
{
    unsigned char	*ucp;
    int			 n;
    struct aApplEntity	*entp,
			*wentp;
    struct aST2pcb	*wcp;


    /* Default selection of local SAP */
    if (   (sockadrp EQ (struct sockaddr_st2 *) NULL)
	OR (maxsockbytes <= 0) )
    {
	ucp = &( cp->soname.applent.SAP[0] );
	n = cp->soname.applent.SAPBytes;

	for ( ; n > 0 ; n-- )
	    if ( *ucp++ NE 0 )
		return ( 0 );		/* already specified */

	if ( cp->soname.applent.SAPBytes EQ 0 )
	{
	    cp->soname.applent.SAPBytes = hton2 (2);
	    cp->soname.applent.plen = (octet1) 12;
	}

	for (;;)
	{
	    if ( (0xFFFF & st2_stats.sap_next) EQ 0 )
	    {
		st2_stats.sap_next &= ~ 0xFFFF;
		st2_stats.sap_next |= 0xC000 + 1024;
	    }

	    cp->soname.applent.SAP[0] = 0xFF & (st2_stats.sap_next >> 8);
	    cp->soname.applent.SAP[1] = 0xFF &  st2_stats.sap_next++;

	    /* (overly conservative) uniqueness check */
	    ForAllpcbs (wcp)
	    {
		if ( wcp EQ cp )
		    continue;

		if (    (   wcp->soname.applent.SAP[1]
			 EQ  cp->soname.applent.SAP[1])
		    AND (   wcp->soname.applent.SAP[0]
			 EQ cp->soname.applent.SAP[0]) )
		{
		    break;		/* Not unique */
		}
	    } /* end of for all pcbs */

	    if ( wcp EQ &( st2_ConTbl[0] ) )
		break;			/* unique, out of for loop */

	} /* end of for non-unique loop */

	/* ??? defer IPAdr until routing */
	/* ??? default NextPcol (0) */

	cp->soname.applent.pcode = (octet1) STLclAppEnt;
	cp->soname.sa_family = AF_COIP;

	return ( 0 );
    }

    /* Application specification of SAP */
    if ( maxsockbytes > sizeof (cp->soname) )
    {
	BUGRETURN (-,EINVAL,0x30201,int);
    }

    if ( cp->soname.applent.pcode NE (octet1) 0 )
    {
	/* ??? what's wrong with what's already there/merge */
	BUGRETURN (-,EINVAL,0x30202,int);
    }
					/* ??? too restrictive? */
    /* Must give address as an single, local application entity parameter */
    entp = (struct aApplEntity *) &( sockadrp->st_parms[0] );
    if (   (maxsockbytes NE sizeofsockaddr_st2header + (int) entp->plen)
	OR (entp->pcode NE (octet1) STLclAppEnt) )
    {
	BUGRETURN (-,EINVAL,0x30203,int);
    }

    /* If an IP address was specified, make sure its one of our addresses */
    if ( entp->IPAdr NE INADDR_ANY )
    {
	if ( NOT st2_IPAdrFunc( IPAdrFuncValidate, &( entp->IPAdr ) ) )
	    BUGRETURN (,EADDRNOTAVAIL,0x30204,int);
    }

    /* Check for uniqueness of NextPcol and SAP for all STLclAppEnts */
    ForAllpcbs (wcp)
    {
	if ( wcp EQ cp )
	    continue;			/* Don't compare to self */

	wentp = (struct aApplEntity *) &( wcp->soname.applent );

	if ( wentp->pcode NE (octet1) STLclAppEnt )
	    continue;			/* Skip unbound, or remote */

	/* Unique if non-wild protocols differ */
	if (    (wentp->NextPcol NE entp->NextPcol)
	    AND (wentp->NextPcol NE 0)
	    AND (entp->NextPcol NE 0) )
	    continue;
	/* NextPcols same or wild */

	/* Not unique if either SAP is unspecified ... */
	if ( (wentp->SAPBytes EQ 0) OR (entp->SAPBytes EQ 0) )
	    return ( EADDRINUSE );

	/* ... or the SAP lengths and value match */
	if (    (wentp->SAPBytes EQ entp->SAPBytes)
	    AND (Bcmp (&( wentp->SAP[0] ),&( entp->SAP[0] ),
		       entp->SAPBytes) EQ 0) )
	{
	    if (   (    (cp->sop NE (struct socket *) NULL)
		    AND ((cp->sop->so_options & SO_REUSEADDR) EQ 0))
		OR (    (wcp->sop NE (struct socket *) NULL)
		    AND ((wcp->sop->so_options & SO_REUSEADDR) EQ 0)) )
		return ( EADDRINUSE );
	}
    } /* end of for all pcb uniqueness loop */

    IfLog (DBG_SKT_RUT)
	(void) printf ( "Binding to %s socket\n",
		       (sockadrp->st_options & (int) STOpt1socket)
		       ? "shared" : "normal" );

    Bcopy (sockadrp,&( cp->soname ),maxsockbytes);

    return ( 0 );
}

/*
S* api_makeifname ( sp, pid )
s*
s*	Routine to make an ST-II pseudo device name of the form "ST-II.<pid>".
s*
 */

  static int				/* Next bugid 0x30300 */
api_makeifname( sp, pid )		/* from ApplIfcCreate st2_intr */
  char			*sp;
  int			 pid;
{
    char		*xp = stifname;
    int			 div = 10000;

    while ( *xp NE '\0' )
	*sp++ = *xp++;

    pid &= 0xFFFF;

    if ( pid EQ 0 )
	*sp++ = '0';
    else
    {
	while ( div > 0 )
	{
	    *sp = '0' + pid / div;
	    pid %= div;
	    div /= 10;
	    if ( *sp EQ '0' )
		continue;
	    sp++;
	    break;
	}

	while ( div > 0 )
	{
	    *sp++ = '0' + pid / div;
	    pid %= div;
	    div /= 10;
	}
    }    
    *sp = '\0';
    return;
}

/*
S* api_send ( cp, gifcp, sast2p, sast2len, datap )
s*
s*	Routine to process application request to send data.
s*
 */

  static int				/* Next bugid 0x30402 */
api_send( cp, gifcp, sast2p, sast2len, datap ) /* from st2_usrreq*2 */
  struct aST2pcb	*cp;
  struct aNetIf		*gifcp;		/* ARGSUSED */
  struct sockaddr_st2	*sast2p;	/* may be absent, if present long */
  int			 sast2len;	/* enough for header w/o parms */
  struct mbuf		*datap;		/* we free it, caller verifies */
{
    struct mbuf		*pktp;
    struct aPktDesc	*stpacketp;
    struct aST2Header	*st2hdrp;
    unsigned int	 totalbytes;
    unsigned long	 checksum;
    register octet2	*chkp;
    int			 retcod = 0;
    union {
	unsigned long	 full;
	struct {
	    unsigned short half[2];
	}		 parts;
    }			 splitsum;


    while ( (pktp = datap) NE (struct mbuf *) NULL )
    {
	if ( pktp->m_act NE (struct mbuf *) NULL )
	    BUGSTOP (-,1,0x30401);	/* should never happen */

	/* Remove packet from queue. */
	datap = pktp->m_act;
	pktp->m_act = (struct mbuf *) NULL;

	/* Get an mbuf for the st header, put in front of data */
	/* ??? ought to make upper layer leave enough room! */
	stpacketp = NO_PKTDSCP;
	st2hdrp = GetSpace (struct aST2Header *, &( stpacketp ),
			   PKT_DATA, PKT_MD_NEW, PKT_XA_PR,
			   0,0, sizeofaST2Header, sizeofaSTTimestamp,4);
	if ( st2hdrp EQ (struct aST2Header *) NULL )
	{
	    FreePkts ((struct aPktDesc *) pktp);
	    retcod = ENOBUFS;
	    continue;
	}

	/* Add header mbuf in front of data mbuf chain. */
	stpacketp->pkt_nxtbufp = (struct aPktDesc *) pktp;

	/* Calculate length of the data portion */
	totalbytes = sizeofaST2Header;
	while ( pktp NE (struct mbuf *) NULL )
	{
	    totalbytes += pktp->m_len;
	    pktp = pktp->m_next;
	} /* end of while */

	/* ??? chek if smaller then mtu */

	/* Accumulate checksum as we build header. */
	checksum = 0;

	/* Build protocol version and option fields. */
	st2hdrp->IPSTVer = hton1 (STV2);
	/* priority-per-packet from sast2p->options */
	{   unsigned int temp = 0;
	    unsigned long sec = 0, fract = 0;

	    if (    (sast2len >= sizeofsockaddr_st2header)
		AND (sast2p NE (struct sockaddr_st2 *) NULL) )
	    {
		temp = sast2p->st_options;
		sec = sast2p->st_sec;
		fract = sast2p->st_fract;
	    }

	    st2hdrp->STPriOpt = hton1 (cp->optmask & (temp | cp->optneg));

	    /* Add timestamp if selected; accumulate checksum. */
	    if ( HasSTTimestamp ( st2hdrp ) )
	    {
		unsigned long *stmp;

		/* Make room for the timestamp. */
		stmp = AddSpace (unsigned long *, &( stpacketp ), PKT_MD_PKT,
				 sizeofaSTTimestamp);
		/* ??? if ( stmp EQ (unsigned long *) NULL ) ... */
		totalbytes += sizeofaSTTimestamp;

		/* Get time in usec's and convert to NTP format,
		   allow user to specify. */
		if (   (sec NE (unsigned long) 0)
		    OR (fract NE (unsigned long) 0) )
		{
		    stmp[0] = hton4 (sec);
		    stmp[1] = hton4 (fract);
		}
		else
		{
		    struct timeval tv;


		    uniqtime ( &( tv ) );
		    /* Convert to NTP format; 1970 - 1900 in seconds */
		    stmp[0] = hton4 (tv.tv_sec + 2208988800);
		    stmp[1] = hton4 (  (tv.tv_usec * 0x10c6)
				     + ((tv.tv_usec * 0xf7a) >> 12));
		}

		/* Compute checksum. */
		chkp = (octet2 *) stmp;
		checksum += *chkp++;
		checksum += *chkp++;
		checksum += *chkp++;
		checksum += *chkp++;
	    }
	}

	/* Now we have total bytes, so set it, in network order */
	st2hdrp->TotalBytes = hton2 (totalbytes);

	/* No hid, it came from us.  st2_Forward () will update it. */
	st2hdrp->HID = (HID) 0;

	/* Zero checksum field to compute required checksum value. */
	st2hdrp->STChecksum = 0;

	/* Standard ST header fields now complete, add to checksum. */
	chkp = (octet2 *) st2hdrp;
	checksum += *chkp++;
	checksum += *chkp++;
	checksum += *chkp++;
	checksum += *chkp++;

	/* Handle end-around carry and set header field. */
	splitsum.full = checksum;
	checksum = splitsum.parts.half[0] + splitsum.parts.half[1];
	if ( checksum > 0xffff )	/* Fast carry wrap-around */
	    checksum += (1 - 0x10000);

	kassert (checksum < (unsigned long) 0x10000);
	st2hdrp->STChecksum = ~ checksum;

	/* Pass packet to output routine. */

	(void) st2_Forward( NO_NETIFP, cp, (struct mbuf *) stpacketp,
			   (struct mbuf *) stpacketp, (octet1 *) NULL, 0 );
	/* ??? error */

    } /* end of while packets to go loop */

    return ( retcod );
}

/*
S* ApplIfcCheck ( gifcp )
s*
s*	Routine to verify that an gifc pointer is pointing to a
s*	pseudo interface structure.
s*
 */

  static enum ST2Errors			/* Next bugid 0x30503 */
ApplIfcCheck( gifcp )			/* from st2_ApplIfcBind st2_ToAppl */
  struct aNetIf		*gifcp;
{
    if ( gifcp EQ NO_NETIFP ) {
	BUGRETURN (-,InvalidPointer,0x30501,enum ST2Errors);
    }

    if ( Bcmp (gifcp->ifc_name,stifname,strlen (stifname)) EQ 0 )
	return ( NoError );

    if ( gifcp->ifc_name EQ stiffree )
    {
	kassert (gifcp->ifc_name NE stiffree);
    }
    else
    {
	kassert (Bcmp (gifcp->ifc_name,stifname,strlen (stifname)) EQ 0);
    }

    BUGRETURN (-,InvalidPointer,0x30502,enum ST2Errors);
}

/*
S* ApplIfcCreate ( cp )
s*
s*	Routine to create pseudo-network interface for application.
s*
 */

  static struct aNetIf *		/* Next bugid 0x30600 */
ApplIfcCreate( cp )			/* from st2_ApplIfcCreate st2_usrreq */
  struct aST2pcb	*cp;		/* @Origin,@Target	@GW,@Origin */
{
    struct aNetIf	*ifst2p;
    struct aPktDesc	*pp;


    pp = NO_PKTDSCP;
    ifst2p = GetSpace (struct aNetIf *, &( pp ),
		       PKT_NETIF, PKT_MD_NEW, PKT_XA_PS,
		       0,0, sizeof (struct aNetIf),
		       sizeof (struct ifnet),0);
    if ( ifst2p EQ NO_NETIFP )
    {
	return ( NO_NETIFP );
    }

    Bzero (ifst2p,sizeof (struct aNetIf));

    ifst2p->osifcp = AddSpace (struct ifnet *, &( pp ), PKT_MD_PKT,
			       sizeof (struct ifnet));

    Bzero (ifst2p->osifcp,sizeof (struct ifnet));

    ifst2p->ifc_selfp = pp;
    ifst2p->osifchandle = (unsigned long) ifst2p->osifcp;
    (void) api_makeifname( &( ifst2p->namebuf[0] ), cp->usrpid );
    ifst2p->lclhdrlen = 6;		/* maintain alignment with sa_family */

    /* Make it look like a normal OS network interface */

    ifst2p->osifcp->if_name = &( ifst2p->namebuf[0] );
    ifst2p->osifcp->if_unit = cp->usrfd;
    ifst2p->osifcp->if_mtu = ConfigParm (api_MTU);
    ifst2p->osifcp->if_flags = IFF_UP | IFF_POINTOPOINT | IFF_NOARP
			 | IFF_NOTRAILERS | IFF_RUNNING | IFF_PRIVATE;
    ifst2p->osifcp->if_init = oopssubr;
    ifst2p->osifcp->if_output = st2_ToAppl;
    ifst2p->osifcp->if_ioctl = oopssubr;
    ifst2p->osifcp->if_reset = oopssubr;
    ifst2p->osifcp->if_watchdog = oopssubr;
    ifst2p->osifcp->if_input = oopssubr;
    ifst2p->osifcp->if_ctlin = oopssubr;
    ifst2p->osifcp->if_ctlout = oopssubr;

    IfLog (DBG_DTA_STR)
	printf ( "ApplIfcCreate at 0x%x for cp 0x%x  %s unit %d\n", ifst2p,
		cp, ifst2p->ifc_name, ifst2p->osifcp->if_unit );

    cp->refcnt++;
    cp->refwho |= RefGifccp;
    IfLog (DBG_CNT_PCB)
	printf ( "refcnt: cp = %8x  refcnt = %2d  ApplIfcCreate    RefGifccp\n",
		cp, cp->refcnt );
    ifst2p->ifc_cp = cp;

    return ( (struct aNetIf *) ifst2p );
}

/*
S* st2_ApplIfcBind ( ifst2p, gwcp )
s*
s*	Routine to link application pseudo-network interface to intermediate
s*	ST agent.
s*
 */

  enum ST2Errors			/* Next bugid 0x30703 */
st2_ApplIfcBind( ifst2p, gwcp )		/* from st2_DoConnect st2_PCBUpdate */
  struct aNetIf		*ifst2p;	/* 		ok */
  struct aST2pcb	*gwcp;		/* 		ok */
{
    struct aST2pcb	*cp;		/* of Origin/Gateway or Target */
    int			 oldpri;
    enum ST2Errors	 error;


    if ( (error = ApplIfcCheck( ifst2p )) NE NoError )
	return ( error );

    /* Socket has been closed already. */
    if ( (ifst2p->osifcp->if_flags & IFF_UP) EQ 0 )
    {
	BUGSTOP (,ifst2p->osifcp->if_flags,0x30701); /* ??? vs reopen */
	return ( SAPUnknown );
    }

    oldpri = splnet ();
    {
	/* Already bound to another stream. */
	if ( ifst2p->ifc_gwcp NE NO_PCBP )
	{
	    (void) splx ( oldpri );
	    BUGSTOP (,ifst2p->ifc_gwcp,0x30702);
	    return ( RouteLoop );	/* old path not yet torn down */
	}

	ifst2p->ifc_gwcp = gwcp;
	gwcp->refcnt++;

	if ( ifst2p->ifc_cp EQ gwcp )
	{
	    gwcp->refwho |= RefGifcgwcpo;
	    IfLog (DBG_CNT_PCB)
		printf ( "refcnt: cp = %8x  refcnt = %2d  st2_ApplIfcBind  RefGifcgwcpo\n",
			gwcp, gwcp->refcnt );
	}
	else
	{
	    gwcp->refwho |= RefGifcgwcpt;
	    IfLog (DBG_CNT_PCB)
		printf ( "refcnt: cp = %8x  refcnt = %2d  st2_ApplIfcBind  RefGifcgwcpt\n",
			gwcp, gwcp->refcnt );
	}
    }
    (void) splx ( oldpri );


    /* ??? target aST2pcb ought to be combined with gateway & ?origin */
    /* ??? too soon for dafaulted IPAdrs ... get st2_ToAppl to do it */
    if ( (cp = ifst2p->ifc_cp) NE gwcp ) /* a target */
    {
	Bcopy (&( gwcp->soname ),&( cp->soname ),
					sizeof (struct sockaddr_st2soname));

	Bcopy (&( gwcp->Name ),&( cp->Name ),sizeof (struct aName));

	Bcopy (&( gwcp->Origin ),&( cp->Origin ),sizeof (struct aOrigin));

	Bcopy (&( gwcp->flow_spec ),&( cp->flow_spec ),
			sizeof (struct aFlowSpec3)); /* ??? other versions */

	cp->precedence = gwcp->precedence; /* ??? must implement */
	cp->optapi = gwcp->optapi;	/* ??? from listener */
	cp->detectorIPAdr = gwcp->detectorIPAdr;

    } /* end of copying gateway info to target */

    return ( NoError );
}

/*
S* st2_FlowSpecFromAppl ( apflowp, nflowp )
s*
s*	Routine to import a FlowSpec from an application.
s*
 */

  int					/* Next bugid 0x30800 */
st2_FlowSpecFromAppl( apflowp, nflowp ) /* from st2_DoConnect*2 */
					/*??? should be in flowver3 module */
  struct aFlowSpec	*apflowp;
  struct aFlowSpec3	*nflowp;
{
    struct aFlowSpec3	*apflow3p;


    switch ( apflowp->FlowVer )
    {
    case ntoh1 (STFSVer3):
	/* Byte order provided by the application is net-order */

	apflow3p                 = (struct aFlowSpec3 *) apflowp;
	nflowp->pcode            = apflowp->pcode;
	nflowp->plen             = (apflowp->plen < sizeof (struct aFlowSpec3))
	  ? apflowp->plen : sizeof (struct aFlowSpec3);
	nflowp->FlowVer          = apflowp->FlowVer;
	nflowp->fill1            = apflowp->fill1;
	nflowp->DutyFactor       = apflow3p->DutyFactor;
	nflowp->ErrorRate        = apflow3p->ErrorRate;
	nflowp->Precedence       = apflow3p->Precedence;
	nflowp->Reliability      = apflow3p->Reliability;
	nflowp->Tradeoffs        = apflow3p->Tradeoffs;
	nflowp->RecoveryTimeout  = apflow3p->RecoveryTimeout;
	nflowp->LimitOnCosts     = apflow3p->LimitOnCosts;
	nflowp->LimitOnDelay     = apflow3p->LimitOnDelay;
	nflowp->LimitOnPDUBytes  = apflow3p->LimitOnPDUBytes;
	nflowp->LimitOnPDURate   = apflow3p->LimitOnPDURate;
	nflowp->MinBytesXRate    = apflow3p->MinBytesXRate;
	nflowp->AccdMeanDelay    = apflow3p->AccdMeanDelay;
	nflowp->AddDelayVariance = apflow3p->AddDelayVariance;
	nflowp->DesPDUBytes      = apflow3p->DesPDUBytes;
	nflowp->DesPDURate       = apflow3p->DesPDURate;

	return ( 0 );

    default:
	return ( EINVAL );
    }
}

/*
S* st2_ApplIfcCreate ( listencp, arcp )
s*
s*	Routine to create a new socket for an application.
s*
 */

  struct aST2pcb *			/* Next bugid 0x30904 */
st2_ApplIfcCreate( cp, arcp )		/* from st2_PCBUpdate */
  struct aST2pcb	*cp;		/* Of listening socket */
  struct aRouteChoice	*arcp;
{
    int			 oldpri;
    struct aNetIf	*gifcp;
    struct socket	*sop;


    /* st_option STOpt1socket means use single socket,
     * rather than lister+multiple accept
     */

    /* ??? since cannot do a connect () after listen (), ought
     * to check OPENING, OPEN, CLOSING & 1sock, too.
     */

    if ( (cp->soname.st_options & (int) STOpt1socket) EQ 0 )
    {					/* st2_newconn */
	IfLog (DBG_SKT_RUT)
	    (void) printf ( "...queuing new socket\n" );

	/* Note that sonewconn calls PRU_ATTACH, so the socket
	 * returned is a new correctly bound a aST2pcb.
	 */
	/* ??? calls soreserve twice ??? spl level */
	sop = sonewconn ( cp->sop );
	if (   (sop EQ (struct socket *) NULL)
	    OR ((cp = sotoST2pcb (sop)) EQ NO_PCBP) )
	{
	    BUGSTOP (-,CantGetResrc,0x30901);
	    arcp->flags |= RouteFailed | RouteNoSkt;
	    arcp->error = CantGetResrc;
	    return ( cp /*??? NO_PCBP*/ );
	}

	cp->usrpid = 0;			/* was not in user context */
	cp->usrfd = 65535;		/* will correct at accept () */

	cp->flags |= LocalTarget;
	arcp->gifcp = cp->gifcp;
    }
    else				/* ??? not yet support 1sock */
    {
	BUGSTOP (-,Inconsistency,0x30902);

	/* Add another pseudo interface to socket aST2pcb */
	gifcp = ApplIfcCreate( cp );

	if ( gifcp EQ NO_NETIFP )
	{
	    BUGSTOP (-,CantGetResrc,0x30903);
	    arcp->flags |= RouteFailed | RouteNoSkt;
	    arcp->error = CantGetResrc;
	}
	else
	{
	    oldpri = splnet ();
	    {
		gifcp->ifc_nextp = cp->gifcp;
		cp->gifcp = gifcp;
	    }
	    (void) splx ( oldpri );

	    cp->flags |= LocalTarget;
	    arcp->gifcp = gifcp;
	}
    }

    return ( cp );
}

/*
S* oopssubr ()
s*
s*	Dummy routine for unused pseudo-network interface vector entries.
s*
 */

  static int				/* Next bugid 0x30a02 */
oopssubr()				/* from */
					/* bad dsp via ApplIfcCreate */
{
	BUGRETURN (-,45/*EOPNOTSUPP*/,0x30a01,int);
}

/*
S* st2_ctlinput ( "prc_req,sockaddrp" )
s*
s*	NOTYET
s*
 */
#ifdef NOTYET
|
|  int					/* Next bugid 0x30b00 */
|st2_ctlinput(/*"prc_req,sockaddrp"*/)	/* from <nowhere/vector> */
|{
|}
#endif NOTYET

/*
S* st2_ctloutput ( "op, socketp, level, optname, mbufpp" )
s*
s*	Implements getsockopt () and setsockopt ().
s*	NOTYET
s*
 */

  int					/* Next bugid 0x30c00 */
st2_ctloutput( rop, so, level, optname, mp ) /* from */
  int			  rop;		/* PRCO_GETOPT or PRCO_SETOPT */
  struct socket		 *so;		/* ARGSUSED */
  int			  level,	/* 1190 for us */
			  optname;
  struct mbuf		**mp;
{
#ifdef  STIIDEBUG
    (void) printf ( "Call st2_ctloutput, rop %d, lev %d, optname %d\n",
		    rop, level, optname );
#endif  STIIDEBUG
    switch ( rop )
    {
    case PRCO_GETOPT:
 	break;

    case PRCO_SETOPT:
	break;
  }

    return ( EOPNOTSUPP /*0*/ );
}

/*
S* st2_drain ()
s*
s*	NOTYET
s*
 */
#ifdef NOTYET
|	
|	  int
|st2_drain()			/* from <nowhere/vector> */
|	{
|	}
#endif NOTYET

/*
S* st2_input ( mbufp, ifp )
s*
s*	NOTYET
s*
 */
#ifdef NOTYET
|
|  int					/* Next bugid 0x30d02 */
|st2_input( mbufp, ifp )			/* from <nowhere/vector> */
|  struct aPktDesc	*mbufp;
|  struct ifnet		*ifp;		/* ARGSUSED */
|{
|    /* When get here, mbufp has (pulled-up) ip header w/len = data len */
|    /* and packet split across mbufs, at least once, second half contained */
|    /* ifnetp, 2-bytes pad, ethernet header, ip header, st2 header, scmp */
|
|    BUGSTOP (-,mbufp,0x30d01);
|    FreePkts (mbufp);
|
|    return ( 0 );
|}
#endif NOTYET

/*
S* st2_output ( mbufp, ifp )
s*
s*	NOTYET
s*
 */
#ifdef NOTYET
|
|  int					/* Next bugid 0x30e02 */
|st2_output( inpcbp , mbufp )		/* from <nowhere/vector> */
|  struct inpcb		*inpcbp;	/* ARGSUSED */
|  struct aPktDesc	*mbufp;
|{
|    BUGSTOP (-,mbufp,0x30e01);
|    FreePkts (mbufp);
|
|    /* to call ip output, mbuf has ip header w/o options */
|    /* with iplen = 20+data, ttl, prot, src, dst, data... */
|
|    return ( 0 );
|}
#endif NOTYET

/*
S* st2_ToAppl ( ifst2p, pktp, ifsockadr )
s*
s*	Network input routine for a application pseudo-network interface.
s*
 */
/* This is called via "if_output" for our pseudo interface. */

  int					/* Next bugid 0x30f11 */
st2_ToAppl( gifcp, pktp, ifsockadr )	/* from st2_Forward st2_CMPOutput */
					/* dsp if_output via ApplIfcCreate */
  struct aNetIf		*gifcp;		/* NOT osifchandle */
  struct aPktDesc	*pktp;		/* we free */
  struct sockaddr_st2	*ifsockadr;	/* ARGSUSED */
{
    char		*pb;
    int			 error = 0,
			 hdrbytes,
			 msgbytes;
    long		 totalbytes;
    unsigned long	 mandatory[1],
			*timestamp;
    struct aDetectorIPAdr *diap;
    struct aOrigin	*originp;
    struct aPktDesc	*ctrlp,
			*msgtailp,
			*pp,
			*sonamep;
    struct aReasonCode	*rcp;
    struct aSCMPHeader	*scmhdrp;
    struct aST2Header	*st2hdrp;
    struct aST2pcb	*cp;
    struct cmsghdr	*cmsghdrp;
    struct sockaddr_st2	*cpysadrp;
    struct socket	*so;


    if ( ApplIfcCheck( gifcp ) NE NoError )
    {
	FreePkts (pktp);
	return ( EINVAL );
    }

    /* We expect exactly one packet in chain, but let's be careful. */
    /* Multiple packets corrupts statistics -- all delivered or all dropped */
    msgtailp = pktp;

 more:

    if ( msgtailp EQ NO_PKTDSCP )
    {
	return ( error );
    }
#ifdef STIIDEBUG
    else if ( msgtailp->pkt_nxtmsgp NE NO_PKTDSCP )
    {
	BUGSTOP (,msgtailp->pkt_nxtmsgp,0x30f01); /* Got a chain */
    }
#endif STIIDEBUG

    pktp = msgtailp;
    msgtailp = pktp->pkt_nxtmsgp;
    pktp->pkt_nxtmsgp = NO_PKTDSCP;

    /* If interface is down (i.e., socket closed), just return. */
    if ( (gifcp->osifcp->if_flags & IFF_UP) EQ 0 )
    {
	BUGSTOP (,gifcp->osifcp->if_flags,0x30f02);
	FreePkts (pktp);
	error = ENETDOWN;
	goto more;
    }

    /* These are other indications that the application has disconnected.
     *  But someone should have killed the IFF_UP bit!
     */
    if (   ((cp = gifcp->ifc_cp) EQ NO_PCBP)
	OR ((so = cp->sop) EQ (struct socket *) NULL) )
    {
	BUGSTOP (-,cp,0x30f03);
	FreePkts (pktp);
	error = ENETDOWN;
	goto more;
    }

    /* Get rid of any non-data portion in the packet we receive. */
    while (    (pktp->pkt_type NE PKT_DATA)
	   AND (pktp->pkt_type NE PKT_IF) )
    {
	BUGSTOP (,pktp->pkt_type,0x30f04);
	pp = pktp->pkt_nxtbufp;
	pktp->pkt_nxtbufp = NO_PKTDSCP;
	FreePkt (pktp);
	pktp = pp;
	if ( pktp EQ NO_PKTDSCP )
	{
	    error = EINVAL;
	    goto more;
	}
    }

    /* Get pointer to ST-II header (and insist it is an ST-II header) */
    st2hdrp = pkt2data (pktp, struct aST2Header *);
    hdrbytes = sizeofaST2Header;

    if ( st2hdrp->IPSTVer NE STV2 )
    {
	FreePkts (pktp);
	error = EINVAL;
	goto more;
    }

    if ( HasSTTimestamp ( st2hdrp ) )
    {
	timestamp = Mkp (unsigned long *, st2hdrp, hdrbytes);
	hdrbytes += sizeofaSTTimestamp;
    }
    else
	timestamp = 0;

    /* ??? verify suficient length */
    /* HID zero means ST-II CMP packet. */
    if ( st2hdrp->HID NE (HID) 0 )
    {					/* Data packet */
	scmhdrp = (struct aSCMPHeader *) NULL;
    }
    else
    {					/* Control packet */
	scmhdrp = Mkp (struct aSCMPHeader *, st2hdrp, hdrbytes);
	hdrbytes += sizeofaSCMPHeader;

	switch ( scmhdrp->OpCode )
	{
	/*
	 * Connect and accept both indicate the stream is
	 * (partially) established so data and control traffic
	 * are now valid.
	 */
	case hton1 (OpConnect):		/* 5 we are target */

	    /* ??? duplicate processing, filter to next layer */

	    Bcopy (&( gifcp->ifc_gwcp->soname ),&( gifcp->ifc_cp->soname ),
				sizeof (struct sockaddr_st2soname));
	    gifcp->ifc_cp->soname.applent.pcode = (int) STRmtAppEnt;

	    Bcopy (&( gifcp->ifc_gwcp->Name ),&( gifcp->ifc_cp->Name ),
				sizeof (struct aName));

	    Bcopy (&( gifcp->ifc_gwcp->Origin ),&( gifcp->ifc_cp->Origin ),
				sizeof (struct aOrigin));

	    Bcopy (&( gifcp->ifc_gwcp->flow_spec ),&( gifcp->ifc_cp->flow_spec ),
			sizeof (struct aFlowSpec3)); /* ??? other versions */

	    /* ??? must implement */
	    gifcp->ifc_cp->precedence = gifcp->ifc_gwcp->precedence;

	    /* ??? from listener */
	    gifcp->ifc_cp->optapi = gifcp->ifc_gwcp->optapi;
	    gifcp->ifc_cp->detectorIPAdr = gifcp->ifc_gwcp->detectorIPAdr;

	    goto makeopen;

	case hton1 (OpAccept):		/* 1 we are origin */
	    if (  ((int) scmhdrp->Options & (int) STTSR)
		< gifcp->ifc_cp->optlvl )
	    {
		gifcp->ifc_cp->optlvl = (int) scmhdrp->Options & (int) STTSR;
		switch ( gifcp->ifc_cp->optlvl )
		{
		default:
		case STTSROpt:		/* Cannot be .LT. */
		    BUGSTOP (-,Inconsistency,0x30f05);
		    /* Fall through */

		case STTSRNotImp:
		case STTSRNo:		/* force it off */
		    gifcp->ifc_cp->optmask &= (unsigned char ) ~ STDataTBit;
		    break;

		case STTSRYes:		/* Force it on */
		    gifcp->ifc_cp->optneg = (unsigned char) STDataTBit;
		    break;
		} /* end of lower timestamp level switch */
	    }
    makeopen:
	    if ( (so->so_state & SS_ISCONNECTED) EQ 0 )
	    {
		SetState (PCB,cp,pcb_state,CINS_OPEN);
		soisconnected ( so );
		IfLog (DBG_SKT_RUT)
		    (void) printf ( "Socket %x now connected\n", so );
	    }
	    break;


	/*
	 * Disconnect, Notify and Refuse have a ReasonCode and
	 * Detector IP Addressin the header that must be reformatted
	 * to be passed to the application.
	 */
	case hton1 (OpDisconnect):	/* 6 we are target */
	case hton1 (OpNotify):		/* 14 */
	case hton1 (OpRefuse):		/* 15 we are origin */
	    break;

	case hton1 (OpChange):		/* 3 we are downstrean */
	case hton1 (OpChangeRequest):	/* 4 we are upstream */
	    break;

	default:
	    BUGSTOP (-,ntoh1 (scmhdrp->OpCode),0x30f06);
	    FreePkts (pktp);
	    error = EINVAL;
	    goto more;
	} /* end of control switch */
    }

    totalbytes = ntoh2 (st2hdrp->TotalBytes);
    IfLog (DBG_SKT_RUT)
	(void) printf ( "ST2 Data length %d\n", totalbytes );

    /*
     * Trim excess padding to match actual packet length.  This will
     * generally have been done already for control packets, but data
     * packets are not checked at the lower layer.  Data packets might
     * be fragmented across multiple buffers if the driver provided them
     * to the lower ST layer that way.
     */
    for ( pp = pktp ; pp NE NO_PKTDSCP ; pp = pp->pkt_nxtbufp )
    {
	if ( pp->pkt_length > totalbytes )
	    pp->pkt_length = totalbytes;
	totalbytes -= pp->pkt_length;
    }

    /* It's a truncated packet if more data was indicated in ST header. */
    if ( totalbytes > 0 )
    {
	/* count truncated (data) pkts */
	BUGSTOP (,pktp->pkt_length - ntoh2 (st2hdrp->TotalBytes),0x30f07);
	FreePkts (pktp);		/* drop truncated (data) */
	error = EINVAL;
	goto more;			/* ??? data vs control, so_error */
    }

    /* Skip over the header; but leave for copying to sockaddr */
    kassert (pktp->pkt_length >= hdrbytes);
    /* ??? what if it fails */

    pktp->pkt_offset += hdrbytes;
    pktp->pkt_length -= hdrbytes;
    if (  pktp->pkt_offset
	> (OffsetOf (pkt_prepad,struct aPktDesc) + sizeof (pktp->pkt_prepad)) )
	pktp->pkt_prepad += hdrbytes;
    else
	BUGSTOP (-,Inconsistency,0x30f08);

    /* Passing message to Unix socket layer requires we chain soname
     * and rights just before the data.
     */
    if ( so->so_proto->pr_flags & PR_ADDR )
    {
	sonamep = NO_PKTDSCP;
	originp = &( gifcp->ifc_gwcp->Origin );
	cpysadrp = GetSpace (struct sockaddr_st2 *, &( sonamep ),
			PKT_SONAME, PKT_MD_NEW, PKT_XA_PS, 0,0,
			sizeofsockaddr_st2header + (int) originp->plen, 0,0);
	if ( cpysadrp EQ (struct sockaddr_st2 *) NULL )
	{
	    BUGSTOP (-,1,0x30f09);
	    FreePkts (pktp);
	    error = ENOBUFS;
	    goto more;
	}

	IfLog (DBG_SKT_RUT)
	    (void) printf ( "...Sockaddr length %d\n", sonamep->pkt_length );

	cpysadrp->sa_family  = ConfigParm (coip_family);
	cpysadrp->st_request = 0;
	cpysadrp->st_options = ntoh1 (st2hdrp->STPriOpt);
	cpysadrp->st_handle  = 0;
	Bcopy (originp,&( cpysadrp->st_parms[0] ),originp->plen);
	cpysadrp->st_parms[0].pcode = (int) STRmtAppEnt;

	if ( timestamp NE 0 )
	{
	    cpysadrp->st_sec = ntoh4 (timestamp[0]);
	    cpysadrp->st_fract = ntoh4 (timestamp[1]);
	}
	else
	{
	    cpysadrp->st_sec = 0;
	    cpysadrp->st_fract = 0;
	}

	if ( scmhdrp NE (struct aSCMPHeader *) NULL )
	{
	    cpysadrp->st_request = scmhdrp->OpCode;
	    /* overrides ST Header info */
	    cpysadrp->st_options = scmhdrp->Options;

	    if ( (so->so_proto->pr_flags & PR_RIGHTS) NE 0 )
	    {
		/* ??? never have rights w/o addr? */
		ctrlp = NO_PKTDSCP;
		msgbytes = sizeofCMsgHdr;
		for ( pp = pktp ; pp NE NO_PKTDSCP ; pp = pp->pkt_nxtbufp )
		    msgbytes += pp->pkt_length;

		mandatory[0] = st2_parmManByOp[ (int) ntoh1 (scmhdrp->OpCode) ];
		if ( IsSet32 (mandatory, (int) STReasonCode) )
		    msgbytes += sizeof (struct aReasonCode);
		if ( IsSet32 (mandatory, (int) STDetectorIPAdr) )
		    msgbytes += sizeof (struct aDetectorIPAdr);

		cmsghdrp = GetSpace (struct cmsghdr *, &( ctrlp ),
				     PKT_CONTROL, PKT_MD_NEW, PKT_XA_PS,
				     0,0, msgbytes, 0,0);
		if ( cmsghdrp EQ (struct cmsghdr *) NULL )
		{
		    BUGSTOP (-,1,0x30f10);
		    FreePkts (sonamep);
		    FreePkts (pktp);
		    error = ENOBUFS;
		    goto more;
		}

		IfLog (DBG_SKT_RUT)
		    (void) printf ( "...Rights length %d\n", msgbytes );

		cmsghdrp->cmsg_len = msgbytes;
		cmsghdrp->cmsg_level = SOL_STII;
		cmsghdrp->cmsg_type = (int) scmhdrp->OpCode;	/* APIEQ */

		pb = (char *) &( cmsghdrp->cmsg_data[0] );

		if ( IsSet32 (mandatory, (int) STReasonCode) )
		{
		    rcp = Mkp (struct aReasonCode *,pb,0);
		    rcp->pcode = (octet1) STReasonCode;
		    rcp->plen = (octet1) sizeof (struct aReasonCode);
		    rcp->reason = scmhdrp->peropcode2; /* net order */
		    pb += sizeof (struct aReasonCode);
		}

		if ( IsSet32 (mandatory, (int) STDetectorIPAdr) )
		{
		    diap = Mkp (struct aDetectorIPAdr *,pb,0);
		    diap->pcode = (octet1) STDetectorIPAdr;
		    diap->plen = (octet1) sizeof (struct aDetectorIPAdr);
		    diap->pad2 = 0;
		    diap->IPAdr = scmhdrp->peropcode4; /* net order */
		    pb += sizeof (struct aDetectorIPAdr);
		}

		while ( pktp NE NO_PKTDSCP )
		{
		    pp = pktp->pkt_nxtbufp;
		    Bcopy (pkt2data (pktp, char *),pb,pktp->pkt_length);
		    pb += pktp->pkt_length;
		    FreePkt (pktp);
		    pktp = pp;
		    /* ??? wouldn't it be easier to switch MT_DATA to
		           MT_RIGHTS (and correct stats) */
		}
		pktp = ctrlp;

	    } /* end of PR_RIGHTS */
	} /* end of control (scmhdrp) */

	sonamep->pkt_nxtbufp = pktp;
	pktp = sonamep;

    } /* end of PR_ADDR */

    /* ??? INLINE */
    /* Try to queue packet for application */
    if ( stsbappend( &( so->so_rcv ), pktp ) )
	sorwakeup ( so );		/* Succeeded, wakeup reader */
    else
    {
	FreePkts (pktp);		/* Failed, drop packet */
	error = ENOBUFS;
    }

    goto more;
}

/*
S* st2_usrreq ( sop, req, mp, namp, rightsp )
s*
s*	this function processes a user's request to the protocol
s*
s*	so - socket involved
s*	req - request # (defined in .h file)
s*	m - data (depends on request)
s*	nam - sockaddr_st2
s*	rights - access rights (not used in mchip or inet!)
s*
 */

  int					/* Next bugid 0x31003 */
st2_usrreq( sop, req, mp, namp, rightsp ) /* from */
  struct socket		*sop;
  int			 req;
  struct mbuf		*mp,
			*namp,
			*rightsp;
{
    int			 ctltyp,
			 error = 0,
			 namtyp,
			 oldpri;
    struct aST2pcb	*cp;
    struct cmsghdr	*cmsghdrp;
    enum STRequests	streq;


    IfLog (DBG_SKT_RUT)
    {
	namtyp = -1;
	ctltyp = -1;

	if ( namp NE (struct mbuf *) NULL )
	{
	    namtyp = (mtod (namp,struct sockaddr_st2 *))->st_request;
	}

	if ( rightsp NE (struct mbuf *) NULL )
	{
	    ctltyp = (mtod (rightsp,struct cmsghdr *))->cmsg_type;
	}

	(void) printf ( "usrreq %d (%s): %d (%s)/%d (%s) to socket %x",
		       req,
		       (req < DimensionOf (usrreqnames))
		       ? usrreqnames[ req ] : "?",
		       namtyp, (namtyp EQ -1) ? "na" :
		       (namtyp < DimensionOf (streqnames))
		       ? streqnames[ namtyp ] : "?",
		       ctltyp, (ctltyp EQ -1) ? "na" :
		       (ctltyp < DimensionOf (streqnames))
		       ? streqnames[ ctltyp ] : "?", sop );
    }

#ifdef NOTYET
|	    if ( req EQ PRU_CONTROL )	/* XXX here we catch ioctl calls */
|	    {
|		st2_ioctl ( sop, mp, namp, rightsp );
|		/* ??? return */
|	    }
#endif NOTYET


    /*
     * set processor priority and look up the protocol control block
     * for this socket.   If there isn't a PCB and this isn't PRU_ATTACH
     * then return an error!  
     */

    oldpri = splnet ();
    cp = sotoST2pcb (sop);		/* get PCB from socket struct */
    if ( cp NE NO_PCBP )
    {
	IfLog (DBG_SKT_RUT)
	    (void) printf ( ": pcb_state %d (%s)\n", cp->pcb_state,
			    (st2_npcb_states > (int) cp->pcb_state )
			    ? st2_pcb_states[ (int) cp->pcb_state ].namep
			    : "??" );
    }
    else
    {
	IfLog (DBG_SKT_RUT)
	    (void) printf ( "\n" );

	if ( req NE PRU_ATTACH )
	{
	    (void) splx ( oldpri );
	    return ( EINVAL );	/* an error */
	}
    }

    /*
     * now we switch on the user's request (from the socket layer)
     */

    switch ( req )
    {
    /* ------------------------------------------------------------ */
    /* pr_usrreq ( so, PRU_ATTACH, (struct mbuf *)0, (struct mbuf *)proto,
		   (struct mbuf *)0 ); */
					/* from socreate */
    case PRU_ATTACH:	/*  0 */	/* attach COIP to this (new) socket */
	if ( cp NE NO_PCBP )
	{
	    error = EISCONN;		/* already have a PCB? */
	    break;
	}

	/* reserve socket space for the socket [see top of the file] */

	/* ??? should be f (flowspec) */
	if ( (error = soreserve ( sop, ConfigParm (sendspace),
				  ConfigParm (recvspace) )) NE 0 )
	    break;

	/* create the PCB for this socket and init it */
	if ( (cp = st2_PCBGet()) EQ NO_PCBP )
	{
	    error = ENOBUFS;
	    /* ??? undo soreserve */
	    break;
	}

	/* ??? shouldn't need this if LISTEN & not 1sock */
	/* attach pseudo interface for to-socket traffic */
	/* pseudo interface to/from application */
	cp->gifcp = ApplIfcCreate( cp );
	if ( cp->gifcp EQ NO_NETIFP )
	{
	    st2_PCBDel( cp );
	    error = ENOBUFS;
	}

	cp->sop = sop;
	cp->refcnt++;
	cp->refwho |= RefSocket;
	sop->so_pcb = (caddr_t) cp;	/* link into socket */
	IfLog (DBG_CNT_PCB)
	    printf ( "refcnt: cp = %8x  refcnt = %2d  pru_attach       RefSocket\n",
		    cp, cp->refcnt );

	cp->flags |= LocalConnection;

	cp->soname.applent.NextPcol = (int) namp; /* Might be zero */
    
	/* fill in usrpid, primarily diagnostic */
	cp->usrpid = u.u_procp->p_pid;
	cp->usrfd = -1;
	break;

    /* ------------------------------------------------------------ */
    /* pr_usrreq (so, PRU_BIND, (struct mbuf *)0, namp, (struct mbuf *)0); */

    case PRU_BIND:	/*  2 */	/* from sobind */
	error = api_lbind( cp, mtod (namp,struct sockaddr_st2 *), namp->m_len );
	/* EADDRNOTAVAIL EINVAL ENOBUFS 0 */
	break;

    /* ------------------------------------------------------------ */
    /* pr_usrreq ( so, PRU_LISTEN, (struct mbuf *)0, (struct mbuf *)0,
		   (struct mbuf *)0 ); */

    case PRU_LISTEN:	/*  3 */	/* from solisten */

	UpdateApplId ();

	cp->flags |= LocalListener;	/* ??? also LocalTarget if 1sock */

	error = api_lbind( cp, (struct sockaddr_st2 *) NULL, 0 );
	/* EADDRNOTAVAIL EINVAL ENOBUFS 0 */

	if ( error EQ 0 )
	{
	    SetState (PCB,cp,pcb_state,CINS_LISTEN);
	}
	break; 

    /* ------------------------------------------------------------ */
    /* pr_usrreq (so, PRU_ACCEPT, (struct mbuf *)0, nam, (struct mbuf *)0); */

    case PRU_ACCEPT:	/*  5 */	/* from soaccept */
    {	struct sockaddr_st2 *sastp = mtod (namp,struct sockaddr_st2 *);
	struct aOrigin *Originp = &( cp->Origin );

	UpdateApplId ();

	cp->flags |= LocalTarget;

	kassert (Originp->plen + sizeofsockaddr_st2header <= MLEN);

	/* return origin to application */
	namp->m_len = sizeofsockaddr_st2header + (int) Originp->plen;
	Bzero (sastp,sizeof (*sastp));
	sastp->sa_family = ConfigParm (coip_family);
	sastp->st_options = cp->optapi;
	Bcopy (Originp,&( sastp->st_parms[0] ),Originp->plen);
	error = 0;
	break;
    }

    /* ------------------------------------------------------------ */
    /* pr_usrreq (so, PRU_CONNECT, (struct mbuf *)0, nam, (struct mbuf *)0); */
					/* from soconnect */
    case PRU_CONNECT:	/*  4 */	/* a connect () request */

	UpdateApplId ();

	/* ??? should just incorporate new targets into list & start processing
	   any in "before-open" state */

	if ( NOT IsState (PCB,cp,pcb_state,CINS_PREOPEN) )
	{				/* ??? or CINS_LISTEN, STOpt1socket */
	    error = EALREADY;
	    break;
	}

	cp->flags |= LocalOrigin;

	/* if  not bound then do so */
        if ( (error = api_lbind( cp, (struct sockaddr_st2 *) NULL, 0 )) NE 0 )
	    /* EADDRNOTAVAIL EINVAL ENOBUFS 0 */
            break;

	/*
	 * now we call st2_DoConnect ().   This will setup the rest
	 * of the PCB stuff and then call the routing/resource
	 * allocation module.
	 */
	error = st2_DoConnect( cp, mtod (namp,struct sockaddr_st2 *),
			       namp->m_len, (struct aFlowSpec *) NULL,
			       (struct aUserData *) NULL );
	/* NB: cp may no longer be valid */

	if ( (error EQ 0) AND (sop->so_pcb NE (caddr_t) NULL) )
	{
	    soisconnecting ( sop );
	    if (    ((cp = sotoST2pcb (sop)) NE NO_PCBP)
		AND IsState (PCB,cp,pcb_state,CINS_OPEN) )
	    {
		soisconnected ( sop );
	    }
	}
	/* else ??? mark socket as errored & wakeup */

	break;				/* done! */

    /* ------------------------------------------------------------ */
    /* pr_usrreq (so, PRU_SEND, top, addr, control); */
					/* from sosend w/o MSG_OOB */
    case PRU_SEND:	/*  9 */	/* we dispose of mp */
	streq = STReqUnspec;

	/* For Data: nam or no nam;
	/*  Control: nam w/o rights, rights w/o nam, or name & rights */
	if ( namp NE (struct mbuf *) NULL )
	{
	    if ( namp->m_len < sizeofsockaddr_st2header )
	    {
		error = EINVAL;
		FreePkts ((struct aPktDesc *) mp);
		break;
	    }
	    streq = (enum STRequests)
			(mtod (namp,struct sockaddr_st2 *))->st_request;
	}

	cmsghdrp = (struct cmsghdr *) NULL;

	if ( rightsp NE (struct mbuf *) NULL )
	{
	    if ( rightsp->m_len < sizeofCMsgHdr )
	    {
		error = EINVAL;
		FreePkts ((struct aPktDesc *) mp);
		break;
	    }

	    cmsghdrp = mtod (rightsp,struct cmsghdr *);

	    /* ??? Is there something else we should do if level is wrong? */
	    if ( cmsghdrp->cmsg_level NE SOL_STII )
	    {
		BUGSTOP (,cmsghdrp->cmsg_level,0x31001);
		FreePkts ((struct aPktDesc *) mp);
		error = EINVAL;
		break;
	    }
	    if ( streq EQ STReqUnspec )
		streq = (enum STRequests) cmsghdrp->cmsg_type;
	}

	if (   (rightsp NE (struct mbuf *) NULL)
	    OR (streq NE STReqUnspec)
	    OR (mp EQ (struct mbuf *) NULL) )
	{
	    error = api_control( sop, cp, mtod (namp,struct sockaddr_st2 *),
				 namp->m_len, cmsghdrp,
				 (struct aPktDesc *) mp );
	}
	else
	{
	    /* ??? Explicitly allow zero size packets */
	    if ( mp EQ (struct mbuf *) NULL )
	    {
		error = EINVAL;
		BUGSTOP (-,error,0x31002);
		break;
	    }

	    if ( namp NE (struct mbuf *) NULL )
	    {
		error = api_send( cp, NO_NETIFP,
			mtod (namp,struct sockaddr_st2 *), namp->m_len, mp );
	    }
	    else
	    {
		error = api_send( cp, NO_NETIFP,
			(struct sockaddr_st2 *) NULL, 0, mp );
	    }
	}
	break;

    /* ------------------------------------------------------------ */
    /* pr_usrreq ( so, PRU_RCVD, (struct mbuf *)0, (struct mbuf *)0,
		   (struct mbuf *)0 ); */
		/* from soreceive w/o MSG_OOB & so_state &
		   SS_ISCONFIRMING AND uio->uio_resid */
    /* pr_usrreq ( so, PRU_RCVD, (struct mbuf *)0, (struct mbuf *)flags,
		   (struct mbuf *)0, (struct mbuf *)0); */
		/* from soreceive w/o MSG_OOB & w/ MSG_PEEK & PR_WANTRCVD */

    case PRU_RCVD:	/*  8 */
	break;

    /* ------------------------------------------------------------ */
    /* so1: pr_usrreq (so1, PRU_CONNECT2, (struct mbuf *)0, (struct mbuf *)so2,
			(struct mbuf *)0); */

    case PRU_CONNECT2:	/* 17 */	/* from soconnect2 */
	error = EOPNOTSUPP;
	break;

    /* ------------------------------------------------------------ */
    case PRU_SENSE:	/* 12 */
	((struct stat *) mp)->st_blksize = sop->so_snd.sb_hiwat;
	/* ??? rest of structure */
	break;

    /* ------------------------------------------------------------ */
    /* pr_usrreq ( so, PRU_SHUTDOWN, (struct mbuf *)0, (struct mbuf *)0,
		   (struct mbuf *)0)); */
					/* from soshutdown & FWRITE */
    case PRU_SHUTDOWN:	/*  7 */	/* won't send any more data */
	socantsendmore ( sop );
	break;

    /* ------------------------------------------------------------ */
    /* Only get here on graceless close (ApplAbort) */
    /* Gracefull disconnect is via PRU_SEND of STReqDisconnect */
    /* from sodisconnect
       via soclose when SO_ISCONNECTED & ! SO_ISDISCONNECTING or
       via soconnect when ... */
    /* pr_usrreq ( so, PRU_DISCONNECT, (struct mbuf *)0, (struct mbuf *)0,
		   (struct mbuf *)0); */

    case PRU_DISCONNECT:/*  6 */	/* disconnect from peer */
	soisdisconnecting ( sop );	/* update state & wakeup sleepers */

	if ( NOT IsState (PCB,cp,pcb_state,CINS_POSTACC) )
	{
	    (void) st2_DoDisconnect( cp, ApplAbort, NO_TRGP,
				   (struct aUserData *) NULL );
					/* does disconnect_Forwarder */
	    /* cp may be invalid */	/* 	st2_PCBUpdate */
	}

	/* Let subsequent call to PRU_DETACH do finish the job */
	break;

    /* ------------------------------------------------------------ */
    /* from soabort via soclose when SO_ACCEPTCONN for so_q[0] */
    /* pr_usrreq ( so, PRU_ABORT, (struct mbuf *)0, (struct mbuf *)0,
		   (struct mbuf *)0)); */

    case PRU_ABORT:	/* 10 */
	soisdisconnecting ( sop );	/* update state & wakeup sleepers */
	/* Falls through */

    /* ------------------------------------------------------------ */
    /* from soclose when so_pcb */
    /* pr_usrreq ( so, PRU_DETACH, (struct mbuf *)0, (struct mbuf *)0,
		   (struct mbuf *)0); */

    case PRU_DETACH:	/*  1 */

	/* Break link from pcb so we don't get anything back */
	if ( sop->so_pcb NE (caddr_t) NULL )
	{
	    sop->so_pcb = (caddr_t) NULL;
	    cp->sop = (struct socket *) NULL;
	    cp->refcnt--;
	    cp->refwho &= ~ RefSocket;
	    IfLog (DBG_CNT_PCB)
		printf ( "refcnt: cp = %8x  refcnt = %2d  pru_detach       RefSocket\n",
			 cp, cp->refcnt );
	}

	/* Now tell the stream to go away */
	if ( NOT IsState (PCB,cp,pcb_state,CINS_POSTACC) )
	{
	    (void) st2_DoDisconnect( cp, ApplAbort, NO_TRGP,
				   (struct aUserData *) NULL );
					/* does disconnect_Forwarder */
	    /* cp may be invalid */	/* 	st2_PCBUpdate */
	}


	soisdisconnected ( sop );	/* update state & wakeup sleepers */


	/* Noop for PRU_DETACH since SS_NOFDREF is not yet set - soclose
	   will actually free it; actually freed for PRU_ABORT */
	sofree ( sop );

	break;

    /* ------------------------------------------------------------ */
    /* pr_usrreq (so, PRU_SEND, top, addr, control); */

    case PRU_SENDOOB:	/* 14 */	/* from sosend w/ MSG_OOB */
	/* Falls through */

    /* ------------------------------------------------------------ */
    /* pr_usrreq (so, PRU_SOCKADDR, 0, m, 0) */
    /* ??? this is broken, as we are not told how big the user's buffer is,
       and the return code doesn't seem to check either */

    case PRU_SOCKADDR:	/* 15 */	/* from getsockname */
	/* Falls through */

    /* ------------------------------------------------------------ */
    /* pr_usrreq (so, PRU_PEERADDR, 0, m, 0) */
    /* ??? this is broken, as we are not told how big the user's buffer is,
       and the return code doesn't seem to check either */

    case PRU_PEERADDR:	/* 16 */	/* from getpeername */
	/* Falls through */

    /* ------------------------------------------------------------ */
    /* pr_usrreq ( so, PRU_RCVOOB, m, (struct mbuf *)(flags & MSG_PEEK),
		   (struct mbuf *)0); */

    case PRU_RCVOOB:	/* 13 */	/* from soreceive w/ MSG_OOB */
	/* Falls through */

    case PRU_PROTOSEND:			/* ??? do we freeup any mp */
    case PRU_CONTROL:
    case PRU_SLOWTIMO:
    case PRU_FASTTIMO:
    case PRU_PROTORCV:
    default:
#ifdef  STIIDEBUG
	(void) printf ( "st2_usrreq: unknown req: %d\n", req );
#endif  STIIDEBUG
	error = EOPNOTSUPP;
	break;

    } /* end of switch req */

    (void) splx ( oldpri );
    return ( error );
}

/*
S* stsbappend ( sb, pktp )
s*
s*	Space permitting, adjust socket byte counts & queue pkt for
s*	application and return TRUE. Otherwise return FALSE for caller
s*	to dispose of packet.
s*
 * We have to do this outselves because sbappendaddr () assumes the sockaddr
 * is exactly 16 bytes long and sbappend () assumes there is no MT_SONAME
 * mbuf in the chain.
 */

  static int				/* Next bugid 0x31103 */
stsbappend( sb, pktp )			/* from st2_ToAppl */
  struct sockbuf	*sb;
  struct aPktDesc	*pktp;
{
    struct aPktDesc	*pp;
    long		 space = 0;


    for ( pp = pktp ; pp NE NO_PKTDSCP ; pp = pp->pkt_nxtbufp )
	space += pp->pkt_length;

    if ( space > sbspace ( sb ) )
    {
	BUGSTOP (,space,0x31101);
	BUGSTOP (,sbspace ( sb ),0x31102);
	return ( 0 );
    }

    for ( pp = pktp ; pp NE NO_PKTDSCP ; pp = pp->pkt_nxtbufp )
	sballoc ( sb, (struct mbuf *) pp );

    if ( (pp = (struct aPktDesc *) sb->sb_mb) NE NO_PKTDSCP )
    {
	while ( pp->pkt_nxtmsgp )
	    pp = pp->pkt_nxtmsgp;

	pp->pkt_nxtmsgp = pktp;
    }
    else
	sb->sb_mb = (struct mbuf *) pktp;

    return ( 1 );
}

/*
S* TargByApplIfc ( gifcp, cp )
s*
s*	Routine to locate intermediate Agent target which is linked to a
s*	given application pseudo-network interface.
s*
 */

  static struct Target *		/* Next bugid 0x31200 */
TargByApplIfc( gifcp, cp )		/* from api_control*2 */
  struct aNetIf		*gifcp;
  struct aST2pcb	*cp;
{
    struct aVLink	*vlinkp;
    struct Target	*targetp;


    ForAllTargets (cp, targetp)
    {
	if (    ((vlinkp = targetp->vlinkp) NE NO_VLINKP)
	    AND (   vlinkp->route.nxthopp->gifcp
		 EQ (struct aNetIf *) gifcp) )
	    return ( targetp );
    }
    return ( NO_TRGP );
}

/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
