#ifndef	_ST2_ROUTING_H_
#define _ST2_ROUTING_H_
#ifndef lint
static char rcsid_st2_routing_h[] = "\
@(#)  $Header: st2_routing.h,v 1.10 91/11/26 23:17:00 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint


/*
M* st2_routing.h  Definitions for ST-II Routing.
M*
 */

/*
m* Status:
m*   Features:
m*   Untested Features:
m*   Restrictions/Bugs:
m*   Things to do:
m*
 */


/* Module Revision History
 *
 *	bugid	b,0
 * $Log:	st2_routing.h,v $
 * Revision 1.10  91/11/26  23:17:00  clynn
 * *** empty log message ***
 * 
 * Revision 1.9  91/11/04  09:23:17  clynn
 * Updated for Public Domain Release.  Major changes: addition of Source
 * Routing, IP Encapsulation, HELLO protocol between neighbors, tracking
 * of neighbors, detection of component or agent failures & notification
 * to applications.  More consistant naming and format, including making
 * all external names begin with "st2_".  Moved some routines and data
 * structures to reduce external references.
 * 
 * Added flags for modes of ST-II communication.
 * 
 * Revision 1.8  91/05/28  16:17:23  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 * Revision 1.1  91/03/15  18:33:09  clynn
 * Initial revision
 */


/* External definitions */

#include <sys/types.h>
#include <sys/socket.h>
#include <net/route.h>

#include "st2_api.h"		/* Global Definitions, before st2.h */
#include "st2.h"		/* ST-II Implementation definitions,
				 * includes st2_api.h, before st2_cmp.h */
#include "st2_cmp.h"		/* SCM Protocol definitions */


#ifdef DOCUMENTATION

struct rtentry {
	u_long	rt_hash;		/* to speed lookups */
	struct	sockaddr rt_dst;	/* key */
	struct	sockaddr rt_gateway;	/* value */
	short	rt_flags;		/* up/down?, host/net */
	short	rt_refcnt;		/* # held references */
	u_long	rt_use;			/* raw # packets forwarded */
	struct	ifnet *rt_ifp;		/* the answer: interface to use */
};

#define	RTF_UP		0x1		/* route useable */
#define	RTF_GATEWAY	0x2		/* destination is a gateway */
#define	RTF_HOST	0x4		/* host entry (net otherwise) */
#define RTF_REINSTATE	0x8		/* re-instate route after timeout */
#define	RTF_DYNAMIC	0x10		/* created dynamically (redirect) */
#define	RTF_MODIFIED	0x20		/* modified dynamically (redirect) */

#endif DOCUMENTATION

/* Additions to Routing Table ... */

#define RTFX_ENCAP	0x8000		/* Use IP encapsulation */
#define RTFX_NOTIP	0x4000		/* Not an IP route */
#define RTFX_ST		0x2000		/* An ST route */
#define RTFX_SRCRUT	0x1000		/* Has aSrcRutp/SrcRutLen */

#define RTEGetaSrcRutp(rep)	((struct aSrcRut *) NULL) /* Not yet */
#define RTEGetSrcRutLen(rep)	(0)


#endif	_ST2_ROUTING_H_

/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
